Package: enrichit
Title: 'C++' Implementations of Functional Enrichment Analysis
Version: 0.2.1
Authors@R: person(given = "Guangchuang", family = "Yu",
               email = "guangchuangyu@gmail.com", role  = c("aut", "cre"))
Maintainer: Guangchuang Yu <guangchuangyu@gmail.com>
Description: Fast implementations of functional enrichment analysis methods using 'C++' via 'Rcpp'.
    Currently provides Over-Representation Analysis (ORA), Gene Set Enrichment Analysis (GSEA),
    Weighted Enrichment Analysis for ORA and GSEA, Network-based Set Enrichment Analysis (NSEA),
    multi-layer network-based enrichment, and multi-omics integration workflows. Additional
    features include early fusion at the feature level, late fusion at the pathway level,
    multi-omics contribution tracing, topology-aware explanation helpers, Bayesian term
    selection, and extremely fast Random Walk with Restart (RWR) using 'RcppEigen'. The
    enrichment methods build on GSEA by Subramanian et al. (2005)
    <doi:10.1073/pnas.0506580102>, the multilevel strategy derived from 'fgsea'
    by Korotkevich et al. (2021) <doi:10.1101/060012>, and network-based
    enrichment ideas described by Glaab et al. (2012)
    <doi:10.1093/bioinformatics/bts389>.
License: Artistic-2.0
Depends: R (>= 3.5.0)
Imports: Matrix, methods, Rcpp (>= 1.0.10), rlang, stats, yulab.utils
        (> 0.2.1)
LinkingTo: Rcpp, RcppEigen
Suggests: AnnotationDbi, BiasedUrn, clusterProfiler, DOSE, fgsea, gson,
        qvalue, testthat
Encoding: UTF-8
URL: https://yulab-smu.top/biomedical-knowledge-mining-book/
Config/roxygen2/version: 8.0.0
NeedsCompilation: yes
Packaged: 2026-08-04 17:28:59 UTC; HUAWEI
Author: Guangchuang Yu [aut, cre]
Repository: CRAN
Date/Publication: 2026-08-04 18:40:02 UTC

# Generated by roxygen2: do not edit by hand

S3method("$",enrichResult)
S3method("$",gseaResult)
S3method("[",enrichResult)
S3method("[",gseaResult)
S3method("[[",enrichResult)
S3method("[[",gseaResult)
S3method(as.data.frame,enrichResult)
S3method(as.data.frame,gseaResult)
S3method(dim,enrichResult)
S3method(dim,gseaResult)
S3method(geneID,compareClusterResult)
S3method(geneID,enrichResult)
S3method(geneID,gseaResult)
S3method(geneInCategory,compareClusterResult)
S3method(geneInCategory,enrichResult)
S3method(geneInCategory,gseaResult)
S3method(head,enrichResult)
S3method(head,gseaResult)
S3method(tail,enrichResult)
S3method(tail,gseaResult)
export(EXTID2NAME)
export(aggregate_enrichment)
export(aggregate_omics)
export(bayes_enrich)
export(bayes_summary)
export(classify_omics_pattern)
export(collapse_multilayer_scores)
export(extract_mnsea_subnetwork)
export(geneID)
export(geneInCategory)
export(get_mnsea_contribution)
export(get_omics_contribution)
export(gsea)
export(gseaScores)
export(gsea_gson)
export(gsfilter)
export(harmonize_ids)
export(mnsea)
export(mnsea_gson)
export(nsea)
export(nsea_gson)
export(ora)
export(ora_gson)
export(prepare_multilayer_network)
export(prepare_network)
export(propagate_multilayer)
export(select_features_for_ora)
export(setReadable)
exportClasses(compareClusterResult)
exportClasses(enrichResult)
exportClasses(gseaResult)
exportClasses(mnseaResult)
exportClasses(nseaResult)
exportMethods(show)
exportMethods(summary)
importClassesFrom(methods,data.frame)
importFrom(Matrix,Diagonal)
importFrom(Matrix,colSums)
importFrom(Matrix,rowSums)
importFrom(Matrix,sparseMatrix)
importFrom(Rcpp,sourceCpp)
importFrom(methods,.hasSlot)
importFrom(methods,is)
importFrom(methods,new)
importFrom(methods,show)
importFrom(rlang,check_installed)
importFrom(stats,median)
importFrom(stats,p.adjust)
importFrom(stats,pchisq)
importFrom(stats,pnorm)
importFrom(stats,qnorm)
importFrom(stats,setNames)
importFrom(utils,head)
importFrom(utils,str)
importFrom(utils,tail)
importFrom(yulab.utils,yulab_msg)
useDynLib(enrichit, .registration = TRUE)

# enrichit: C++ Implementations of Functional Enrichment Analysis

`enrichit` is part of the `clusterProfiler` family, serving as the underlying algorithm implementation layer. It focuses on fast core computation, standardized result objects, and reusable data-preparation layers for downstream visualization packages such as `enrichplot`.

The package now covers not only classical enrichment workflows such as **ORA** and **GSEA**, but also **weighted enrichment**, **network propagation-based enrichment**, **multi-omics early/late fusion**, **multi-layer topology fusion**, and **explanation-ready data extraction**.

## Installation

You can install the development version of `enrichit` from GitHub using `devtools`:

```r
# install.packages("devtools")
devtools::install_github("YuLab-SMU/enrichit")
```

## Scope

`enrichit` is designed around four layers:

- **Core enrichment engines**: ORA, GSEA, weighted ORA/GSEA, and GSON-aware variants.
- **Network-aware enrichment**: single-layer `nsea()` and multi-layer `mnsea()` workflows based on Random Walk with Restart.
- **Multi-omics integration**: early fusion at the feature level and late fusion at the pathway level.
- **Explanation-ready outputs**: contribution tables and topology-aware extraction helpers prepared for visualization in `enrichplot`.

## Feature Map

- **High performance core**: key algorithms are implemented in `C++` via `Rcpp`, with sparse network propagation powered by `RcppEigen`.
- **ORA**: standard hypergeometric ORA with optional weighted ORA through Wallenius' noncentral hypergeometric distribution.
- **GSEA**: multilevel, permutation, and adaptive strategies for ranked enrichment analysis.
- **GSON support**: native `ora_gson()` and `gsea_gson()` interfaces for structured gene set collections.
- **NSEA**: `nsea()` and `nsea_gson()` for network-ranked enrichment on a single graph, including `mode = "signed"` for bidirectional propagation.
- **Multi-layer topology fusion**: `mnsea()` and `mnsea_gson()` for multiplex or heterogeneous network propagation across multiple layers.
- **Multi-omics early fusion**: `aggregate_omics()`, `harmonize_ids()`, and `select_features_for_ora()` for feature-level integration before enrichment.
- **Multi-omics late fusion**: `aggregate_enrichment()` for pathway-level aggregation of multiple enrichment results.
- **Contribution tracing**: `get_omics_contribution()`, `classify_omics_pattern()`, and `get_mnsea_contribution()` for explanation-oriented summaries.
- **Topology-aware extraction**: `extract_mnsea_subnetwork()` for pathway-specific node/edge tables that can be passed to downstream visualization packages.
- **Bayesian compression**: `bayes_enrich()` and `bayes_summary()` for posterior-based term prioritization.

## Main APIs

### Classical enrichment

- `ora()`, `ora_gson()`
- `gsea()`, `gsea_gson()`
- `gseaScores()`

### Weighted enrichment

- `ora(..., weight = )`
- `ora_gson(..., weight = )`
- `gsea(..., weight = )`
- `gsea_gson(..., weight = )`

### Network-aware enrichment

- `prepare_network()`
- `nsea()`, `nsea_gson()`
- `prepare_multilayer_network()`
- `propagate_multilayer()`
- `collapse_multilayer_scores()`
- `mnsea()`, `mnsea_gson()`

### Multi-omics integration

- `aggregate_omics()`
- `harmonize_ids()`
- `select_features_for_ora()`
- `aggregate_enrichment()`

### Explanation helpers

- `get_omics_contribution()`
- `classify_omics_pattern()`
- `get_mnsea_contribution()`
- `extract_mnsea_subnetwork()`

## Result Objects

The package provides the standard enrichment result object model used by the `clusterProfiler` family:

- `enrichResult` for ORA-like workflows
- `gseaResult` for ranked enrichment workflows
- `nseaResult` for single-network propagation plus enrichment
- `mnseaResult` for multi-layer propagation, collapsed scores, and cached explanation tables

These objects are intended to support a clean separation of concerns across the `clusterProfiler` family:

- `enrichit` handles core computation, algorithm implementation, and explanation-ready data preparation
- `clusterProfiler` provides high-level biological interpretation workflows and general enrichment analysis interfaces
- `enrichplot` handles visualization
- `gson` provides a structured gene set resource layer for managing and exchanging gene set collections across the family
- knowledge-base-oriented downstream packages such as `DOSE`, `ReactomePA`, `meshes`, and `MicrobiomeProfiler` provide domain-specific annotation and interpretation layers

## Design Notes

- **Computation first**: this package prioritizes fast and robust numerical routines over plot helpers.
- **Decoupled architecture**: integration layers, propagation layers, and enrichment layers are exposed separately where useful.
- **Stable downstream interface**: explanation helpers return standard tables so that plotting logic can evolve independently in downstream packages.

## ----eval=FALSE---------------------------------------------------------------
# devtools::install_github("YuLab-SMU/enrichit")


## -----------------------------------------------------------------------------
library(enrichit)

# Simulate a universe of 1000 genes
universe <- paste0("Gene", 1:1000)

# Define gene sets
gene_sets <- list(
  PathwayA = paste0("Gene", 1:50),       # Genes 1-50
  PathwayB = paste0("Gene", 800:850)     # Genes 800-850
)

# Select 'significant' genes (e.g., top 20 genes)
# PathwayA should be enriched
sig_genes <- paste0("Gene", 1:20)

# Run ORA
ora_result <- ora(
  gene = sig_genes,
  gene_sets = gene_sets,
  universe = universe
)

# View results
as.data.frame(ora_result)


## -----------------------------------------------------------------------------
# Generate synthetic ranked gene list
set.seed(42)
geneList <- sort(rnorm(1000), decreasing = TRUE)
names(geneList) <- paste0("Gene", 1:1000)

# Define gene sets
# PathwayTop is enriched at the top (positive ES)
# PathwayBottom is enriched at the bottom (negative ES)
gene_sets <- list(
  PathwayTop = names(geneList)[1:50],
  PathwayBottom = names(geneList)[951:1000],
  PathwayRandom = sample(names(geneList), 50)
)

# Run GSEA using the multilevel method
gsea_result <- gsea(
  geneList = geneList,
  gene_sets = gene_sets,
  method = "multilevel",
  nPerm = 1000,    # Base permutations
  minGSSize = 10,
  maxGSSize = 500
)

# View results
head(gsea_result)


## ----eval=FALSE---------------------------------------------------------------
# # Assuming you have a GSON object 'g'
# # result <- gsea_gson(geneList = geneList, gson = g)


## -----------------------------------------------------------------------------
sessionInfo()

#' Gene Set Enrichment Analysis (GSEA)
#'
#' Perform Gene Set Enrichment Analysis (GSEA) using a ranked gene list.
#'
#' @inheritParams enrichit_params
#' @param weight A named numeric vector of weights for genes. The names should match the names of geneList. If provided, the geneList will be multiplied by the weight and resorted before GSEA (default: NULL).
#' @param eps Epsilon for multilevel methods (default: 1e-10). Sets the smallest p-value that can be estimated.
#' @param sampleSize Sample size for multilevel methods (default: 101).
#' @param seed Random seed for reproducibility (default: FALSE). If FALSE, a random seed is generated.
#' @param nPermSimple Number of permutations for the simple method (default: 1000).
#' @param scoreType Type of enrichment score calculation: "std", "pos", "neg" (default: "std").
#'
#' @return A data.frame with columns:
#' - **ID**: Gene set name
#' - **enrichmentScore**: Enrichment Score
#' - **NES**: Normalized Enrichment Score
#' - **pvalue**: Empirical p-value from permutation test
#' - **setSize**: Size of the gene set (number of genes found in geneList)
#' - **nPerm**: (adaptive mode only) Actual number of permutations used
#' - **rank**: Rank at which the maximum enrichment score is attained
#' - **leading_edge**: Leading edge statistics (tags, list, signal)
#' - **core_enrichment**: Genes in the leading edge, separated by '/'
#'
#' @examples
#' # Example data
#' stats <- rnorm(1000)
#' names(stats) <- paste0("Gene", 1:1000)
#' stats <- sort(stats, decreasing = TRUE)
#' 
#' gs1 <- paste0("Gene", 1:50)
#' gs2 <- paste0("Gene", 500:550)
#' gene_sets <- list(Pathway1 = gs1, Pathway2 = gs2)
#' 
#' # Use default fixed permutation method
#' result <- gsea(geneList=stats, gene_sets=gene_sets, nPerm=100)
#' 
#' # Use adaptive permutation for more accurate p-values
#' \donttest{
#' result_adaptive <- gsea(geneList=stats, gene_sets=gene_sets, adaptive=TRUE)
#' }
#'
#' @export
gsea <- function(geneList, gene_sets, 
                 weight = NULL,
                 minGSSize = 10,
                 maxGSSize = 500,
                 nPerm = 1000, 
                 exponent = 1.0, 
                 method = "multilevel",
                 adaptive = FALSE, 
                 minPerm = 101, 
                 maxPerm = 100000, 
                 pvalThreshold = 0.1, 
                 eps = 1e-10,
                 sampleSize = 101,
                 seed = FALSE,
                 nPermSimple = 1000,
                 scoreType = "std",
                 verbose = TRUE) {
    
    gene_sets <- validate_gene_sets(gene_sets)
    
    if (!is.null(weight)) {
        if (!is.numeric(weight) || is.null(names(weight))) {
            stop("weight must be a named numeric vector")
        }
        
        common_genes <- intersect(names(geneList), names(weight))
        if (length(common_genes) == 0) {
            stop("No common genes found between geneList and weight")
        }
        if (verbose && length(common_genes) < length(geneList)) {
            message("Weight provided: ", length(common_genes), " out of ", length(geneList), " genes are weighted.")
        }
        
        geneList <- geneList[common_genes]
        weight_aligned <- weight[common_genes]
        
        if (any(weight_aligned < 0, na.rm = TRUE)) {
            warning("Negative weights detected. Taking absolute values.")
            weight_aligned <- abs(weight_aligned)
        }
        
        # Apply weight
        geneList <- geneList * weight_aligned
        # Re-sort to maintain preranked requirement
        geneList <- sort(geneList, decreasing = TRUE)
    }
    
    method <- match.arg(method, c("sample", "permute", "multilevel"))
    scoreType <- match.arg(scoreType, c("std", "pos", "neg"))

    prepared <- prepare_gsea_inputs(geneList, scoreType, exponent)
    geneList <- prepared$geneList
    multilevelRanks <- prepared$multilevelRanks
    sampleSize <- normalize_multilevel_sample_size(sampleSize)
    if (isFALSE(seed)) {
        seed <- sample.int(1e9, 1)
    }
    
    # Intersect with geneList to get overlapping genes;
    # minGSSize/maxGSSize should constrain the overlap, not the raw set size
    # see https://github.com/YuLab-SMU/clusterProfiler/issues/824
    gene_sets <- lapply(gene_sets, intersect, names(geneList))

    # Filter by size
    idx <- get_geneSet_index(gene_sets, minGSSize, maxGSSize)
    if (sum(idx) == 0) {
        if (verbose) {
            msg <- paste("No gene sets have size between", minGSSize, "and", maxGSSize, "...")
            message(msg)
            message("--> return NULL...")
        }
        return (NULL)
    }
    gene_sets <- gene_sets[idx]   

    gene_set_names <- names(gene_sets)
    
    # Call appropriate C++ function
    if (method == "multilevel") {
        result <- gsea_multilevel_cpp(geneList = multilevelRanks, gene_sets = gene_sets, gene_set_names = gene_set_names,
                                 minPerm = minPerm, maxPerm = maxPerm, pvalThreshold = pvalThreshold,
                                 exponent = 1.0, method = method, eps = eps, sampleSize = sampleSize, seed = seed,
                                 nPermSimple = nPermSimple, scoreType = scoreType)
    } else if (adaptive) {
        result <- gsea_adaptive_cpp(geneList, gene_sets, gene_set_names, 
                                    minPerm, maxPerm, pvalThreshold, exponent, method, seed)
    } else {
        result <- gsea_cpp(geneList, gene_sets, gene_set_names, nPerm, exponent, method, seed)
    }
    
    # Rename columns to standard names
    if (!"ID" %in% names(result) && "GeneSet" %in% names(result)) names(result)[names(result) == "GeneSet"] <- "ID"
    if (!"enrichmentScore" %in% names(result) && "ES" %in% names(result)) names(result)[names(result) == "ES"] <- "enrichmentScore"
    if (!"pvalue" %in% names(result) && "PValue" %in% names(result)) names(result)[names(result) == "PValue"] <- "pvalue"
    if (!"setSize" %in% names(result) && "Size" %in% names(result)) names(result)[names(result) == "Size"] <- "setSize"

    # Add setSize if missing
    if (!"setSize" %in% names(result)) {
        # Calculate size based on gene_sets (filtered)
        # Note: gene_sets order must match result ID order
        # Assuming result ID matches gene_set_names order passed to C++
        # If C++ preserves order (it does), we can map
        set_sizes <- sapply(gene_sets, length)
        result$setSize <- set_sizes[result$ID]
    }

    # Add NES if missing (placeholder)
    if (!"NES" %in% names(result)) {
        result$NES <- NA
    }

    if (!all(c("rank", "leading_edge", "core_enrichment") %in% names(result))) {
        ledge_rank <- integer(nrow(result))
        ledge_str <- character(nrow(result))
        ledge_core <- character(nrow(result))
        for (i in seq_len(nrow(result))) {
            sid <- result$ID[[i]]
            gs <- gene_sets[[sid]]
            ledge <- gsea_leading_edge_details(geneList, gs, exponent = exponent, scoreType = scoreType)
            ledge_rank[[i]] <- ledge$rank
            ledge_str[[i]] <- ledge$leading_edge
            ledge_core[[i]] <- ledge$core_enrichment
        }

        if (!"rank" %in% names(result)) result$rank <- ledge_rank
        if (!"leading_edge" %in% names(result)) result$leading_edge <- ledge_str
        if (!"core_enrichment" %in% names(result)) result$core_enrichment <- ledge_core
    }

    # Sort by absolute NES (descending) or pvalue (ascending)
    if (nrow(result) > 0) {
        if (!all(is.na(result$NES))) {
            result <- result[order(abs(result$NES), decreasing = TRUE), ]
        } else {
             result <- result[order(result$pvalue, decreasing = FALSE), ]
        }
    }

    if (method == "multilevel" && any(is.na(result$pvalue))) {
        warning(
            "There were ", sum(is.na(result$pvalue)),
            " pathways for which P-values were not calculated properly due to unbalanced gene-level statistic values. ",
            "For such pathways pvalue, NES and log2err are set to NA. You can try to increase nPermSimple."
        )
    }
    if (method == "multilevel" && any(!is.na(result$pvalue) & result$pvalue == eps & is.na(result$log2err))) {
        warning(
            "For some pathways, in reality P-values are less than ", eps,
            ". You can set the eps argument to zero for better estimation."
        )
    }
    
    return(result)
}

normalize_multilevel_sample_size <- function(sampleSize) {
    sampleSize <- as.integer(sampleSize[[1]])
    if (!is.finite(sampleSize)) {
        stop("sampleSize must be a finite integer")
    }
    sampleSize <- max(sampleSize, 3L)
    if (sampleSize %% 2L == 0L) {
        sampleSize <- sampleSize + 1L
    }
    sampleSize
}

prepare_gsea_inputs <- function(geneList, scoreType, exponent) {
    if (!is.numeric(geneList) || is.null(names(geneList))) {
        stop("geneList must be a named numeric vector")
    }
    if (any(!is.finite(geneList))) {
        stop("Not all stats values are finite numbers")
    }
    if (is.unsorted(rev(geneList))) {
        warning("geneList is not sorted in descending order. Sorting it now.")
        geneList <- sort(geneList, decreasing = TRUE)
    }

    ties <- sum(duplicated(geneList[geneList != 0]))
    if (ties != 0) {
        warning(
            "There are ties in the preranked stats (",
            round(ties * 100 / length(geneList), digits = 2),
            "% of the list). The order of those tied genes will be arbitrary, which may produce unexpected results."
        )
    }
    if (all(geneList > 0) && scoreType == "std") {
        warning(
            "All values in the stats vector are greater than zero and scoreType is \"std\", maybe you should switch to scoreType = \"pos\"."
        )
    }

    multilevelRanks <- scale_fgsea_ranks(geneList, exponent)

    list(
        geneList = geneList,
        multilevelRanks = multilevelRanks
    )
}

scale_fgsea_ranks <- function(geneList, exponent) {
    ranks <- abs(geneList)^exponent

    # Mirror fgsea::prepareStats(): scale to integer-like values based on the
    # total weight rather than a fixed multiplier, which keeps the multilevel
    # backend numerically closer to the long-used fgsea implementation.
    scaleCoeff <- 2^30 / sum(ranks)
    if (scaleCoeff >= 1) {
        scaleCoeff <- floor(scaleCoeff)
    }

    ranks <- round(ranks * scaleCoeff)
    storage.mode(ranks) <- "integer"
    structure(ranks, names = names(geneList))
}


#' generic function for gene set enrichment analysis
#'
#'
#' @title gsea_gson
#' @inheritParams enrichit_params
#' @param weight A named numeric vector of weights for genes.
#' @param ... Additional parameters passed to gsea()
#' @return gseaResult object
#' @author Guangchuang Yu
#' @export
gsea_gson <- function(geneList,
                 gson,
                 weight = NULL,
                 nPerm = 1000,
                 exponent = 1.0,
                 minGSSize = 10,
                 maxGSSize = 500,
                 pvalueCutoff = 0.05,
                 pAdjustMethod = "BH",
                 method = "multilevel",
                 adaptive = FALSE,
                 minPerm = 101,
                 maxPerm = 100000,
                 pvalThreshold = 0.1,
                 verbose = TRUE,
                 ...) {

    if (!inherits(gson, "GSON")) {
        stop("gson should be a GSON object")
    }

    # Ensure geneList is sorted
    if (is.unsorted(rev(geneList))) {
        if (verbose) {
            warning("geneList is not sorted in descending order. Sorting it now.")
        }
        geneList <- sort(geneList, decreasing = TRUE)
    }

    ## query external ID to Term ID
    gene <- names(geneList)
    
    # Extract gene sets from GSON
    gsid2gene <- gson@gsid2gene
    
    # ID Match Check
    if (!check_gene_id(gene, gsid2gene)) {
        return(NULL)
    }

    # Prepare Gene Sets
    geneSets <- split(gsid2gene$gene, gsid2gene$gsid)
    
    gsea_res <- gsea(geneList = geneList, 
                     gene_sets = geneSets, 
                     weight = weight,
                     minGSSize = minGSSize,
                     maxGSSize = maxGSSize,
                     nPerm = nPerm, 
                     exponent = exponent, 
                     method = method,
                     adaptive = adaptive,
                     minPerm = minPerm,
                     maxPerm = maxPerm,
                     pvalThreshold = pvalThreshold,
                     verbose = verbose,
                     ...)
                     
    if (is.null(gsea_res) || nrow(gsea_res) == 0) {
        return(NULL)
    }

    # Add Description
    gsid2name <- gson@gsid2name
    if (!is.null(gsid2name) && "ID" %in% names(gsea_res)) {
        description <- gsid2name$name[match(gsea_res$ID, gsid2name$gsid)]
        na_idx <- is.na(description)
        description[na_idx] <- gsea_res$ID[na_idx]
        gsea_res$Description <- description
    } else {
        if (!"Description" %in% names(gsea_res)) {
            gsea_res$Description <- gsea_res$ID
        }
    }

    # Calculate p.adjust
    gsea_res$p.adjust <- p.adjust(gsea_res$pvalue, method=pAdjustMethod)
    
    # Calculate qvalue
    gsea_res$qvalue <- calculate_qvalue(gsea_res$pvalue)
    
    # Filter by pvalueCutoff
    if (!is.null(pvalueCutoff)) {
        gsea_res <- gsea_res[!is.na(gsea_res$pvalue) & gsea_res$pvalue <= pvalueCutoff, ]
    }
    
    if (nrow(gsea_res) == 0) {
        return(NULL)
    }
    
    # Reorder columns
    expected_cols <- c("ID", "Description", "setSize", "enrichmentScore", "NES", "pvalue", "p.adjust", "qvalue", "rank", "leading_edge", "core_enrichment")
    
    # Ensure all expected columns exist to avoid error
    missing_cols <- setdiff(expected_cols, names(gsea_res))
    if (length(missing_cols) > 0) {
        warning("Missing columns in GSEA result: ", paste(missing_cols, collapse=", "), 
                ". Filling with NA. (Note: 'qvalue' may be missing if p-value distribution does not allow q-value calculation)")
        for (col in missing_cols) {
            gsea_res[[col]] <- NA
        }
    }

    if ("qvalues" %in% names(gsea_res) && !"qvalue" %in% names(gsea_res)) {
        gsea_res$qvalue <- gsea_res$qvalues
    }

    other_cols <- setdiff(names(gsea_res), expected_cols)
    gsea_res <- gsea_res[, c(expected_cols, other_cols)]
    
    # Set row names
    # Ensure IDs are robust (handle NA and duplicates)
    ids <- as.character(gsea_res$ID)
    
    # Handle NA IDs (convert to string "NA")
    if (any(is.na(ids))) {
        warning("NA values detected in gene set IDs. Replacing with string 'NA'.")
        ids[is.na(ids)] <- "NA"
    }
    
    # Handle Duplicate IDs
    if (any(duplicated(ids))) {
        dup_ids <- unique(ids[duplicated(ids)])
        warning("Duplicate gene set IDs detected: ", paste(head(dup_ids), collapse=", "), 
                "... (Total ", length(dup_ids), "). Unique suffixes added.")
        ids <- make.unique(ids)
    }
    
    rownames(gsea_res) <- ids
    
    params <- list(pvalueCutoff = pvalueCutoff,
                   nPerm = nPerm,
                   pAdjustMethod = pAdjustMethod,
                   exponent = exponent,
                   minGSSize = minGSSize,
                   maxGSSize = maxGSSize)
                   
    res <- new("gseaResult",
               result = gsea_res,
               organism = if (!is.null(gson@species)) gson@species else "UNKNOWN",
               setType = if (!is.null(gson@gsname)) gsub(".*;", "", gson@gsname) else "UNKNOWN",
               geneSets = geneSets,
               geneList = geneList,
               keytype = if (!is.null(gson@keytype)) gson@keytype else "UNKNOWN",
               permScores = matrix(), 
               params = params,
               gene2Symbol = character(), 
               readable = FALSE
              )
              
    return(res)
}

library(testthat)
library(enrichit)

test_that("GSEA function works correctly with both methods", {
  # Create synthetic data
  # 1000 genes, sorted
  stats <- sort(rnorm(1000), decreasing = TRUE)
  names(stats) <- paste0("Gene", 1:1000)
  
  # Create a gene set enriched at the top (should have positive ES)
  # Top 20 genes + some random ones
  gs_top <- c(names(stats)[1:20], names(stats)[sample(100:1000, 30)])
  
  # Create a gene set enriched at the bottom (should have negative ES)
  # Bottom 20 genes + some random ones
  gs_bottom <- c(names(stats)[981:1000], names(stats)[sample(1:900, 30)])
  
  # Create a random gene set (should have low ES)
  gs_random <- names(stats)[sample(1:1000, 50)]
  
  gene_sets <- list(
    TopEnriched = gs_top,
    BottomEnriched = gs_bottom,
    Random = gs_random
  )
  
  set.seed(123)
  
  # Test "sample" method (default)
  res_sample <- gsea(geneList = stats, gene_sets = gene_sets, nPerm = 100, method = "sample")
  
  expect_true(is.data.frame(res_sample))
  expect_true(all(c("ID", "enrichmentScore", "NES", "pvalue", "setSize", "rank", "leading_edge", "core_enrichment") %in% colnames(res_sample)))
  
  top_res <- res_sample[res_sample$ID == "TopEnriched", ]
  expect_gt(top_res$enrichmentScore, 0)
  expect_lt(top_res$pvalue, 0.05)
  
  # Test "permute" method
  res_permute <- gsea(geneList = stats, gene_sets = gene_sets, nPerm = 100, method = "permute")
  
  expect_true(is.data.frame(res_permute))
  
  top_res_perm <- res_permute[res_permute$ID == "TopEnriched", ]
  expect_gt(top_res_perm$enrichmentScore, 0)
  expect_lt(top_res_perm$pvalue, 0.05)
  
  # Compare NES (sample method usually produces higher NES magnitude for enriched sets)
  # Note: with small nPerm and synthetic data, this might not always hold, but generally true.
  # We just check that they are somewhat different but consistent in sign.
  expect_equal(sign(top_res$NES), sign(top_res_perm$NES))
  
  # Check that method argument validation works
  expect_error(gsea(geneList = stats, gene_sets = gene_sets, method = "invalid"))
})

test_that("Adaptive GSEA works correctly", {
  stats <- sort(rnorm(1000), decreasing = TRUE)
  names(stats) <- paste0("Gene", 1:1000)
  
  # Highly enriched set
  gs_top <- names(stats)[1:30]
  # Random set
  gs_random <- sample(names(stats), 30)
  
  gene_sets <- list(TopEnriched = gs_top, Random = gs_random)
  
  set.seed(42)
  res_adaptive <- gsea(geneList = stats, gene_sets = gene_sets, adaptive = TRUE, method = "sample",
                       minPerm = 100, maxPerm = 10000, pvalThreshold = 0.2)
  
  expect_true(is.data.frame(res_adaptive))
  expect_true("nPerm" %in% colnames(res_adaptive))
  
  # TopEnriched should have more permutations (significant)
  top_nPerm <- res_adaptive[res_adaptive$ID == "TopEnriched", "nPerm"]
  random_nPerm <- res_adaptive[res_adaptive$ID == "Random", "nPerm"]
  
  # Significant sets should use more permutations than initial minPerm
  expect_gte(top_nPerm, 100)
  # Random sets might stop early or not
  expect_gte(random_nPerm, 100)
})

test_that("Multilevel GSEA works correctly with new parameters", {
  stats <- sort(rnorm(1000), decreasing = TRUE)
  names(stats) <- paste0("Gene", 1:1000)
  
  # Top enriched
  gs_top <- names(stats)[1:30]
  # Bottom enriched
  gs_bottom <- names(stats)[971:1000]
  
  gene_sets <- list(Top = gs_top, Bottom = gs_bottom)
  
  # Test scoreType = "pos"
  set.seed(123)
  res_pos <- suppressWarnings(gsea(geneList = stats, gene_sets = gene_sets, method = "multilevel", 
                  scoreType = "pos", nPermSimple = 1000))
  
  expect_true(is.data.frame(res_pos))
  top_pos <- res_pos[res_pos$ID == "Top", ]
  expect_lt(top_pos$pvalue, 0.05)
  
  # Test scoreType = "neg"
  res_neg <- suppressWarnings(gsea(geneList = stats, gene_sets = gene_sets, method = "multilevel", 
                  scoreType = "neg", nPermSimple = 1000))
  
  expect_true(is.data.frame(res_neg))
  bottom_neg <- res_neg[res_neg$ID == "Bottom", ]
  expect_lt(bottom_neg$pvalue, 0.05)
  
  # Test scoreType = "std" (default)
  res_std <- suppressWarnings(gsea(geneList = stats, gene_sets = gene_sets, method = "multilevel", 
                  scoreType = "std", nPermSimple = 1000))
  
  expect_true(is.data.frame(res_std))
  expect_lt(res_std[res_std$ID == "Top", "pvalue"], 0.05)
  expect_lt(res_std[res_std$ID == "Bottom", "pvalue"], 0.05)
  
  # Check if nPermSimple parameter is accepted and works (by checking it doesn't crash)
  res_simple <- suppressWarnings(gsea(geneList = stats, gene_sets = gene_sets, method = "multilevel", 
                     nPermSimple = 500))
  expect_true(is.data.frame(res_simple))
})

test_that("Multilevel GSEA stays close to fgsea reference results", {
  skip_if_not_installed("fgsea")

  set.seed(1)
  stats <- rnorm(2000)
  names(stats) <- paste0("Gene", seq_along(stats))
  stats <- sort(stats, decreasing = TRUE)

  gene_sets <- list(
    Top = c(names(stats)[1:20], names(stats)[sample(100:2000, 20)]),
    Bottom = c(names(stats)[1981:2000], names(stats)[sample(1:1900, 20)]),
    Mixed = sample(names(stats), 40)
  )

  enrichit_res <- suppressWarnings(
    gsea(
      geneList = stats,
      gene_sets = gene_sets,
      method = "multilevel",
      eps = 0,
      sampleSize = 101,
      nPermSimple = 1000,
      scoreType = "std"
    )
  )
  fgsea_res <- suppressWarnings(
    fgsea::fgseaMultilevel(
      pathways = gene_sets,
      stats = stats,
      eps = 0,
      sampleSize = 101,
      nPermSimple = 1000,
      scoreType = "std",
      gseaParam = 1,
      minSize = 10,
      maxSize = 500
    )
  )
  fgsea_res <- as.data.frame(fgsea_res)

  cmp <- merge(
    enrichit_res[, c("ID", "enrichmentScore", "NES", "pvalue")],
    fgsea_res[, c("pathway", "ES", "NES", "pval")],
    by.x = "ID",
    by.y = "pathway",
    sort = FALSE
  )

  expect_equal(cmp$enrichmentScore, cmp$ES, tolerance = 1e-6)
  expect_equal(cmp$NES.x, cmp$NES.y, tolerance = 0.08)
  expect_equal(log10(cmp$pvalue), log10(cmp$pval), tolerance = 1)
})

test_that("weighted GSEA accepts lightweight gene weights", {
  set.seed(99)
  stats <- sort(rnorm(200), decreasing = TRUE)
  names(stats) <- paste0("Gene", seq_along(stats))
  gene_sets <- list(
    Top = names(stats)[1:20],
    Mixed = names(stats)[c(10:19, 120:129)]
  )
  weight <- setNames(rep(1, length(stats)), names(stats))
  weight[names(stats)[1:20]] <- 2

  res <- gsea(
    geneList = stats,
    gene_sets = gene_sets,
    weight = weight,
    nPerm = 50,
    method = "sample",
    verbose = FALSE
  )

  expect_true(is.data.frame(res))
  expect_true(all(c("ID", "enrichmentScore", "pvalue") %in% colnames(res)))
  expect_true("Top" %in% res$ID)
})

test_that("gsea_gson sorts the input gene list and keeps pathway descriptions", {
  skip_if_not_installed("gson")

  gsid2gene <- data.frame(
    gsid = c("Top", "Top", "Bottom", "Bottom"),
    gene = c("Gene1", "Gene2", "Gene5", "Gene6"),
    stringsAsFactors = FALSE
  )
  gsid2name <- data.frame(
    gsid = c("Top", "Bottom"),
    name = c("Top pathway", "Bottom pathway"),
    stringsAsFactors = FALSE
  )
  gson_obj <- gson::gson(
    gsid2gene = gsid2gene,
    gsid2name = gsid2name,
    species = "test",
    gsname = "test",
    version = "test",
    accessed_date = as.character(Sys.Date()),
    keytype = "SYMBOL"
  )

  stats <- c(Gene4 = 0.1, Gene1 = 3, Gene6 = -2, Gene2 = 2, Gene5 = -1)
  weight <- c(Gene1 = 2, Gene2 = 2, Gene4 = 1, Gene5 = 1, Gene6 = 1)

  res <- gsea_gson(
    geneList = stats,
    gson = gson_obj,
    weight = weight,
    pvalueCutoff = 1,
    minGSSize = 1,
    maxGSSize = 10,
    method = "sample",
    nPerm = 30,
    verbose = FALSE
  )

  expect_s4_class(res, "gseaResult")
  expect_equal(names(res@geneList), names(sort(stats, decreasing = TRUE)))
  expect_equal(unname(res@geneList), unname(sort(stats, decreasing = TRUE)))
  expect_equal(res@result$Description[match("Top", res@result$ID)], "Top pathway")
  expect_true(all(c("p.adjust", "qvalue", "core_enrichment") %in% colnames(res@result)))
})

test_that("gseaScores returns signed enrichment scores and fortify output", {
  geneList <- c(A = 4, B = 3, C = 1, D = -2, E = -4)

  expect_gt(gseaScores(geneList, c("A", "B")), 0)
  expect_lt(gseaScores(geneList, c("D", "E")), 0)
  expect_equal(gseaScores(geneList, character(0)), 0)

  running <- gseaScores(geneList, c("A", "B"), fortify = TRUE)
  expect_true(is.data.frame(running))
  expect_equal(nrow(running), length(geneList))
  expect_true(all(c("x", "runningScore", "position") %in% colnames(running)))
})

#' Aggregate multi-omics gene/protein-level statistics
#'
#' Aggregate multi-omics or multi-source statistics into a unified object for downstream enrichment analysis.
#'
#' @param x A list of named numeric vectors, a data.frame, or a matrix. Row names (or names for vectors) must represent feature IDs.
#' @param method Character, aggregation method. One of "fisher", "stouffer", "mean", or "max_abs".
#' @param input Character, input type. One of "pvalue" or "signed_score".
#' @param feature_type Character, type of the features (e.g., "gene", "protein"). Default is "gene".
#' @param conflict_policy Character, strategy to handle directional conflicts when input is "signed_score". One of "keep_all" (default, ignore conflicts), "strict" (set to NA if any signs conflict), or "penalty" (divide final score by 2 if signs conflict).
#' @param ... Additional arguments.
#'
#' @return An object of class `omics_aggregated` containing `score`, `pvalue` (if input is "pvalue"), `input_type`, `feature_type`, and `feature_id`.
#' @export
#' @importFrom stats p.adjust pchisq pnorm qnorm median
#' @importFrom rlang check_installed
aggregate_omics <- function(x, method = c("fisher", "stouffer", "brown", "mean", "weighted_mean", "max_abs"), 
                            input = c("pvalue", "signed_score"), feature_type = "gene", 
                            conflict_policy = c("keep_all", "strict", "penalty"), ...) {
    method <- match.arg(method)
    input <- match.arg(input)
    conflict_policy <- match.arg(conflict_policy)
    
    # Convert input to a matrix of features x omics
    if (is.list(x) && !is.data.frame(x)) {
        all_features <- unique(unlist(lapply(x, names)))
        mat <- matrix(NA_real_, nrow = length(all_features), ncol = length(x))
        rownames(mat) <- all_features
        colnames(mat) <- names(x)
        for (i in seq_along(x)) {
            mat[names(x[[i]]), i] <- as.numeric(x[[i]])
        }
    } else if (is.data.frame(x) || is.matrix(x)) {
        mat <- as.matrix(x)
        if (is.null(rownames(mat))) stop("Input must have row names representing feature IDs")
    } else {
        stop("x must be a list of named vectors, a data.frame, or a matrix")
    }
    
    res <- list(
        input_type = input,
        feature_type = feature_type,
        feature_id = rownames(mat)
    )
    
    if (input == "pvalue") {
        if (method == "fisher") {
            # Fisher's method: -2 * sum(ln(p))
            agg_p <- apply(mat, 1, function(p) {
                p <- p[!is.na(p)]
                if (length(p) == 0) return(NA_real_)
                p <- pmax(p, .Machine$double.xmin)
                stat <- -2 * sum(log(p))
                pchisq(stat, df = 2 * length(p), lower.tail = FALSE)
            })
        } else if (method == "stouffer") {
            # Stouffer's method: sum(Z) / sqrt(k)
            agg_p <- apply(mat, 1, function(p) {
                p <- p[!is.na(p)]
                if (length(p) == 0) return(NA_real_)
                p <- pmax(p, .Machine$double.xmin)
                p <- pmin(p, 1 - .Machine$double.eps)
                z <- qnorm(p, lower.tail = FALSE)
                z_stouffer <- sum(z) / sqrt(length(z))
                pnorm(z_stouffer, lower.tail = FALSE)
            })
        } else if (method == "brown") {
            # Brown's method (Empirical Brown's Method)
            args <- list(...)
            if (!is.null(args$cov_matrix)) {
                cov_mat <- args$cov_matrix
            } else {
                # Compute empirical covariance of -2 * log(p) across all features
                logP <- -2 * log(pmax(mat, .Machine$double.xmin))
                cov_mat <- stats::cov(logP, use = "pairwise.complete.obs")
                cov_mat[is.na(cov_mat)] <- 0
            }
            
            agg_p <- apply(mat, 1, function(p) {
                idx <- which(!is.na(p))
                k <- length(idx)
                if (k == 0) return(NA_real_)
                if (k == 1) return(p[idx])
                
                p_vals <- pmax(p[idx], .Machine$double.xmin)
                X <- -2 * sum(log(p_vals))
                
                VX <- sum(cov_mat[idx, idx])
                
                # If empirical variance is <= 4k, fallback to Fisher's independent assumption
                if (is.na(VX) || VX <= 4 * k) {
                    return(pchisq(X, df = 2 * k, lower.tail = FALSE))
                }
                
                c_factor <- VX / (4 * k)
                df <- (8 * k^2) / VX
                
                pchisq(X / c_factor, df = df, lower.tail = FALSE)
            })
        } else {
            stop("For input='pvalue', method must be 'fisher', 'stouffer', or 'brown'")
        }
        res$pvalue <- agg_p
        # Convert aggregated pvalue to score for ranking
        res$score <- -log10(pmax(agg_p, .Machine$double.xmin))
        
    } else if (input == "signed_score") {
        
        has_conflict <- apply(mat, 1, function(s) {
            s <- s[!is.na(s) & s != 0]
            if (length(s) <= 1) return(FALSE)
            return(length(unique(sign(s))) > 1)
        })
        
        if (method == "mean") {
            agg_s <- rowMeans(mat, na.rm = TRUE)
        } else if (method == "weighted_mean") {
            args <- list(...)
            weights <- args$weights
            if (is.null(weights)) {
                warning("method='weighted_mean' requires 'weights' argument. Falling back to simple mean.")
                agg_s <- rowMeans(mat, na.rm = TRUE)
            } else {
                if (length(weights) != ncol(mat)) {
                    stop("Length of 'weights' must match the number of omics columns.")
                }
                agg_s <- apply(mat, 1, function(s) {
                    idx <- !is.na(s)
                    if (sum(idx) == 0) return(NA_real_)
                    sum(s[idx] * weights[idx]) / sum(weights[idx])
                })
            }
        } else if (method == "max_abs") {
            agg_s <- apply(mat, 1, function(s) {
                s <- s[!is.na(s)]
                if (length(s) == 0) return(NA_real_)
                s[which.max(abs(s))]
            })
        } else {
            stop("For input='signed_score', method must be 'mean', 'weighted_mean', or 'max_abs'")
        }
        
        if (conflict_policy == "strict") {
            agg_s[has_conflict] <- NA_real_
        } else if (conflict_policy == "penalty") {
            agg_s[has_conflict] <- agg_s[has_conflict] / 2
        }
        
        res$score <- agg_s
    }
    
    # Remove features with NA scores
    valid_idx <- !is.na(res$score)
    res$feature_id <- res$feature_id[valid_idx]
    res$score <- res$score[valid_idx]
    names(res$score) <- res$feature_id
    
    if (!is.null(res$pvalue)) {
        res$pvalue <- res$pvalue[valid_idx]
        names(res$pvalue) <- res$feature_id
    }
    
    res$original_matrix <- mat[valid_idx, , drop = FALSE]
    
    class(res) <- "omics_aggregated"
    return(res)
}

#' Harmonize feature IDs to a target space
#'
#' Map protein-level or other feature-level statistics to a unified gene-level space.
#'
#' @param x A structured result from `aggregate_omics()`.
#' @param mapping A data.frame with `source_id` and `target_id` columns.
#' @param from Character, source feature type. Default is "protein".
#' @param to Character, target feature type. Default is "gene".
#' @param collapse Character, method to collapse multiple source IDs mapped to a single target ID. One of "max_abs", "mean", or "min_p".
#'
#' @return A harmonized `omics_aggregated` object.
#' @export
harmonize_ids <- function(x, mapping, from = "protein", to = "gene", collapse = c("max_abs", "mean", "min_p")) {
    collapse <- match.arg(collapse)
    if (!inherits(x, "omics_aggregated")) {
        stop("x must be an omics_aggregated object")
    }
    
    if (!all(c("source_id", "target_id") %in% colnames(mapping))) {
        stop("mapping must be a data.frame with 'source_id' and 'target_id' columns")
    }
    
    df <- data.frame(source_id = x$feature_id, score = x$score, stringsAsFactors = FALSE)
    if (!is.null(x$pvalue)) {
        df$pvalue <- x$pvalue
    }
    
    merged <- merge(df, mapping, by = "source_id", all.x = FALSE)
    if (nrow(merged) == 0) {
        warning("No IDs mapped successfully.")
        return(NULL)
    }
    
    res_list <- split(merged, merged$target_id)
    target_ids <- names(res_list)
    scores <- numeric(length(target_ids))
    pvals <- if (!is.null(x$pvalue)) numeric(length(target_ids)) else NULL
    
    for (i in seq_along(res_list)) {
        sub_df <- res_list[[i]]
        if (collapse == "max_abs") {
            idx <- which.max(abs(sub_df$score))
            scores[i] <- sub_df$score[idx]
            if (!is.null(pvals)) pvals[i] <- sub_df$pvalue[idx]
        } else if (collapse == "mean") {
            scores[i] <- mean(sub_df$score, na.rm = TRUE)
            if (!is.null(pvals)) pvals[i] <- mean(sub_df$pvalue, na.rm = TRUE)
        } else if (collapse == "min_p") {
            if (is.null(pvals)) stop("collapse='min_p' requires pvalue input in the aggregated object")
            idx <- which.min(sub_df$pvalue)
            scores[i] <- sub_df$score[idx]
            pvals[i] <- sub_df$pvalue[idx]
        }
    }
    
    names(scores) <- target_ids
    if (!is.null(pvals)) names(pvals) <- target_ids
    
    res <- list(
        input_type = x$input_type,
        feature_type = to,
        feature_id = target_ids,
        score = scores
    )
    if (!is.null(pvals)) res$pvalue <- pvals
    
    class(res) <- "omics_aggregated"
    return(res)
}

#' Select features for ORA
#'
#' Convert continuous aggregated statistics into a discrete list of genes and a universe for Over-Representation Analysis.
#'
#' @param x A structured result from `aggregate_omics()` or `harmonize_ids()`.
#' @param cutoff Numeric, the threshold to apply.
#' @param by Character, metric to apply the threshold on. One of "pvalue" or "score".
#' @param ... Additional arguments.
#'
#' @return A list containing `gene` (the selected feature IDs) and `universe` (all feature IDs).
#' @export
select_features_for_ora <- function(x, cutoff = 0.05, by = c("pvalue", "score"), ...) {
    by <- match.arg(by)
    if (!inherits(x, "omics_aggregated")) {
        stop("x must be an omics_aggregated object")
    }
    
    if (by == "pvalue") {
        if (is.null(x$pvalue)) stop("x does not contain pvalue. Use by='score' instead.")
        gene <- x$feature_id[x$pvalue < cutoff]
    } else if (by == "score") {
        gene <- x$feature_id[x$score > cutoff]
    }
    
    list(
        gene = gene,
        universe = x$feature_id
    )
}

#' Aggregate multiple enrichment results (Late Fusion)
#'
#' Combine pathway-level enrichment results from multiple omics or independent analyses.
#' P-values of identical pathways are merged using statistical methods (e.g., Brown's method).
#'
#' @param res_list A named list of enrichment result objects (e.g., `enrichResult`, `gseaResult`, `nseaResult`).
#' @param method Character, aggregation method for p-values. One of "brown", "fisher", or "stouffer".
#' @param ... Additional arguments passed to `aggregate_omics` (e.g., `cov_matrix` for Brown's method).
#'
#' @return An `enrichResult` object containing the aggregated p-values, FDR, and combined gene lists.
#' @export
aggregate_enrichment <- function(res_list, method = c("brown", "fisher", "stouffer"), ...) {
    method <- match.arg(method)
    
    if (!is.list(res_list) || length(res_list) < 2) {
        stop("res_list must be a list of at least two enrichment result objects.")
    }
    
    if (is.null(names(res_list))) {
        names(res_list) <- paste0("Omics_", seq_along(res_list))
    }
    
    meta <- list(organism = "UNKNOWN", ontology = "UNKNOWN", keytype = "UNKNOWN")
    df_list <- lapply(res_list, function(x) {
        if (inherits(x, c("enrichResult", "gseaResult", "nseaResult"))) {
            if (inherits(x, "enrichResult")) {
                meta$organism <<- x@organism
                meta$ontology <<- x@ontology
                meta$keytype <<- x@keytype
            } else {
                meta$organism <<- x@organism
                meta$ontology <<- x@setType
                meta$keytype <<- x@keytype
            }
            df <- x@result
        } else if (is.data.frame(x)) {
            df <- x
        } else {
            stop("Elements in res_list must be enrichResult, gseaResult, nseaResult, or data.frame.")
        }
        
        if (!all(c("ID", "pvalue") %in% colnames(df))) {
            stop("Each enrichment result must contain at least 'ID' and 'pvalue' columns.")
        }
        
        if (!"Description" %in% colnames(df)) {
            df$Description <- df$ID
        }
        
        df$ID <- as.character(df$ID)
        df <- df[!duplicated(df$ID), , drop = FALSE]
        df
    })
    
    non_empty <- vapply(df_list, nrow, integer(1)) > 0
    if (!all(non_empty)) {
        df_list <- df_list[non_empty]
        res_list <- res_list[non_empty]
    }
    if (length(df_list) < 2) {
        stop("At least two non-empty enrichment results are required.")
    }
    
    # Use the shared pathway universe to avoid silently treating missing terms as tested-but-null.
    common_ids <- Reduce(intersect, lapply(df_list, function(df) df$ID))
    if (length(common_ids) == 0) {
        return(.new_late_fusion_result(
            result_df = data.frame(),
            method = method,
            gene_sets = list(),
            meta = meta
        ))
    }
    
    # Build p-value matrix (rows = pathways, cols = omics)
    p_mat <- matrix(NA_real_, nrow = length(common_ids), ncol = length(df_list))
    rownames(p_mat) <- common_ids
    colnames(p_mat) <- names(res_list)
    
    # Build description map and combined gene lists
    desc_map <- setNames(rep("", length(common_ids)), common_ids)
    
    gene_list_map <- setNames(vector("list", length(common_ids)), common_ids)
    for (id in common_ids) {
        gene_list_map[[id]] <- character(0)
    }
    
    for (i in seq_along(df_list)) {
        df <- df_list[[i]]
        df <- df[match(common_ids, df$ID), , drop = FALSE]
        p_mat[, i] <- df$pvalue
        
        # Map descriptions (take the first non-NA/non-empty encountered)
        unmapped <- desc_map == "" | is.na(desc_map)
        if (any(unmapped)) {
            desc_map[unmapped] <- df$Description[unmapped]
        }
        
        # Combine genes from 'geneID' (ORA) or 'core_enrichment' (GSEA)
        gene_col <- if ("geneID" %in% colnames(df)) "geneID" else if ("core_enrichment" %in% colnames(df)) "core_enrichment" else NULL
        if (!is.null(gene_col)) {
            for (j in seq_len(nrow(df))) {
                genes <- unlist(strsplit(as.character(df[[gene_col]][j]), "/", fixed = TRUE))
                genes <- genes[nzchar(genes) & !is.na(genes)]
                gene_list_map[[common_ids[j]]] <- unique(c(gene_list_map[[common_ids[j]]], genes))
            }
        }
    }
    
    agg_res <- aggregate_omics(p_mat, method = method, input = "pvalue", ...)
    combined_p <- agg_res$pvalue
    
    # Drop pathways with NA aggregated p-value
    valid_idx <- !is.na(combined_p)
    combined_p <- combined_p[valid_idx]
    common_ids <- names(combined_p)
    
    padj <- stats::p.adjust(combined_p, method = "BH")
    qval <- calculate_qvalue(combined_p)
    
    combined_genes <- vapply(common_ids, function(id) paste(gene_list_map[[id]], collapse = "/"), character(1))
    counts <- vapply(common_ids, function(id) length(gene_list_map[[id]]), integer(1))
    
    res_df <- data.frame(
        ID = common_ids,
        Description = desc_map[common_ids],
        pvalue = combined_p,
        p.adjust = padj,
        qvalue = qval,
        geneID = combined_genes,
        Count = counts,
        stringsAsFactors = FALSE
    )
    
    res_df <- res_df[order(res_df$pvalue), ]
    rownames(res_df) <- res_df$ID
    
    gene_sets <- setNames(
        lapply(common_ids, function(id) gene_list_map[[id]]),
        common_ids
    )
    
    .new_late_fusion_result(
        result_df = res_df,
        method = method,
        gene_sets = gene_sets,
        meta = meta
    )
}

.new_late_fusion_result <- function(result_df, method, gene_sets, meta) {
    all_genes <- unique(unlist(gene_sets, use.names = FALSE))
    if (length(all_genes) == 0) {
        all_genes <- character(0)
    }
    
    new("enrichResult",
        result = result_df,
        pvalueCutoff = 1.0,
        pAdjustMethod = "BH",
        qvalueCutoff = 1.0,
        organism = meta$organism,
        ontology = if (isTRUE(nzchar(meta$ontology))) meta$ontology else "LateFusion",
        gene = all_genes,
        keytype = meta$keytype,
        universe = all_genes,
        gene2Symbol = character(0),
        geneSets = gene_sets,
        readable = FALSE,
        termsim = matrix(0, nrow = 0, ncol = 0),
        method = paste0("late_fusion_", method),
        dr = list()
    )
}

library(testthat)
library(enrichit)

make_enrich_result_for_test <- function(result_df, pvalue_cutoff = 0.05) {
  all_genes <- unique(unlist(strsplit(result_df$geneID, "/", fixed = TRUE), use.names = FALSE))
  gene_sets <- setNames(
    strsplit(result_df$geneID, "/", fixed = TRUE),
    result_df$ID
  )

  new("enrichResult",
    result = result_df,
    pvalueCutoff = pvalue_cutoff,
    pAdjustMethod = "BH",
    qvalueCutoff = 1,
    organism = "human",
    ontology = "TEST",
    gene = all_genes,
    keytype = "SYMBOL",
    universe = all_genes,
    gene2Symbol = character(0),
    geneSets = gene_sets,
    readable = FALSE,
    termsim = matrix(0, nrow = 0, ncol = 0),
    method = "ORA",
    dr = list()
  )
}

test_that("aggregate_enrichment combines shared pathways with Fisher method", {
  res1 <- data.frame(
    ID = c("Path1", "Path2"),
    Description = c("Path 1", "Path 2"),
    pvalue = c(0.001, 0.20),
    geneID = c("GeneA/GeneB", "GeneC"),
    Count = c(2, 1),
    stringsAsFactors = FALSE
  )
  res2 <- data.frame(
    ID = c("Path1", "Path2"),
    Description = c("Path 1", "Path 2"),
    pvalue = c(0.01, 0.03),
    core_enrichment = c("GeneB/GeneD", "GeneC/GeneE"),
    stringsAsFactors = FALSE
  )

  agg <- suppressWarnings(
    aggregate_enrichment(list(rna = res1, prot = res2), method = "fisher")
  )

  expect_s4_class(agg, "enrichResult")
  expect_equal(agg@result$ID, c("Path1", "Path2"))

  expected_p <- c(
    stats::pchisq(-2 * sum(log(c(0.001, 0.01))), df = 4, lower.tail = FALSE),
    stats::pchisq(-2 * sum(log(c(0.20, 0.03))), df = 4, lower.tail = FALSE)
  )
  expect_equal(agg@result$pvalue, expected_p, tolerance = 1e-12)
  expect_equal(agg@result$Count, c(3L, 2L))
  expect_equal(agg@result$geneID[agg@result$ID == "Path1"], "GeneA/GeneB/GeneD")
  expect_equal(agg@result$geneID[agg@result$ID == "Path2"], "GeneC/GeneE")
})

test_that("aggregate_enrichment reads raw enrichResult tables instead of filtered as.data.frame output", {
  res1_df <- data.frame(
    ID = c("Path1", "Path2"),
    Description = c("Path 1", "Path 2"),
    pvalue = c(0.001, 0.20),
    p.adjust = c(0.002, 0.20),
    qvalue = c(0.002, 0.20),
    geneID = c("GeneA/GeneB", "GeneC"),
    Count = c(2, 1),
    stringsAsFactors = FALSE
  )
  res2_df <- data.frame(
    ID = c("Path1", "Path2"),
    Description = c("Path 1", "Path 2"),
    pvalue = c(0.02, 0.03),
    p.adjust = c(0.02, 0.03),
    qvalue = c(0.02, 0.03),
    geneID = c("GeneB/GeneD", "GeneC/GeneE"),
    Count = c(2, 2),
    stringsAsFactors = FALSE
  )

  res1 <- make_enrich_result_for_test(res1_df, pvalue_cutoff = 0.05)
  res2 <- make_enrich_result_for_test(res2_df, pvalue_cutoff = 0.05)

  expect_equal(as.data.frame(res1)$ID, "Path1")

  agg <- suppressWarnings(
    aggregate_enrichment(list(res1, res2), method = "fisher")
  )

  expect_s4_class(agg, "enrichResult")
  expect_equal(sort(agg@result$ID), c("Path1", "Path2"))
  expect_equal(agg@ontology, "TEST")
  expect_equal(agg@organism, "human")
})

test_that("aggregate_omics supports Brown method with explicit covariance", {
  p_mat <- matrix(
    c(0.01, 0.02,
      0.20, 0.05),
    nrow = 2,
    byrow = TRUE,
    dimnames = list(c("Gene1", "Gene2"), c("rna", "prot"))
  )
  cov_mat <- matrix(c(4, 1, 1, 4), nrow = 2)

  res <- aggregate_omics(
    p_mat,
    method = "brown",
    input = "pvalue",
    cov_matrix = cov_mat
  )

  expect_s3_class(res, "omics_aggregated")
  expect_equal(names(res$pvalue), c("Gene1", "Gene2"))

  x1 <- -2 * sum(log(c(0.01, 0.02)))
  x2 <- -2 * sum(log(c(0.20, 0.05)))
  expected <- stats::pchisq(c(x1, x2) / 1.25, df = 3.2, lower.tail = FALSE)

  expect_equal(unname(res$pvalue), expected, tolerance = 1e-12)
  expect_equal(
    unname(res$score),
    -log10(expected),
    tolerance = 1e-12
  )
})

test_that("aggregate_omics handles signed conflicts and weighted mean", {
  score_mat <- matrix(
    c(1, 3,
      2, -4,
      -2, -6),
    nrow = 3,
    byrow = TRUE,
    dimnames = list(c("Gene1", "Gene2", "Gene3"), c("rna", "prot"))
  )

  strict_res <- aggregate_omics(
    score_mat,
    method = "mean",
    input = "signed_score",
    conflict_policy = "strict"
  )
  expect_equal(strict_res$feature_id, c("Gene1", "Gene3"))
  expect_equal(unname(strict_res$score), c(2, -4))

  penalty_res <- aggregate_omics(
    score_mat,
    method = "mean",
    input = "signed_score",
    conflict_policy = "penalty"
  )
  expect_equal(unname(penalty_res$score), c(2, -0.5, -4))

  weighted_res <- aggregate_omics(
    score_mat,
    method = "weighted_mean",
    input = "signed_score",
    weights = c(1, 3)
  )
  expect_equal(unname(weighted_res$score), c(2.5, -2.5, -5), tolerance = 1e-12)
})

test_that("harmonize_ids and select_features_for_ora keep the workflow lightweight", {
  agg <- structure(
    list(
      input_type = "pvalue",
      feature_type = "protein",
      feature_id = c("P1", "P2", "P3"),
      score = c(P1 = 4, P2 = 2, P3 = 1),
      pvalue = c(P1 = 1e-4, P2 = 1e-2, P3 = 0.20)
    ),
    class = "omics_aggregated"
  )

  mapping <- data.frame(
    source_id = c("P1", "P2", "P3"),
    target_id = c("G1", "G1", "G2"),
    stringsAsFactors = FALSE
  )

  harm <- harmonize_ids(agg, mapping, collapse = "min_p")

  expect_s3_class(harm, "omics_aggregated")
  expect_equal(harm$feature_type, "gene")
  expect_equal(harm$feature_id, c("G1", "G2"))
  expect_equal(unname(harm$pvalue), c(1e-4, 0.20))
  expect_equal(unname(harm$score), c(4, 1))

  ora_input <- select_features_for_ora(harm, cutoff = 0.05, by = "pvalue")
  expect_equal(ora_input$gene, "G1")
  expect_equal(ora_input$universe, c("G1", "G2"))
})

test_that("get_omics_contribution extracts pathway-level source statistics", {
  agg <- structure(
    list(
      input_type = "pvalue",
      feature_type = "gene",
      feature_id = c("GeneA", "GeneB", "GeneC"),
      score = c(GeneA = 4, GeneB = 2, GeneC = 1),
      pvalue = c(GeneA = 1e-4, GeneB = 1e-2, GeneC = 0.2),
      original_matrix = matrix(
        c(1e-4, 1e-2,
          1e-2, 0.2,
          0.2, 0.3),
        nrow = 3,
        byrow = TRUE,
        dimnames = list(c("GeneA", "GeneB", "GeneC"), c("rna", "prot"))
      )
    ),
    class = "omics_aggregated"
  )

  res_df <- data.frame(
    ID = c("Path1", "Path2"),
    Description = c("Path 1", "Path 2"),
    pvalue = c(0.001, 0.2),
    p.adjust = c(0.002, 0.2),
    qvalue = c(0.002, 0.2),
    geneID = c("GeneA/GeneB", "GeneC"),
    Count = c(2, 1),
    stringsAsFactors = FALSE
  )
  res <- make_enrich_result_for_test(res_df, pvalue_cutoff = 1)

  contrib <- get_omics_contribution(res, agg, pathway_id = "Path1")

  expect_true(is.data.frame(contrib))
  expect_equal(contrib$Feature, c("GeneA", "GeneB"))
  expect_true(all(c("rna", "prot", "Aggregated_Score", "Aggregated_Pvalue") %in% colnames(contrib)))
  expect_equal(contrib$Aggregated_Score, c(4, 2))
})

test_that("classify_omics_pattern labels merged pathways by single-omics support", {
  merged_df <- data.frame(
    ID = c("PathEnhanced", "PathRNA", "PathShared", "PathNS"),
    Description = c("Enhanced", "RNA only", "Shared", "Not sig"),
    pvalue = c(0.01, 0.01, 0.01, 0.2),
    p.adjust = c(0.01, 0.01, 0.01, 0.2),
    qvalue = c(0.01, 0.01, 0.01, 0.2),
    geneID = c("GeneA", "GeneB", "GeneC", "GeneD"),
    Count = c(1, 1, 1, 1),
    stringsAsFactors = FALSE
  )
  rna_df <- data.frame(
    ID = c("PathRNA", "PathShared"),
    Description = c("RNA only", "Shared"),
    pvalue = c(0.01, 0.02),
    p.adjust = c(0.01, 0.02),
    qvalue = c(0.01, 0.02),
    geneID = c("GeneB", "GeneC"),
    Count = c(1, 1),
    stringsAsFactors = FALSE
  )
  prot_df <- data.frame(
    ID = c("PathShared"),
    Description = c("Shared"),
    pvalue = c(0.03),
    p.adjust = c(0.03),
    qvalue = c(0.03),
    geneID = c("GeneC"),
    Count = c(1),
    stringsAsFactors = FALSE
  )

  merged <- make_enrich_result_for_test(merged_df, pvalue_cutoff = 1)
  rna <- make_enrich_result_for_test(rna_df, pvalue_cutoff = 1)
  prot <- make_enrich_result_for_test(prot_df, pvalue_cutoff = 1)

  classified <- classify_omics_pattern(
    merged_res = merged,
    single_res = list(rna = rna, prot = prot),
    p_cutoff = 0.05,
    by = "p.adjust"
  )

  out <- classified@result$Omics_Pattern
  names(out) <- classified@result$ID

  expect_equal(unname(out["PathEnhanced"]), "Enhanced (1+1>2)")
  expect_equal(unname(out["PathRNA"]), "rna-Specific")
  expect_equal(unname(out["PathShared"]), "Shared")
  expect_equal(unname(out["PathNS"]), "Not Significant")
})

#' Bayesian term selection for enrichment results
#'
#' `bayes_enrich()` adds a model-based selection layer on top of ORA results.
#' It estimates the posterior probability that each candidate term is an
#' active biological program explaining the observed input genes.
#'
#' The implementation uses a lightweight Metropolis-Hastings sampler over
#' binary latent term states. Given active terms, each gene is modeled as
#' observed with probability `1 - false_negative` if covered by at least one
#' active term, and with probability `false_positive` otherwise. The prior
#' probability that a candidate term is active is `prior`.
#'
#' This is intended as a result-compression and interpretation layer, not as a
#' replacement for ORA p-values.
#'
#' @param x An `enrichResult` object, typically from `ora_gson()` or a package
#'   that builds on enrichit, such as clusterProfiler.
#' @param candidate Candidate terms to include. `"significant"` uses
#'   `as.data.frame(x)`, `"all"` uses `x@result`, `"top"` uses the top
#'   `n_terms` rows from `x@result` ordered by `by`; or provide a character
#'   vector of term IDs.
#' @param n_terms Maximum number of candidate terms when `candidate = "top"` or
#'   when more candidates are supplied than this value. Use `Inf` to disable.
#' @param by Column used to order candidate terms.
#' @param prior Prior probability that a term is active.
#' @param false_positive Probability of observing a gene not covered by active
#'   terms.
#' @param false_negative Probability of missing a gene covered by active terms.
#' @param n_iter Total number of MCMC iterations.
#' @param burnin Number of initial iterations discarded.
#' @param thin Keep one sample every `thin` iterations after burn-in.
#' @param posterior_cutoff Terms with posterior greater than or equal to this
#'   value are marked active.
#' @param seed Optional random seed.
#' @param verbose Print sampler progress.
#' @return The input `enrichResult` object with additional columns in
#'   `@result`: `posterior`, `posterior_odds`, `bayes_rank`,
#'   `bayes_active`, `bayes_covered_gene`, and `bayes_covered_count`.
#' @export
bayes_enrich <- function(x,
                         candidate = c("top", "significant", "all"),
                         n_terms = 200,
                         by = "p.adjust",
                         prior = 0.1,
                         false_positive = 0.01,
                         false_negative = 0.1,
                         n_iter = 5000,
                         burnin = 1000,
                         thin = 1,
                         posterior_cutoff = 0.5,
                         seed = NULL,
                         verbose = FALSE) {
    if (!is(x, "enrichResult")) {
        stop("x should be an enrichResult object.", call. = FALSE)
    }

    .validate_bayes_enrich_args(
        prior = prior,
        false_positive = false_positive,
        false_negative = false_negative,
        n_iter = n_iter,
        burnin = burnin,
        thin = thin,
        posterior_cutoff = posterior_cutoff
    )

    all_res <- x@result
    if (nrow(all_res) == 0L) {
        return(x)
    }

    selected <- .select_bayes_terms(
        x = x,
        candidate = candidate,
        n_terms = n_terms,
        by = by
    )
    term_ids <- selected$term_ids

    gene_sets <- x@geneSets[term_ids]
    gene_sets <- gene_sets[!vapply(gene_sets, is.null, logical(1))]
    gene_sets <- lapply(gene_sets, as.character)
    term_ids <- names(gene_sets)

    if (length(term_ids) == 0L) {
        stop("No candidate terms have gene sets in x@geneSets.", call. = FALSE)
    }

    observed <- unique(as.character(x@gene))
    observed <- observed[!is.na(observed) & observed != ""]
    if (length(observed) == 0L) {
        stop("No observed genes available in x@gene.", call. = FALSE)
    }

    universe <- unique(as.character(x@universe))
    universe <- universe[!is.na(universe) & universe != ""]
    if (length(universe) == 0L) {
        universe <- unique(c(observed, unlist(gene_sets, use.names = FALSE)))
    }

    model_input <- .build_bayes_enrich_input(
        gene_sets = gene_sets,
        observed = observed,
        universe = universe
    )

    if (length(model_input$terms) == 0L) {
        stop("No candidate term overlaps the analysis universe.", call. = FALSE)
    }

    posterior <- .sample_active_terms(
        term_genes = model_input$term_genes,
        observed = model_input$observed,
        prior = prior,
        false_positive = false_positive,
        false_negative = false_negative,
        n_iter = n_iter,
        burnin = burnin,
        thin = thin,
        seed = seed,
        verbose = verbose
    )

    active <- posterior >= posterior_cutoff
    odds <- posterior / pmax(1 - posterior, .Machine$double.eps)
    ranks <- rank(-posterior, ties.method = "first")
    covered <- .posterior_covered_genes(
        term_genes = model_input$term_genes,
        observed = model_input$observed,
        genes = model_input$genes
    )

    out <- data.frame(
        ID = model_input$terms,
        posterior = posterior,
        posterior_odds = odds,
        bayes_rank = ranks,
        bayes_active = active,
        bayes_covered_gene = covered$gene,
        bayes_covered_count = covered$count,
        stringsAsFactors = FALSE
    )

    res <- merge(all_res, out, by = "ID", all.x = TRUE, sort = FALSE)
    ord <- match(all_res$ID, res$ID)
    res <- res[ord, , drop = FALSE]

    x@result <- res
    x@method <- if (length(x@method)) {
        paste0(x@method, "; bayes_enrich")
    } else {
        "bayes_enrich"
    }
    x
}

#' Summarize Bayesian enrichment results
#'
#' Return a data frame sorted by posterior probability from a result processed
#' by `bayes_enrich()`. This is a convenience wrapper around sorting
#' `as.data.frame(x)` by decreasing `posterior`.
#'
#' @param x An `enrichResult` object processed by `bayes_enrich()`.
#' @param active Logical. If `TRUE`, keep only rows with `bayes_active = TRUE`.
#' @param n Number of rows to return. Use `Inf` to return all rows.
#' @return A data frame ordered by decreasing `posterior`.
#' @export
bayes_summary <- function(x, active = FALSE, n = Inf) {
    if (!is(x, "enrichResult")) {
        stop("x should be an enrichResult object.", call. = FALSE)
    }

    res <- as.data.frame(x)
    if (!"posterior" %in% names(res)) {
        stop("x does not contain a posterior column. Run bayes_enrich(x) first.",
             call. = FALSE)
    }

    if (isTRUE(active)) {
        if (!"bayes_active" %in% names(res)) {
            stop("x does not contain a bayes_active column. Run bayes_enrich(x) first.",
                 call. = FALSE)
        }
        res <- res[!is.na(res$bayes_active) & res$bayes_active, , drop = FALSE]
    }

    if ("p.adjust" %in% names(res)) {
        ord <- order(-res$posterior, res$p.adjust, na.last = TRUE)
    } else {
        ord <- order(-res$posterior, na.last = TRUE)
    }
    res <- res[ord, , drop = FALSE]

    if (is.finite(n) && nrow(res) > n) {
        res <- res[seq_len(n), , drop = FALSE]
    }
    rownames(res) <- NULL
    res
}

.validate_bayes_enrich_args <- function(prior,
                                        false_positive,
                                        false_negative,
                                        n_iter,
                                        burnin,
                                        thin,
                                        posterior_cutoff) {
    probs <- c(
        prior = prior,
        false_positive = false_positive,
        false_negative = false_negative,
        posterior_cutoff = posterior_cutoff
    )
    if (any(!is.finite(probs)) || any(probs <= 0 | probs >= 1)) {
        stop("prior, false_positive, false_negative, and posterior_cutoff must be in (0, 1).",
             call. = FALSE)
    }
    if (n_iter <= 1L || burnin < 0L || burnin >= n_iter || thin < 1L) {
        stop("Require n_iter > 1, 0 <= burnin < n_iter, and thin >= 1.",
             call. = FALSE)
    }
}

.select_bayes_terms <- function(x, candidate, n_terms, by) {
    candidate_choices <- c("top", "significant", "all")
    all_res <- x@result
    if (is.character(candidate) &&
        (length(candidate) > 1L && !all(candidate %in% candidate_choices) ||
            length(candidate) == 1L && !candidate %in% candidate_choices)) {
        term_ids <- intersect(candidate, all_res$ID)
    } else {
        candidate <- match.arg(candidate, candidate_choices)
        res <- switch(candidate,
            significant = as.data.frame(x),
            all = all_res,
            top = all_res
        )
        term_ids <- res$ID
        if (candidate == "top") {
            if (!by %in% names(res)) {
                stop("Column specified by 'by' not found in enrichment result.", call. = FALSE)
            }
            vals <- res[[by]]
            ord <- if (is.numeric(vals)) order(vals, na.last = TRUE) else order(as.character(vals), na.last = TRUE)
            term_ids <- res$ID[ord]
        }
    }

    if (is.finite(n_terms) && length(term_ids) > n_terms) {
        term_ids <- term_ids[seq_len(n_terms)]
    }
    list(term_ids = unique(as.character(term_ids)))
}

.build_bayes_enrich_input <- function(gene_sets, observed, universe) {
    universe <- unique(universe)
    observed <- intersect(unique(observed), universe)
    if (length(observed) == 0L) {
        stop("No observed genes overlap the analysis universe.", call. = FALSE)
    }

    gene_sets <- lapply(gene_sets, function(gs) intersect(unique(gs), universe))
    keep <- vapply(gene_sets, length, integer(1)) > 0L
    gene_sets <- gene_sets[keep]

    terms <- names(gene_sets)
    genes <- unique(c(universe, observed))
    gene_index <- stats::setNames(seq_along(genes), genes)

    term_genes <- lapply(gene_sets, function(gs) unname(gene_index[gs]))
    observed_index <- unname(gene_index[observed])

    list(
        terms = terms,
        genes = genes,
        term_genes = term_genes,
        observed = observed_index
    )
}

.sample_active_terms <- function(term_genes,
                                 observed,
                                 prior,
                                 false_positive,
                                 false_negative,
                                 n_iter,
                                 burnin,
                                 thin,
                                 seed,
                                 verbose) {
    if (!is.null(seed)) {
        set.seed(seed)
    }

    n_terms <- length(term_genes)
    n_genes <- max(unlist(term_genes, use.names = FALSE), observed)
    y <- logical(n_genes)
    y[observed] <- TRUE

    active <- stats::rbinom(n_terms, size = 1L, prob = prior) == 1L
    cover_count <- integer(n_genes)
    if (any(active)) {
        for (idx in which(active)) {
            cover_count[term_genes[[idx]]] <- cover_count[term_genes[[idx]]] + 1L
        }
    }

    log_prior_on <- log(prior)
    log_prior_off <- log1p(-prior)
    log_fp <- log(false_positive)
    log_tn <- log1p(-false_positive)
    log_tp <- log1p(-false_negative)
    log_fn <- log(false_negative)

    gene_loglik <- function(cc, idx = NULL) {
        yy <- if (is.null(idx)) y else y[idx]
        covered <- cc > 0L
        ifelse(yy & covered, log_tp,
            ifelse(yy & !covered, log_fp,
                ifelse(!yy & covered, log_fn, log_tn)))
    }

    current_gene_score <- gene_loglik(cover_count)
    kept <- integer(n_terms)
    n_kept <- 0L

    for (iter in seq_len(n_iter)) {
        j <- sample.int(n_terms, 1L)
        affected <- term_genes[[j]]
        proposal_cover <- cover_count
        if (!active[j]) {
            proposal_cover[affected] <- proposal_cover[affected] + 1L
            prior_delta <- log_prior_on - log_prior_off
        } else {
            proposal_cover[affected] <- proposal_cover[affected] - 1L
            prior_delta <- log_prior_off - log_prior_on
        }

        proposal_gene_score <- gene_loglik(proposal_cover[affected], affected)
        score_delta <- sum(proposal_gene_score - current_gene_score[affected]) +
            prior_delta

        if (log(stats::runif(1L)) < score_delta) {
            active[j] <- !active[j]
            cover_count <- proposal_cover
            current_gene_score[affected] <- proposal_gene_score
        }

        if (iter > burnin && ((iter - burnin) %% thin == 0L)) {
            kept <- kept + as.integer(active)
            n_kept <- n_kept + 1L
        }

        if (verbose && iter %% 1000L == 0L) {
            message("bayes_enrich iteration ", iter, "/", n_iter)
        }
    }

    kept / n_kept
}

.posterior_covered_genes <- function(term_genes, observed, genes) {
    gene <- character(length(term_genes))
    count <- integer(length(term_genes))
    for (i in seq_along(term_genes)) {
        covered <- intersect(term_genes[[i]], observed)
        count[[i]] <- length(covered)
        gene[[i]] <- paste(genes[covered], collapse = "/")
    }
    list(gene = gene, count = count)
}

library(enrichit)

make_enrich_result <- function(gene_sets,
                               gene,
                               universe,
                               pvalueCutoff = 1,
                               qvalueCutoff = 1) {
    result <- data.frame(
        ID = names(gene_sets),
        Description = names(gene_sets),
        GeneRatio = paste0(
            vapply(gene_sets, function(x) length(intersect(x, gene)), integer(1)),
            "/",
            length(gene)
        ),
        BgRatio = vapply(gene_sets, function(x) paste0(length(x), "/", length(universe)), character(1)),
        RichFactor = 1,
        FoldEnrichment = 1,
        zScore = 0,
        pvalue = seq_along(gene_sets) / 10,
        p.adjust = seq_along(gene_sets) / 10,
        qvalue = seq_along(gene_sets) / 10,
        geneID = vapply(gene_sets, function(x) paste(intersect(x, gene), collapse = "/"), character(1)),
        Count = vapply(gene_sets, function(x) length(intersect(x, gene)), integer(1)),
        stringsAsFactors = FALSE
    )

    new(
        "enrichResult",
        result = result,
        pvalueCutoff = pvalueCutoff,
        pAdjustMethod = "BH",
        qvalueCutoff = qvalueCutoff,
        organism = "UNKNOWN",
        ontology = "UNKNOWN",
        gene = as.character(gene),
        keytype = "UNKNOWN",
        universe = as.character(universe),
        geneSets = gene_sets,
        readable = FALSE
    )
}

test_that("bayes_enrich annotates enrichResult with posterior columns", {
    gene_sets <- list(
        T1 = c("g1", "g2", "g3", "g4"),
        T2 = c("g1", "g2", "g3"),
        T3 = c("g8", "g9", "g10")
    )
    gene <- c("g1", "g2", "g3")
    universe <- paste0("g", 1:10)
    x <- make_enrich_result(gene_sets, gene, universe)

    y <- bayes_enrich(
        x,
        n_terms = Inf,
        n_iter = 300,
        burnin = 100,
        seed = 1
    )

    res <- as.data.frame(y)
    expect_s4_class(y, "enrichResult")
    expect_true(all(c(
        "posterior", "posterior_odds", "bayes_rank",
        "bayes_active", "bayes_covered_gene", "bayes_covered_count"
    ) %in% names(res)))
    expect_true(all(res$posterior[!is.na(res$posterior)] >= 0))
    expect_true(all(res$posterior[!is.na(res$posterior)] <= 1))
    expect_true(res$posterior[res$ID == "T2"] > res$posterior[res$ID == "T3"])

    tab <- bayes_summary(y)
    expect_true(all(diff(tab$posterior) <= 0))
    expect_identical(names(tab), names(res))

    top1 <- bayes_summary(y, n = 1)
    expect_equal(nrow(top1), 1)

    active_tab <- bayes_summary(y, active = TRUE)
    expect_true(all(active_tab$bayes_active))
})

test_that("bayes_enrich candidate modes use the expected term space", {
    gene_sets <- list(
        T1 = c("g1", "g2", "g3"),
        T2 = c("g1", "g2"),
        T3 = c("g8", "g9")
    )
    x <- make_enrich_result(
        gene_sets,
        gene = c("g1", "g2", "g3"),
        universe = paste0("g", 1:10),
        pvalueCutoff = 0.2,
        qvalueCutoff = 0.2
    )

    selected_sig <- .select_bayes_terms(
        x,
        candidate = "significant",
        n_terms = Inf,
        by = "p.adjust"
    )
    selected_all <- .select_bayes_terms(
        x,
        candidate = "all",
        n_terms = Inf,
        by = "p.adjust"
    )

    expect_identical(selected_sig$term_ids, as.character(as.data.frame(x)$ID))
    expect_identical(selected_all$term_ids, as.character(x@result$ID))
    expect_lt(length(selected_sig$term_ids), length(selected_all$term_ids))
})

test_that("bayes_enrich validates inputs", {
    gene_sets <- list(T1 = c("g1", "g2"))
    x <- make_enrich_result(
        gene_sets,
        gene = c("g1"),
        universe = c("g1", "g2", "g3")
    )

    expect_error(
        bayes_enrich(x, prior = 1),
        "must be in"
    )
    expect_error(
        bayes_enrich(x, n_iter = 10, burnin = 10),
        "Require"
    )
    expect_error(
        bayes_summary(x),
        "Run bayes_enrich"
    )
})

#' Get cached contribution tables from a `mnseaResult`
#'
#' @param res A `mnseaResult` object.
#' @param pathway_id Optional pathway ID. If `NULL`, returns all pathways for
#'   `level = "pathway"` and uses the top pathway for `level = "feature"`.
#' @param level One of `"pathway"` or `"feature"`.
#'
#' @return A data.frame containing cached contribution information.
#' @export
get_mnsea_contribution <- function(res, pathway_id = NULL, level = c("pathway", "feature")) {
    level <- match.arg(level)
    
    if (!inherits(res, "mnseaResult")) {
        stop("res must be a mnseaResult object.")
    }
    
    if (level == "pathway") {
        df <- res@pathway_contribution
        if (is.null(pathway_id)) {
            return(df)
        }
        return(df[df$ID == pathway_id, , drop = FALSE])
    }
    
    if (is.null(pathway_id)) {
        if (nrow(res@result) == 0) {
            stop("The mnseaResult is empty.")
        }
        pathway_id <- as.character(res@result$ID[1])
        message("pathway_id not provided. Using the top pathway: ", pathway_id)
    }
    
    df <- res@feature_contribution
    df[df$ID == pathway_id, , drop = FALSE]
}

#' show method for `gseaResult` instance
#'
#' @name show
#' @docType methods
#' @rdname show-methods
#'
#' @title show method
#' @return message
#' @importFrom methods show
#' @exportMethod show
#' @usage show(object)
#' @author Guangchuang Yu <https://yulab-smu.top>
setMethod("show", signature(object="gseaResult"),
          function (object){
              params <- object@params
              cat("#\n# Gene Set Enrichment Analysis\n#\n")
              .print_common_info(object)
              
              cat("#...@geneList", "\t")
              str(object@geneList)
              cat("#...nPerm", "\t", params$nPerm, "\n")
              cat(sprintf("#...pvalues adjusted by '%s' with cutoff < %s\n", 
                          params$pAdjustMethod, params$pvalueCutoff))
              cat(sprintf("#...%d enriched terms found\n", nrow(object@result)))
              str(object@result)
              cat("#...Citation\n")
              print_citation_msg(object@setType)
          }
)

#' show method for `nseaResult` instance
#'
#' @name show
#' @aliases show,nseaResult-method
#' @docType methods
#' @rdname show-methods
#'
#' @title show method
#' @param object A `nseaResult` instance.
#' @return message
#' @importFrom utils str
#' @importFrom methods show
#' @exportMethod show
#' @usage show(object)
#' @author Guangchuang Yu <https://yulab-smu.top>
setMethod("show", signature(object="nseaResult"),
          function (object){
              params <- object@params
              cat("#\n# Network-based Set Enrichment Analysis (NSEA)\n#\n")
              .print_common_info(object)
              
              cat("#...@mode", "\t", object@mode, "\n")
              cat(sprintf("#...@network\t%d nodes, %d edges\n", 
                          nrow(object@network), sum(object@network > 0)))
              cat(sprintf("#...@rwr\t%d iterations to converge (restart_prob = %s)\n", 
                          object@iterations, object@restart_prob))
              
              cat("#...@geneList (input)", "\t")
              str(object@geneList)
              cat("#...@diffusion_scores", "\t")
              str(object@diffusion_scores)
              
              cat(sprintf("#...pvalues adjusted by '%s' with cutoff < %s\n", 
                          params$pAdjustMethod, params$pvalueCutoff))
              cat(sprintf("#...%d enriched terms found\n", nrow(object@result)))
              str(object@result)
              cat("#...Citation\n")
              print_citation_msg(object@setType)
          }
)

#' show method for `mnseaResult` instance
#'
#' @name show
#' @aliases show,mnseaResult-method
#' @docType methods
#' @rdname show-methods
#'
#' @title show method
#' @param object A `mnseaResult` instance.
#' @return message
#' @importFrom utils str
#' @importFrom methods show
#' @exportMethod show
#' @usage show(object)
#' @author Guangchuang Yu <https://yulab-smu.top>
setMethod("show", signature(object="mnseaResult"),
          function (object){
              params <- object@params
              cat("#\n# Multi-layer Network-based Set Enrichment Analysis (MNSEA)\n#\n")
              .print_common_info(object)
              
              cat("#...@mode", "\t", object@mode, "\n")
              cat("#...@collapse", "\t", object@collapse_method, "\n")
              cat("#...@output_space", "\t", object@output_space, "\n")
              if (nzchar(object@target_layer)) {
                  cat("#...@target_layer", "\t", object@target_layer, "\n")
              }
              cat(sprintf("#...@layers\t%d layers, %d coupling edges\n",
                          length(object@layer_scores), nrow(object@coupling_table)))
              cat(sprintf("#...@rwr\t%d iterations to converge (restart_prob = %s)\n",
                          object@iterations, object@restart_prob))
              cat(sprintf("#...@explain\t%d pathway rows, %d feature rows cached\n",
                          nrow(object@pathway_contribution), nrow(object@feature_contribution)))
              
              cat("#...@collapsed_scores", "\t")
              str(object@collapsed_scores)
              cat(sprintf("#...pvalues adjusted by '%s' with cutoff < %s\n",
                          params$pAdjustMethod, params$pvalueCutoff))
              cat(sprintf("#...%d enriched terms found\n", nrow(object@result)))
              str(object@result)
              cat("#...Citation\n")
              print_citation_msg(object@setType)
          }
)


#' show method for `enrichResult` instance
#'
#' @name show
#' @docType methods
#' @rdname show-methods
#'
#' @title show method
#' @param object A `enrichResult` instance.
#' @return message
#' @importFrom utils str
#' @importFrom methods show
#' @exportMethod show
#' @usage show(object)
#' @author Guangchuang Yu <https://yulab-smu.top>
setMethod("show", signature(object="enrichResult"),
        function (object){
              
              cat("#\n# over-representation test\n#\n")
              .print_common_info(object)
              
              cat("#...@gene", "\t")
              str(object@gene)
              cat(sprintf("#...pvalues adjusted by '%s' with cutoff < %s\n", 
                          object@pAdjustMethod, object@pvalueCutoff))
              
              object <- get_enriched(object)
              n <- nrow(object@result)
              cat(sprintf("#...%d enriched terms found\n", n))
              if (n > 0) str(object@result)
              cat("#...Citation\n")
              print_citation_msg(object@ontology)
        }
)

#' @importFrom methods .hasSlot
.print_common_info <- function(object) {
    cat("#...@organism", "\t", object@organism, "\n")
    
    # Handle ontology or setType depending on object type
    if (.hasSlot(object, "ontology")) {
        cat("#...@ontology", "\t", object@ontology, "\n")
    } else if (.hasSlot(object, "setType")) {
        cat("#...@setType", "\t", object@setType, "\n")
    }
    
    kt <- object@keytype
    if (kt != "UNKNOWN") {
        cat("#...@keytype", "\t", kt, "\n")
    }
}


print_citation_msg <- function(ontology) {
    refs <- yulab.utils:::ref_knownledge()

    if (ontology == "HDO" || ontology == "NCG") {
        citation_msg <- refs["DOSE"]
    } else if (ontology == "Reactome") {
        citation_msg <- refs["ReactomePA"]
    } else if (ontology == "MeSH") {
        citation_msg <- refs["meshes"]
    } else {
        citation_msg <- refs["clusterProfiler_NP"]
    }
    cat(citation_msg, "\n\n")
}

#' Calculate GSEA Running Enrichment Scores
#'
#' @param geneList a named numeric vector of gene statistics (e.g., t-statistics or log-fold changes), sorted in decreasing order.
#' @param geneSet a character vector of gene IDs belonging to the gene set.
#' @param exponent a numeric value defining the weight of the running enrichment score. Default is 1.
#' @param fortify logical. If TRUE, returns a data frame with columns `x`, `runningScore`, and `position`. 
#' If FALSE (default), returns the enrichment score (ES).
#'
#' @return If `fortify = TRUE`, a data frame containing the running enrichment scores and positions. 
#' If `fortify = FALSE`, a numeric value representing the Enrichment Score (ES).
#'
#' @export
#' @author Guangchuang Yu
gseaScores <- function(geneList, geneSet, exponent=1, fortify=FALSE) {
    geneSet <- intersect(geneSet, names(geneList))
    
    if (length(geneSet) == 0) {
        # If no genes in set match geneList, ES is 0? Or NA?
        # User code doesn't handle this explicitly, but intersection would be empty.
        # If empty, hits are all false.
        # But let's let the C++ code handle it or return 0 here.
        if (fortify) {
            return(data.frame(x = seq_along(geneList), runningScore = 0, position = 0))
        }
        return(0)
    }
    
    in_set <- names(geneList) %in% geneSet
    
    res <- gsea_scores_cpp(geneList, in_set, exponent)
    
    if (fortify) {
        return(res)
    }
    
    # Calculate ES
    # ES is the maximum deviation from zero
    max.ES <- max(res$runningScore)
    min.ES <- min(res$runningScore)
    
    if (abs(max.ES) > abs(min.ES)) {
        return(max.ES)
    } else {
        return(min.ES)
    }
}

#' Over-Representation Analysis (ORA)
#'
#' Perform over-representation analysis using hypergeometric test (Fisher's exact test).
#'
#' @param gene Character vector of differentially expressed genes (or gene list of interest).
#' @param universe Character vector of background genes (e.g., all genes in the platform).
#' @param weight A named numeric vector of weights for background genes. If provided, Weighted ORA will be performed using Wallenius' noncentral hypergeometric distribution (requires 'BiasedUrn' package). The names should match the universe genes.
#' @inheritParams enrichit_params
#'
#' @return A data.frame with columns:
#' \item{GeneSet}{Gene set name}
#' \item{SetSize}{Number of genes in the gene set (intersected with universe)}
#' \item{DEInSet}{Number of differentially expressed genes in the gene set}
#' \item{DESize}{Total number of differentially expressed genes in universe}
#' \item{PValue}{Raw p-value from hypergeometric test}
#'
#' @examples
#' # Example data
#' de_genes <- c("Gene1", "Gene2", "Gene3", "Gene4", "Gene5")
#' all_genes <- paste0("Gene", 1:1000)
#' 
#' gs1 <- paste0("Gene", 1:50)
#' gs2 <- paste0("Gene", 51:150)
#' gs3 <- paste0("Gene", 151:300)
#' gene_sets <- list(Pathway1 = gs1, Pathway2 = gs2, Pathway3 = gs3)
#' 
#' result <- ora(gene=de_genes, gene_sets=gene_sets, universe=all_genes)
#' head(result)
#'
#' @export
ora <- function(gene, gene_sets, universe, weight = NULL) {
    
    # Validate inputs
    if (!is.character(gene)) {
        stop("gene must be a character vector")
    }
    if (!is.character(universe)) {
        stop("universe must be a character vector")
    }
    
    gene_sets <- validate_gene_sets(gene_sets)
    
    # Remove duplicates
    gene <- unique(gene)
    universe <- unique(universe)
    
    gene_set_names <- names(gene_sets)
    
    # Call C++ function through Rcpp (using the Rcpp-generated ora_cpp function)
    result <- ora_cpp(gene, universe, gene_sets, gene_set_names)
    
    if (!is.null(weight)) {
        if (!is.numeric(weight) || is.null(names(weight))) {
            stop("weight must be a named numeric vector")
        }
        rlang::check_installed('BiasedUrn', 'for weighted ORA.')
        
        weight_map <- weight[universe]
        names(weight_map) <- universe
        if (any(is.na(weight_map))) {
            warning("Some background genes do not have weights. Assigning median weight.")
            weight_map[is.na(weight_map)] <- median(weight, na.rm = TRUE)
        }
        if (any(weight_map <= 0, na.rm = TRUE)) {
            warning("Weights must be strictly positive for BiasedUrn. Adjusting non-positive weights.")
            min_pos <- min(weight_map[weight_map > 0], na.rm = TRUE)
            if (is.na(min_pos) || min_pos <= 0) min_pos <- 1e-6
            weight_map[weight_map <= 0] <- min_pos
        }
        
        pvals <- numeric(nrow(result))
        for (i in seq_len(nrow(result))) {
            gs_name <- result$GeneSet[i]
            gset <- gene_sets[[gs_name]]
            genes_in_set <- intersect(gset, universe)
            
            m1 <- length(genes_in_set)
            m2 <- length(universe) - m1
            x <- result$DEInSet[i]
            n <- result$DESize[i]
            
            if (m1 == 0 || x == 0) {
                pvals[i] <- 1.0
                next
            }
            
            w_in <- mean(weight_map[genes_in_set], na.rm = TRUE)
            w_out <- mean(weight_map[setdiff(universe, genes_in_set)], na.rm = TRUE)
            
            odds <- w_in / w_out
            if (is.na(odds) || !is.finite(odds)) odds <- 1.0
            
            pvals[i] <- 1 - BiasedUrn::pWNCHypergeo(x - 1, m1, m2, n, odds)
        }
        result$PValue <- pvals
    }
    
    # Rename columns to standard names
    # C++ returns: GeneSet, SetSize, DEInSet, DESize, UniverseSize, PValue, geneID
    names(result)[names(result) == "GeneSet"] <- "ID"
    names(result)[names(result) == "PValue"] <- "pvalue"
    names(result)[names(result) == "DEInSet"] <- "Count"
    
    # Calculate derived columns
    # GeneRatio = Count / DESize
    result$GeneRatio <- paste0(result$Count, "/", result$DESize)
    
    # BgRatio = SetSize / UniverseSize
    result$BgRatio <- paste0(result$SetSize, "/", result$UniverseSize)
    
    # RichFactor = Count / SetSize
    result$RichFactor <- result$Count / result$SetSize
    
    # FoldEnrichment = (Count/DESize) / (SetSize/UniverseSize)
    result$FoldEnrichment <- (result$Count / result$DESize) / (result$SetSize / result$UniverseSize)
    
    # Sort by p-value
    result <- result[order(result$pvalue), ]
    rownames(result) <- NULL
    
    return(result)
}

library(testthat)
library(enrichit)

# Define a helper function to compute expected p-values using phyper
expected_pvalue <- function(set_size, overlap, de_total, universe_size) {
  # hypergeometric: probability of >= overlap successes
  # q = overlap - 1, lower.tail = FALSE
  phyper(overlap - 1, set_size, universe_size - set_size, de_total, lower.tail = FALSE)
}

test_that("ORA function returns correct p-values and format", {
  # Example data from documentation
  de_genes <- c("Gene1", "Gene2", "Gene3", "Gene4", "Gene5")
  all_genes <- paste0("Gene", 1:1000)
  gs1 <- paste0("Gene", 1:50)
  gs2 <- paste0("Gene", 51:150)
  gs3 <- paste0("Gene", 151:300)
  gene_sets <- list(Pathway1 = gs1, Pathway2 = gs2, Pathway3 = gs3)

  result <- ora(gene = de_genes, gene_sets = gene_sets, universe = all_genes)

  # Verify columns exist
  expected_cols <- c("ID", "GeneRatio", "BgRatio", "RichFactor", "FoldEnrichment", "pvalue", "geneID", "Count")
  expect_true(all(expected_cols %in% colnames(result)))

  # Manual calculations
  manual_p1 <- expected_pvalue(set_size = 50, overlap = 5, de_total = 5, universe_size = 1000)
  manual_p2 <- expected_pvalue(set_size = 100, overlap = 0, de_total = 5, universe_size = 1000)
  manual_p3 <- expected_pvalue(set_size = 150, overlap = 0, de_total = 5, universe_size = 1000)

  # Compare with ORA output (allow tiny numerical tolerance)
  expect_equal(result$pvalue[result$ID == "Pathway1"], manual_p1, tolerance = 1e-10)
  expect_equal(result$pvalue[result$ID == "Pathway2"], manual_p2, tolerance = 1e-10)
  expect_equal(result$pvalue[result$ID == "Pathway3"], manual_p3, tolerance = 1e-10)
  
  # Check content of new columns for Pathway1
  p1_res <- result[result$ID == "Pathway1", ]
  expect_equal(p1_res$Count, 5)
  expect_equal(p1_res$GeneRatio, "5/5")
  expect_equal(p1_res$BgRatio, "50/1000")
  expect_equal(p1_res$RichFactor, 5/50)
  expect_equal(p1_res$FoldEnrichment, (5/5) / (50/1000))
  
  # Check that all expected genes are present in geneID
  genes_in_id <- strsplit(p1_res$geneID, "/")[[1]]
  expect_true(all(c("Gene1", "Gene2", "Gene3", "Gene4", "Gene5") %in% genes_in_id))
  expect_equal(length(genes_in_id), 5)
})

test_that("ora_gson excludes zero-overlap sets from multiple testing correction", {
  gsid2gene <- data.frame(
    gsid = c("set1", "set1", "set2", "set2", "set3", "set3"),
    gene = c("Gene1", "Gene2", "Gene2", "Gene3", "Gene4", "Gene5"),
    stringsAsFactors = FALSE
  )
  gsid2name <- data.frame(
    gsid = c("set1", "set2", "set3"),
    name = c("Set 1", "Set 2", "Set 3"),
    stringsAsFactors = FALSE
  )
  gson_obj <- gson::gson(
    gsid2gene = gsid2gene,
    gsid2name = gsid2name,
    species = "test",
    gsname = "test",
    version = "test",
    accessed_date = as.character(Sys.Date()),
    keytype = "UNKNOWN"
  )

  res <- ora_gson(
    gene = c("Gene1", "Gene2"),
    pvalueCutoff = 1,
    pAdjustMethod = "BH",
    universe = c("Gene1", "Gene2", "Gene3", "Gene4", "Gene5"),
    minGSSize = 1,
    maxGSSize = 10,
    qvalueCutoff = 1,
    gson = gson_obj
  )

  expect_s4_class(res, "enrichResult")
  expect_equal(res@result$ID, c("set1", "set2"))
  expect_true(all(res@result$Count > 0))

  raw_p <- c(
    expected_pvalue(set_size = 2, overlap = 2, de_total = 2, universe_size = 5),
    expected_pvalue(set_size = 2, overlap = 1, de_total = 2, universe_size = 5)
  )
  expect_equal(res@result$pvalue, raw_p, tolerance = 1e-10)
  expect_equal(res@result$p.adjust, p.adjust(raw_p, method = "BH"), tolerance = 1e-10)
})

test_that("weighted ORA runs on a small universe", {
  skip_if_not_installed("BiasedUrn")

  de_genes <- c("Gene1", "Gene2", "Gene3")
  all_genes <- paste0("Gene", 1:30)
  gene_sets <- list(
    Pathway1 = paste0("Gene", 1:6),
    Pathway2 = paste0("Gene", 10:18)
  )
  weight <- setNames(rep(1, 30), all_genes)
  weight[c("Gene1", "Gene2", "Gene3", "Gene4", "Gene5", "Gene6")] <- 3

  unweighted <- ora(gene = de_genes, gene_sets = gene_sets, universe = all_genes)
  weighted <- ora(gene = de_genes, gene_sets = gene_sets, universe = all_genes, weight = weight)

  expect_true(is.data.frame(weighted))
  expect_true(all(c("ID", "pvalue", "Count") %in% colnames(weighted)))
  expect_false(isTRUE(all.equal(weighted$pvalue, unweighted$pvalue)))
})

#' Prepare network for repeated NSEA runs
#'
#' @param network edge list (data.frame with 2 or 3 columns) or sparse matrix.
#' @param directed logical, whether the network is directed. Default is FALSE.
#' @param normalize one of "column", "row", or "none". Default is "column".
#'
#' @return A sparse matrix (dgCMatrix) that has been properly formatted and normalized.
#' @importFrom Matrix sparseMatrix
#' @importFrom Matrix colSums
#' @importFrom Matrix rowSums
#' @importFrom Matrix Diagonal
#' @importFrom stats setNames
#' @export
prepare_network <- function(network, directed = FALSE, normalize = "column") {
    if (inherits(network, "sparseMatrix")) {
        A <- network
    } else if (is.data.frame(network) || is.matrix(network)) {
        network <- as.data.frame(network)
        if (ncol(network) < 2) {
            stop("network must have at least 2 columns")
        }
        if (ncol(network) == 2) {
            network$weight <- 1
        }
        
        nodes <- unique(c(as.character(network[[1]]), as.character(network[[2]])))
        node_idx <- setNames(seq_along(nodes), nodes)
        
        i <- node_idx[as.character(network[[1]])]
        j <- node_idx[as.character(network[[2]])]
        x <- as.numeric(network[[3]])
        
        if (!directed) {
            i_all <- c(i, j)
            j_all <- c(j, i)
            x_all <- c(x, x)
        } else {
            i_all <- i
            j_all <- j
            x_all <- x
        }
        
        A <- Matrix::sparseMatrix(i = i_all, j = j_all, x = x_all,
                                  dims = c(length(nodes), length(nodes)),
                                  dimnames = list(nodes, nodes))
    } else {
        stop("network must be a data.frame, matrix or sparseMatrix")
    }
    
    if (normalize == "column") {
        cs <- Matrix::colSums(A)
        cs[cs == 0] <- 1 # avoid division by zero
        A <- A %*% Matrix::Diagonal(x = 1/cs)
    } else if (normalize == "row") {
        rs <- Matrix::rowSums(A)
        rs[rs == 0] <- 1
        A <- Matrix::Diagonal(x = 1/rs) %*% A
    }
    
    return(A)
}

build_nsea_result <- function(base_result,
                              gene_sets,
                              rwr_scores,
                              network,
                              mode,
                              iter,
                              p,
                              organism = "UNKNOWN",
                              setType = "UNKNOWN",
                              keytype = "UNKNOWN",
                              params = NULL) {
    if (inherits(base_result, "gseaResult")) {
        result_df <- base_result@result
        gene_sets <- base_result@geneSets
        organism <- base_result@organism
        setType <- base_result@setType
        keytype <- base_result@keytype
        params <- base_result@params
        gene_list <- base_result@geneList
        perm_scores <- base_result@permScores
        gene2symbol <- base_result@gene2Symbol
        readable <- base_result@readable
        termsim <- base_result@termsim
        method <- base_result@method
        dr <- base_result@dr
    } else if (is.data.frame(base_result)) {
        result_df <- base_result
        if (is.null(params)) {
            params <- list(pvalueCutoff = 1.0, pAdjustMethod = "BH")
        }
        gene_list <- rwr_scores
        perm_scores <- matrix(0, nrow = 0, ncol = 0)
        gene2symbol <- character(0)
        readable <- FALSE
        termsim <- matrix(0, nrow = 0, ncol = 0)
        method <- "NSEA"
        dr <- list()
    } else {
        stop("base_result must be a gseaResult or data.frame.")
    }
    
    if (!"Description" %in% colnames(result_df)) {
        result_df$Description <- result_df$ID
    }
    if (!"p.adjust" %in% colnames(result_df)) {
        result_df$p.adjust <- stats::p.adjust(result_df$pvalue, method = params$pAdjustMethod %||% "BH")
    }
    if (!"qvalue" %in% colnames(result_df)) {
        result_df$qvalue <- calculate_qvalue(result_df$pvalue)
    }
    rownames(result_df) <- as.character(result_df$ID)
    
    new("nseaResult",
        result = result_df,
        organism = organism,
        setType = setType,
        geneSets = gene_sets,
        geneList = gene_list,
        keytype = keytype,
        permScores = perm_scores,
        params = params,
        gene2Symbol = gene2symbol,
        readable = readable,
        termsim = termsim,
        method = method,
        dr = dr,
        network = network,
        diffusion_scores = rwr_scores,
        mode = mode,
        iterations = as.integer(iter),
        restart_prob = p)
}

`%||%` <- function(x, y) {
    if (is.null(x) || length(x) == 0) y else x
}

#' Network-based Gene Set Enrichment Analysis
#'
#' @param geneList named numeric vector. In "evidence" mode, must be non-negative. In "signed" mode, can contain both positive and negative values.
#' @param network edge list (data.frame) or sparse matrix.
#' @param gene_sets list of gene sets.
#' @param mode character, either "evidence" (default) or "signed". If "signed", the network propagation runs separately for positive and negative values.
#' @param p restart probability for RWR (default is 0.5).
#' @param specific_weight logical, whether to apply gene specificity weighting (TF-IDF style) based on gene frequencies in `gene_sets`. Default is FALSE.
#' @param minGSSize minimal size of each gene set for analyzing. default here is 10.
#' @param maxGSSize maximal size of genes annotated for testing. default here is 500.
#' @param threshold convergence threshold for RWR (default is 1e-9).
#' @param maxIter maximal number of RWR iterations (default is 100).
#' @param verbose logical, print messages.
#' @param ... other arguments passed to `gsea()`.
#'
#' @return A `nseaResult` object of NSEA results.
#' @export
nsea <- function(geneList,
                 network,
                 gene_sets,
                 mode = c("evidence", "signed"),
                 p = 0.5,
                 specific_weight = FALSE,
                 minGSSize = 10,
                 maxGSSize = 500,
                 threshold = 1e-9,
                 maxIter = 100,
                 verbose = TRUE,
                 ...) {
    
    mode <- match.arg(mode)
    if (!is.numeric(geneList) || is.null(names(geneList))) {
        stop("geneList must be a named numeric vector")
    }
    
    if (mode == "evidence" && any(geneList < 0)) {
        warning("geneList contains negative values but mode is 'evidence'. Negative values will be propagated as is, which might violate RWR assumptions. Consider using mode = 'signed'.")
    }
    
    if (verbose) message("Preparing network...")
    A <- prepare_network(network)
    nodes <- rownames(A)
    
    common_nodes <- intersect(names(geneList), nodes)
    if (length(common_nodes) == 0) {
        stop("No overlapping genes between geneList and network.")
    }
    
    if (mode == "evidence") {
        v <- rep(0, length(nodes))
        names(v) <- nodes
        v[common_nodes] <- geneList[common_nodes]
        
        sum_v <- sum(v)
        if (sum_v > 0) {
            v <- v / sum_v
        } else {
            stop("The sum of geneList scores in the network is zero.")
        }
        
        if (verbose) message("Running Random Walk with Restart (RWR)...")
        rwr_res <- rwr_eigen_cpp(A, v, restart = p, threshold = threshold, max_iter = maxIter)
        rwr_scores <- rwr_res$score
        names(rwr_scores) <- nodes
        iter <- as.integer(rwr_res$iterations)
        
    } else {
        # signed mode
        if (verbose) message("Running Signed RWR (Up and Down separately)...")
        
        v_up <- rep(0, length(nodes))
        names(v_up) <- nodes
        v_down <- rep(0, length(nodes))
        names(v_down) <- nodes
        
        genes_up <- common_nodes[geneList[common_nodes] > 0]
        genes_down <- common_nodes[geneList[common_nodes] < 0]
        
        v_up[genes_up] <- geneList[genes_up]
        v_down[genes_down] <- abs(geneList[genes_down])
        
        if (sum(v_up) > 0) v_up <- v_up / sum(v_up)
        if (sum(v_down) > 0) v_down <- v_down / sum(v_down)
        
        rwr_up <- rep(0, length(nodes))
        rwr_down <- rep(0, length(nodes))
        
        iter_up <- 0
        iter_down <- 0
        
        if (sum(v_up) > 0) {
            r_up <- rwr_eigen_cpp(A, v_up, restart = p, threshold = threshold, max_iter = maxIter)
            rwr_up <- r_up$score
            iter_up <- r_up$iterations
        }
        if (sum(v_down) > 0) {
            r_down <- rwr_eigen_cpp(A, v_down, restart = p, threshold = threshold, max_iter = maxIter)
            rwr_down <- r_down$score
            iter_down <- r_down$iterations
        }
        
        rwr_scores <- rwr_up - rwr_down
        names(rwr_scores) <- nodes
        iter <- as.integer(max(iter_up, iter_down))
    }
    
    if (specific_weight) {
        if (verbose) message("Applying gene specificity weighting (TF-IDF)...")
        N <- length(gene_sets)
        gene_freq <- table(unlist(gene_sets))
        df_vals <- as.numeric(gene_freq[names(rwr_scores)])
        df_vals[is.na(df_vals)] <- 1 # Minimum frequency 1
        w <- log(N / df_vals)
        w[w < 0] <- 0
        rwr_scores <- rwr_scores * w
    }
    
    rwr_scores <- sort(rwr_scores, decreasing = TRUE)
    
    if (verbose) message("Running GSEA...")
    if (mode == "evidence") {
        res <- gsea(geneList = rwr_scores,
                    gene_sets = gene_sets,
                    minGSSize = minGSSize,
                    maxGSSize = maxGSSize,
                    scoreType = "pos",
                    ...)
    } else {
        res <- gsea(geneList = rwr_scores,
                    gene_sets = gene_sets,
                    minGSSize = minGSSize,
                    maxGSSize = maxGSSize,
                    scoreType = "std",
                    ...)
    }
    
    build_nsea_result(
        base_result = res,
        gene_sets = gene_sets,
        rwr_scores = rwr_scores,
        network = network,
        mode = mode,
        iter = iter,
        p = p,
        organism = "UNKNOWN",
        setType = "UNKNOWN",
        keytype = "UNKNOWN",
        params = list(
            pvalueCutoff = 1.0,
            pAdjustMethod = "BH",
            minGSSize = minGSSize,
            maxGSSize = maxGSSize
        )
    )
}

#' Network-based GSEA using a GSON object
#'
#' @param geneList named numeric vector. In "evidence" mode, must be non-negative. In "signed" mode, can contain both positive and negative values.
#' @param network edge list (data.frame) or sparse matrix.
#' @param gson a GSON object.
#' @param mode character, either "evidence" (default) or "signed".
#' @param p restart probability for RWR (default is 0.5).
#' @param specific_weight logical, whether to apply gene specificity weighting (TF-IDF style) based on gene frequencies in the GSON object. Default is FALSE.
#' @param minGSSize minimal size of each gene set for analyzing. default here is 10.
#' @param maxGSSize maximal size of genes annotated for testing. default here is 500.
#' @param threshold convergence threshold for RWR (default is 1e-9).
#' @param maxIter maximal number of RWR iterations (default is 100).
#' @param verbose logical, print messages.
#' @param ... other arguments passed to `gsea_gson()`.
#'
#' @return A `nseaResult` object.
#' @export
nsea_gson <- function(geneList,
                      network,
                      gson,
                      mode = c("evidence", "signed"),
                      p = 0.5,
                      specific_weight = FALSE,
                      minGSSize = 10,
                      maxGSSize = 500,
                      threshold = 1e-9,
                      maxIter = 100,
                      verbose = TRUE,
                      ...) {
    
    mode <- match.arg(mode)
    if (!is.numeric(geneList) || is.null(names(geneList))) {
        stop("geneList must be a named numeric vector")
    }
    
    if (verbose) message("Preparing network...")
    A <- prepare_network(network)
    nodes <- rownames(A)
    
    common_nodes <- intersect(names(geneList), nodes)
    if (length(common_nodes) == 0) {
        stop("No overlapping genes between geneList and network.")
    }
    
    if (mode == "evidence") {
        v <- rep(0, length(nodes))
        names(v) <- nodes
        v[common_nodes] <- geneList[common_nodes]
        
        sum_v <- sum(v)
        if (sum_v > 0) {
            v <- v / sum_v
        } else {
            stop("The sum of geneList scores in the network is zero.")
        }
        
        if (verbose) message("Running Random Walk with Restart (RWR)...")
        rwr_res <- rwr_eigen_cpp(A, v, restart = p, threshold = threshold, max_iter = maxIter)
        rwr_scores <- rwr_res$score
        names(rwr_scores) <- nodes
        iter <- as.integer(rwr_res$iterations)
        
    } else {
        if (verbose) message("Running Signed RWR (Up and Down separately)...")
        
        v_up <- rep(0, length(nodes))
        names(v_up) <- nodes
        v_down <- rep(0, length(nodes))
        names(v_down) <- nodes
        
        genes_up <- common_nodes[geneList[common_nodes] > 0]
        genes_down <- common_nodes[geneList[common_nodes] < 0]
        
        v_up[genes_up] <- geneList[genes_up]
        v_down[genes_down] <- abs(geneList[genes_down])
        
        if (sum(v_up) > 0) v_up <- v_up / sum(v_up)
        if (sum(v_down) > 0) v_down <- v_down / sum(v_down)
        
        rwr_up <- rep(0, length(nodes))
        rwr_down <- rep(0, length(nodes))
        
        iter_up <- 0
        iter_down <- 0
        
        if (sum(v_up) > 0) {
            r_up <- rwr_eigen_cpp(A, v_up, restart = p, threshold = threshold, max_iter = maxIter)
            rwr_up <- r_up$score
            iter_up <- r_up$iterations
        }
        if (sum(v_down) > 0) {
            r_down <- rwr_eigen_cpp(A, v_down, restart = p, threshold = threshold, max_iter = maxIter)
            rwr_down <- r_down$score
            iter_down <- r_down$iterations
        }
        
        rwr_scores <- rwr_up - rwr_down
        names(rwr_scores) <- nodes
        iter <- as.integer(max(iter_up, iter_down))
    }
    
    if (specific_weight) {
        if (verbose) message("Applying gene specificity weighting (TF-IDF)...")
        gsid2gene <- gson@gsid2gene
        N <- length(unique(gsid2gene$gsid))
        gene_freq <- table(gsid2gene$gene)
        df_vals <- as.numeric(gene_freq[names(rwr_scores)])
        df_vals[is.na(df_vals)] <- 1 # Minimum frequency 1
        w <- log(N / df_vals)
        w[w < 0] <- 0
        rwr_scores <- rwr_scores * w
    }
    
    rwr_scores <- sort(rwr_scores, decreasing = TRUE)
    
    if (verbose) message("Running GSEA...")
    if (mode == "evidence") {
        res <- gsea_gson(geneList = rwr_scores,
                         gson = gson,
                         minGSSize = minGSSize,
                         maxGSSize = maxGSSize,
                         scoreType = "pos",
                         ...)
    } else {
        res <- gsea_gson(geneList = rwr_scores,
                         gson = gson,
                         minGSSize = minGSSize,
                         maxGSSize = maxGSSize,
                         scoreType = "std",
                         ...)
    }
    
    build_nsea_result(
        base_result = res,
        gene_sets = NULL,
        rwr_scores = rwr_scores,
        network = network,
        mode = mode,
        iter = iter,
        p = p
    )
}

test_that("nsea works", {
    set.seed(123)
    edges <- data.frame(
        from = sample(LETTERS[1:10], 20, replace = TRUE),
        to = sample(LETTERS[1:10], 20, replace = TRUE),
        weight = runif(20)
    )
    edges <- edges[edges$from != edges$to, ]
    
    geneList <- setNames(runif(5), sample(LETTERS[1:10], 5))
    geneList <- sort(geneList, decreasing = TRUE)
    
    gene_sets <- list(
        PathwayA = c("A", "B", "C", "D"),
        PathwayB = c("E", "F", "G", "H"),
        PathwayC = c("I", "J", "A")
    )
    
    res <- nsea(geneList = geneList,
                network = edges,
                gene_sets = gene_sets,
                p = 0.5,
                minGSSize = 2,
                maxGSSize = 10,
                nPermSimple = 1000,
                verbose = FALSE)
    
    expect_s4_class(res, "nseaResult")
    expect_true(nrow(res@result) > 0)
    expect_true("PathwayB" %in% res@result$ID)
    expect_identical(res@mode, "evidence")
    expect_identical(res@iterations, as.integer(res@iterations))
    
    # Test prepare_network manually
    A <- prepare_network(edges)
    expect_s4_class(A, "dgCMatrix")
})

test_that("nsea supports signed mode", {
    edges <- data.frame(
        from = c("A", "A", "B", "C", "D", "E", "F"),
        to = c("B", "C", "D", "D", "E", "F", "A"),
        weight = rep(1, 7),
        stringsAsFactors = FALSE
    )

    geneList <- c(A = 1.2, B = 0.8, D = -1.1, E = -0.7, F = 0.3)
    geneList <- sort(geneList, decreasing = TRUE)

    gene_sets <- list(
        UpPath = c("A", "B", "C"),
        DownPath = c("D", "E", "F")
    )

    res <- nsea(
        geneList = geneList,
        network = edges,
        gene_sets = gene_sets,
        mode = "signed",
        p = 0.5,
        minGSSize = 2,
        maxGSSize = 10,
        nPermSimple = 200,
        verbose = FALSE
    )

    expect_s4_class(res, "nseaResult")
    expect_identical(res@mode, "signed")
    expect_true(nrow(res@result) > 0)
    expect_equal(sort(names(res@diffusion_scores)), sort(unique(c(edges$from, edges$to))))
})

test_that("nsea_gson supports signed mode on a lightweight example", {
    skip_if_not_installed("gson")

    edges <- data.frame(
        from = c("A", "A", "B", "C", "D", "E", "F"),
        to = c("B", "C", "D", "D", "E", "F", "A"),
        weight = rep(1, 7),
        stringsAsFactors = FALSE
    )

    gsid2gene <- data.frame(
        gsid = c("UpPath", "UpPath", "DownPath", "DownPath"),
        gene = c("A", "B", "D", "E"),
        stringsAsFactors = FALSE
    )
    gsid2name <- data.frame(
        gsid = c("UpPath", "DownPath"),
        name = c("Up pathway", "Down pathway"),
        stringsAsFactors = FALSE
    )
    gson_obj <- gson::gson(
        gsid2gene = gsid2gene,
        gsid2name = gsid2name,
        species = "test",
        gsname = "test",
        version = "test",
        accessed_date = as.character(Sys.Date()),
        keytype = "SYMBOL"
    )

    geneList <- c(A = 1.2, B = 0.8, D = -1.1, E = -0.7, F = 0.3)
    geneList <- sort(geneList, decreasing = TRUE)

    res <- nsea_gson(
        geneList = geneList,
        network = edges,
        gson = gson_obj,
        mode = "signed",
        p = 0.5,
        minGSSize = 1,
        maxGSSize = 10,
        pvalueCutoff = 1,
        method = "sample",
        nPerm = 30,
        verbose = FALSE
    )

    expect_s4_class(res, "nseaResult")
    expect_identical(res@mode, "signed")
    expect_true(nrow(res@result) > 0)
    expect_equal(res@result$Description[match("UpPath", res@result$ID)], "Up pathway")
    expect_equal(sort(names(res@diffusion_scores)), sort(unique(c(edges$from, edges$to))))
})

#' Class "compareClusterResult"
#' This class represents the comparison result of gene clusters by GO
#' categories at specific level or GO enrichment analysis.
#'
#'
#' @name compareClusterResult-class
#' @aliases compareClusterResult-class show,compareClusterResult-method summary,compareClusterResult-method plot,compareClusterResult-method
#' @docType class
#' @slot compareClusterResult cluster comparing result
#' @slot geneClusters a list of genes
#' @slot fun one of groupGO, enrichGO and enrichKEGG
#' @slot gene2Symbol gene ID to Symbol
#' @slot keytype Gene ID type
#' @slot readable logical flag of gene ID in symbol or not.
#' @slot .call function call
#' @slot termsim Similarity between term
#' @slot method method of calculating the similarity between nodes
#' @slot dr dimension reduction result
#' @slot organism organism
#' @exportClass compareClusterResult
#' @author Guangchuang Yu \url{https://yulab-smu.top}
#' @exportClass compareClusterResult
#' @seealso 
#'   \code{\linkS4class{enrichResult}}
#' @keywords classes
setClass("compareClusterResult",
         representation = representation(
             compareClusterResult = "data.frame",
             geneClusters = "list",
             fun = "character",
             gene2Symbol    = "character",
             keytype        = "character",
             readable       = "logical",
             .call          = "call",
             termsim        = "matrix",
             method         = "character",
             dr             = "list",
             organism       = "character"
         )
         )

#' Class "enrichResult"
#' This class represents the result of enrichment analysis.
#'
#'
#' @name enrichResult-class
#' @aliases enrichResult-class show,enrichResult-method summary,enrichResult-method
#'
#' @docType class
#' @slot result enrichment analysis
#' @slot pvalueCutoff pvalueCutoff
#' @slot pAdjustMethod pvalue adjust method
#' @slot qvalueCutoff qvalueCutoff
#' @slot organism only "human" supported
#' @slot ontology biological ontology
#' @slot gene Gene IDs
#' @slot keytype Gene ID type
#' @slot universe background gene
#' @slot gene2Symbol mapping gene to Symbol
#' @slot geneSets gene sets
#' @slot readable logical flag of gene ID in symbol or not.
#' @slot termsim Similarity between term
#' @slot method method of calculating the similarity between nodes
#' @slot dr dimension reduction result
#' @exportClass enrichResult
#' @author Guangchuang Yu \url{https://yulab-smu.top}
#' @keywords classes
setClass("enrichResult",
         representation=representation(
             result         = "data.frame",
             pvalueCutoff   = "numeric",
             pAdjustMethod  = "character",
             qvalueCutoff   = "numeric",
             organism       = "character",
             ontology       = "character",
             gene           = "character",
             keytype        = "character",
             universe       = "character",
             gene2Symbol    = "character",
             geneSets       = "list",
             readable       = "logical",
             termsim        = "matrix",
             method         = "character",
             dr             = "list"
             ),
         prototype=prototype(readable = FALSE)
         )



#' Class "gseaResult"
#' This class represents the result of GSEA analysis
#'
#'
#' @name gseaResult-class
#' @aliases gseahResult-class show,gseaResult-method summary,gseaResult-method
#'
#' @docType class
#' @slot result GSEA anaysis
#' @slot organism organism
#' @slot setType setType
#' @slot geneSets geneSets
#' @slot geneList order rank geneList
#' @slot keytype ID type of gene
#' @slot permScores permutation scores
#' @slot params parameters
#' @slot gene2Symbol gene ID to Symbol
#' @slot readable whether convert gene ID to symbol
#' @slot dr dimension reduction result
#' @exportClass gseaResult
#' @author Guangchuang Yu \url{https://yulab-smu.top}
#' @keywords classes
setClass("gseaResult",
         representation   = representation(
             result          = "data.frame",
             organism        = "character",
             setType         = "character",
             geneSets        = "list",
             geneList        = "numeric",
             keytype         = "character",
             permScores      = "matrix",
             params          = "list",
             gene2Symbol     = "character",
             readable        = "logical",
             termsim         = "matrix",
             method          = "character",
             dr              = "list"
         )
         )


#' Class "nseaResult"
#' This class represents the result of Network-based Set Enrichment Analysis (NSEA).
#'
#' @name nseaResult-class
#' @aliases nseaResult-class
#' @docType class
#' @slot result enrichment analysis
#' @slot organism organism label for the enrichment result
#' @slot setType gene set collection type
#' @slot geneSets gene sets
#' @slot geneList order rank geneList
#' @slot keytype ID type of gene
#' @slot permScores permutation score matrix inherited from `gseaResult`
#' @slot gene2Symbol gene ID to symbol mapping
#' @slot readable logical flag of gene ID in symbol or not.
#' @slot termsim Calculation matrix of termsim.
#' @slot method Method of termsim.
#' @slot params parameters
#' @slot dr dimension reduction result
#' @slot network sparse matrix or data.frame representing the underlying network.
#' @slot diffusion_scores numeric vector of RWR diffusion scores for each node.
#' @slot mode character, "evidence" or "signed", describing the RWR propagation mode.
#' @slot iterations integer, the actual number of iterations RWR took to converge.
#' @slot restart_prob numeric, the restart probability used in RWR.
#' @exportClass nseaResult
#' @author Guangchuang Yu \url{https://yulab-smu.top}
setClass("nseaResult",
         contains = "gseaResult",
         representation = representation(
             network = "ANY",
             diffusion_scores = "numeric",
             mode = "character",
             iterations = "integer",
             restart_prob = "numeric"
         )
)

#' Class "mnseaResult"
#' This class represents the result of multi-layer Network-based Set Enrichment Analysis.
#'
#' @name mnseaResult-class
#' @aliases mnseaResult-class
#' @docType class
#' @slot result enrichment analysis
#' @slot organism organism label for the enrichment result
#' @slot setType gene set collection type
#' @slot geneSets gene sets
#' @slot geneList order rank geneList
#' @slot keytype ID type of gene
#' @slot permScores permutation score matrix inherited from `gseaResult`
#' @slot gene2Symbol gene ID to symbol mapping
#' @slot readable logical flag of gene ID in symbol or not.
#' @slot termsim Calculation matrix of termsim.
#' @slot method Method of termsim.
#' @slot params parameters
#' @slot dr dimension reduction result
#' @slot multilayer_network prepared multi-layer network object.
#' @slot layer_scores list of layer-specific diffusion score vectors.
#' @slot collapsed_scores numeric vector used for downstream enrichment.
#' @slot layer_weights numeric vector of layer weights.
#' @slot coupling_table data.frame of inter-layer couplings.
#' @slot mode character, "evidence" or "signed".
#' @slot iterations integer, the actual number of iterations RWR took to converge.
#' @slot restart_prob numeric, the restart probability used in RWR.
#' @slot collapse_method character collapse method used on layer scores.
#' @slot target_layer optional layer name used for downstream export.
#' @slot output_space character output space of collapsed scores.
#' @slot pathway_contribution pathway-by-layer contribution table precomputed for explanation.
#' @slot feature_contribution feature-by-layer contribution table precomputed for explanation.
#' @exportClass mnseaResult
#' @author Guangchuang Yu \url{https://yulab-smu.top}
setClass("mnseaResult",
         contains = "gseaResult",
         representation = representation(
             multilayer_network = "ANY",
             layer_scores = "list",
             collapsed_scores = "numeric",
             layer_weights = "numeric",
             coupling_table = "data.frame",
             mode = "character",
             iterations = "integer",
             restart_prob = "numeric",
             collapse_method = "character",
             target_layer = "character",
             output_space = "character",
             pathway_contribution = "data.frame",
             feature_contribution = "data.frame"
         )
)

#' Extract pathway subnetwork data from a `mnseaResult`
#'
#' @param res A `mnseaResult` object.
#' @param pathway_id Optional pathway ID. If `NULL`, the top pathway is used.
#' @param include_couplings Logical, whether to include inter-layer coupling
#'   edges. Default is `TRUE`.
#' @param include_isolated Logical, whether to keep nodes without retained
#'   edges. Default is `TRUE`.
#'
#' @return A list with `pathway`, `layer_contribution`, `nodes`, and `edges`.
#' @export
extract_mnsea_subnetwork <- function(res,
                                     pathway_id = NULL,
                                     include_couplings = TRUE,
                                     include_isolated = TRUE) {
    if (!inherits(res, "mnseaResult")) {
        stop("res must be a mnseaResult object.")
    }
    if (nrow(res@result) == 0) {
        stop("The mnseaResult is empty.")
    }
    
    if (is.null(pathway_id)) {
        pathway_id <- as.character(res@result$ID[1])
        message("pathway_id not provided. Using the top pathway: ", pathway_id)
    }
    if (!pathway_id %in% res@result$ID) {
        stop("pathway_id not found in the mnseaResult.")
    }
    
    pathway_row <- res@result[res@result$ID == pathway_id, , drop = FALSE]
    layer_contribution <- res@pathway_contribution[
        res@pathway_contribution$ID == pathway_id,
        ,
        drop = FALSE
    ]
    nodes <- res@feature_contribution[
        res@feature_contribution$ID == pathway_id,
        ,
        drop = FALSE
    ]
    
    if (nrow(nodes) == 0) {
        return(list(
            pathway = pathway_row,
            layer_contribution = layer_contribution,
            nodes = nodes,
            edges = data.frame(
                from = character(0),
                to = character(0),
                from_layer = character(0),
                to_layer = character(0),
                from_feature = character(0),
                to_feature = character(0),
                weight = numeric(0),
                edge_type = character(0),
                stringsAsFactors = FALSE
            )
        ))
    }
    
    nodes$node_key <- paste(nodes$layer, nodes$Feature, sep = "::")
    nodes$collapsed_score <- unname(res@collapsed_scores[nodes$Feature])
    layer_weight_lookup <- res@layer_weights
    nodes$layer_weight <- unname(layer_weight_lookup[nodes$layer])
    
    edge_tables <- list()
    edge_idx <- 1L
    layer_names <- unique(nodes$layer)
    
    for (layer in layer_names) {
        layer_nodes <- unique(nodes$Feature[nodes$layer == layer])
        A <- res@multilayer_network$intra_matrices[[layer]]
        valid_nodes <- intersect(layer_nodes, rownames(A))
        if (length(valid_nodes) == 0) {
            next
        }
        
        idx <- match(valid_nodes, rownames(A))
        idx <- idx[!is.na(idx)]
        if (length(idx) == 0) {
            next
        }
        subA <- A[idx, idx, drop = FALSE]
        sm <- Matrix::summary(subA)
        if (nrow(sm) == 0) {
            next
        }
        
        edge_tables[[edge_idx]] <- data.frame(
            from = paste(layer, rownames(A)[idx][sm$i], sep = "::"),
            to = paste(layer, colnames(A)[idx][sm$j], sep = "::"),
            from_layer = layer,
            to_layer = layer,
            from_feature = rownames(A)[idx][sm$i],
            to_feature = colnames(A)[idx][sm$j],
            weight = sm$x,
            edge_type = "intra",
            stringsAsFactors = FALSE
        )
        edge_idx <- edge_idx + 1L
    }
    
    if (include_couplings && nrow(res@coupling_table) > 0) {
        coupling_edges <- res@coupling_table
        node_keys <- unique(nodes$node_key)
        keep <- paste(coupling_edges$from_layer, coupling_edges$from_id, sep = "::") %in% node_keys &
            paste(coupling_edges$to_layer, coupling_edges$to_id, sep = "::") %in% node_keys
        coupling_edges <- coupling_edges[keep, , drop = FALSE]
        
        if (nrow(coupling_edges) > 0) {
            edge_tables[[edge_idx]] <- data.frame(
                from = paste(coupling_edges$from_layer, coupling_edges$from_id, sep = "::"),
                to = paste(coupling_edges$to_layer, coupling_edges$to_id, sep = "::"),
                from_layer = coupling_edges$from_layer,
                to_layer = coupling_edges$to_layer,
                from_feature = coupling_edges$from_id,
                to_feature = coupling_edges$to_id,
                weight = coupling_edges$weight,
                edge_type = "coupling",
                stringsAsFactors = FALSE
            )
        }
    }
    
    edges <- do.call(rbind, edge_tables)
    if (is.null(edges)) {
        edges <- data.frame(
            from = character(0),
            to = character(0),
            from_layer = character(0),
            to_layer = character(0),
            from_feature = character(0),
            to_feature = character(0),
            weight = numeric(0),
            edge_type = character(0),
            stringsAsFactors = FALSE
        )
    }
    
    if (!include_isolated && nrow(edges) > 0) {
        keep_nodes <- unique(c(edges$from, edges$to))
        nodes <- nodes[nodes$node_key %in% keep_nodes, , drop = FALSE]
    }
    
    list(
        pathway = pathway_row,
        layer_contribution = layer_contribution,
        nodes = nodes,
        edges = edges
    )
}

#' @importFrom yulab.utils yulab_msg
check_gene_id <- function(gene, gsid2gene) {
    if (!any(gene %in% gsid2gene$gene)) {
        yulab_msg("--> No gene can be mapped....")
        sg <- unique(gsid2gene$gene[1:100])
        sg <- sample(sg, min(length(sg), 6))
        yulab_msg("--> Expected input gene ID: ", paste0(sg, collapse=','))
        yulab_msg("--> return NULL...")
        return(FALSE)
    }
    return(TRUE)
}

calculate_qvalue <- function(pvals) {
    if (length(pvals) == 0)
        return(numeric(0))

    # Check if qvalue package is available
    if (!requireNamespace("qvalue", quietly = TRUE)) {
        warning("qvalue package not installed. qvalue column will be NA.")
        return(rep(NA_real_, length(pvals)))
    }

    qv <- rep(NA_real_, length(pvals))
    ok <- is.finite(pvals) & !is.na(pvals) & pvals >= 0 & pvals <= 1
    if (any(!ok)) {
        warning("Invalid p-values detected (NA, non-finite, <0, or >1). qvalue will be computed on valid p-values only.")
    }

    p <- pvals[ok]
    if (length(p) == 0) {
        return(qv)
    }

    if (all(p == 1)) {
        qv[ok] <- 1
        return(qv)
    }

    if (all(p == 0)) {
        qv[ok] <- 0
        return(qv)
    }

    qobj <- NULL
    last_err <- NULL
    try_qvalue <- function(expr) {
        tryCatch(expr, error = function(e) {
            last_err <<- e
            NULL
        })
    }

    qobj <- try_qvalue(qvalue::qvalue(p))
    if (!inherits(qobj, "qvalue")) {
        qobj <- try_qvalue(qvalue::qvalue(p, lambda = 0.05, pi0.method = "bootstrap"))
    }
    if (!inherits(qobj, "qvalue")) {
        qobj <- try_qvalue(qvalue::qvalue(p, lambda = seq(0, 0.9, 0.05), pi0.method = "bootstrap"))
    }
    if (!inherits(qobj, "qvalue")) {
        qobj <- try_qvalue(qvalue::qvalue(p, lambda = seq(0, 0.95, 0.05), pi0.method = "bootstrap"))
    }

    if (inherits(qobj, "qvalue")) {
        qv[ok] <- qobj$qvalues
        return(qv)
    }

    if (inherits(last_err, "error")) {
        warning("qvalue::qvalue() failed, returning NA for qvalue. Error: ", conditionMessage(last_err))
    } else {
        warning("qvalue::qvalue() failed, returning NA for qvalue. This may be due to p-value distribution.")
    }
    return(qv)
}

get_geneSet_index <- function(geneSets, minGSSize, maxGSSize) {
    if (is.na(minGSSize) || is.null(minGSSize))
        minGSSize <- 1
    if (is.na(maxGSSize) || is.null(maxGSSize))
        maxGSSize <- Inf 

    ## index of geneSets in used.
    ## logical
    geneSet_size <- sapply(geneSets, length)
    idx <-  minGSSize <= geneSet_size & geneSet_size <= maxGSSize
    return(idx)
}

validate_gene_sets <- function(gene_sets) {
    if (!is.list(gene_sets) || is.null(names(gene_sets))) {
        stop("gene_sets must be a named list")
    }
    gene_sets <- lapply(gene_sets, function(x) {
        if (!is.character(x)) {
            stop("Each element in gene_sets must be a character vector")
        }
        unique(x)
    })
    return(gene_sets)
}

gsea_leading_edge_details <- function(geneList, geneSet, exponent, scoreType) {
    genes <- names(geneList)
    N <- length(geneList)
    geneSet <- unique(intersect(as.character(geneSet), genes))
    if (length(geneSet) == 0 || N == 0) {
        return(list(rank = 0L, leading_edge = "tags=0%, list=0%, signal=0%", core_enrichment = ""))
    }

    in_set <- !is.na(match(genes, geneSet))
    N_H <- sum(in_set)
    if (N_H == 0) {
        return(list(rank = 0L, leading_edge = "tags=0%, list=0%, signal=0%", core_enrichment = ""))
    }

    weights <- abs(geneList)
    if (!isTRUE(all.equal(exponent, 1.0))) {
        weights <- weights^exponent
    }
    N_R <- sum(weights[in_set])
    N_miss <- N - N_H

    hit_inc <- if (N_R == 0) rep(0, N) else (weights * in_set) / N_R
    miss_inc <- if (N_miss == 0) rep(0, N) else (!in_set) / N_miss
    running <- cumsum(hit_inc) - cumsum(miss_inc)

    if (scoreType == "pos") {
        peak_idx <- which.max(running)
        es <- running[peak_idx]
    } else if (scoreType == "neg") {
        peak_idx <- which.min(running)
        es <- running[peak_idx]
    } else {
        max_es <- max(running)
        min_es <- min(running)
        if (abs(max_es) >= abs(min_es)) {
            peak_idx <- which.max(running)
            es <- max_es
        } else {
            peak_idx <- which.min(running)
            es <- min_es
        }
    }

    if (is.na(peak_idx) || length(peak_idx) == 0) {
        return(list(rank = 0L, leading_edge = "tags=0%, list=0%, signal=0%", core_enrichment = ""))
    }
    peak_idx <- peak_idx[[1]]

    if (es >= 0) {
        rank <- as.integer(peak_idx)
        ledge_range <- seq_len(peak_idx)
        list_frac <- peak_idx / N
    } else {
        rank <- as.integer(N - peak_idx + 1)
        ledge_range <- peak_idx:N
        list_frac <- (N - peak_idx + 1) / N
    }

    core_genes <- genes[ledge_range][in_set[ledge_range]]
    hits_in_ledge <- length(core_genes)

    tags_frac <- hits_in_ledge / N_H
    signal_frac <- tags_frac * (1 - list_frac) * (N / max(N_miss, 1))
    leading_edge <- paste0(
        "tags=", round(tags_frac * 100), "%",
        ", list=", round(list_frac * 100), "%",
        ", signal=", round(signal_frac * 100), "%"
    )

    list(
        rank = rank,
        leading_edge = leading_edge,
        core_enrichment = paste0(core_genes, collapse = "/")
    )
}

#' Common parameters for enrichit functions
#'
#' @param geneList A named numeric vector of gene statistics (e.g., log fold change), ranked in descending order.
#' @param gene_sets A named list of gene sets. Each element is a character vector of genes.
#' @param nPerm Number of permutations for p-value calculation (default: 1000).
#' @param exponent Weighting exponent for enrichment score (default: 1.0).
#' @param minGSSize minimal size of each geneSet for analyzing
#' @param maxGSSize maximal size of each geneSet for analyzing
#' @param pvalueCutoff P-value cutoff.
#' @param pAdjustMethod P-value adjustment method (e.g., "BH").
#' @param verbose Logical. Print progress messages.
#' @param gson A GSON object containing gene set information.
#' @param method Permutation method.
#' @param adaptive Logical. Use adaptive permutation.
#' @param minPerm Minimum permutations for adaptive mode.
#' @param maxPerm Maximum permutations for adaptive mode.
#' @param pvalThreshold P-value threshold for early stopping.
#' @keywords internal
#' @name enrichit_params
NULL

get_enriched <- function(object) {

    Over <- object@result

    pvalueCutoff <- object@pvalueCutoff
    if (length(pvalueCutoff) != 0) {
        Over <- Over[ Over$pvalue <= pvalueCutoff, ]
        Over <- Over[ Over$p.adjust <= pvalueCutoff, ]
    }

    qvalueCutoff <- object@qvalueCutoff
    if (length(qvalueCutoff) != 0) {
        if (! any(is.na(Over$qvalue))) {
            if (length(qvalueCutoff) > 0)
                Over <- Over[ Over$qvalue <= qvalueCutoff, ]
        }
    }

    object@result <- Over
    return(object)
}

TERM2NAME <- function(term, gson) {
    if (inherits(gson, "environment")) { 
        PATHID2NAME <- get("PATHID2NAME", envir = gson)
        #if (is.null(PATHID2NAME) || is.na(PATHID2NAME)) {
        if (is.null(PATHID2NAME) || all(is.na(PATHID2NAME))) {
            return(as.character(term))
        }
        res <- PATHID2NAME[term]
        i <-  is.na(res)
        res[i] <- term[i]
    } else if (inherits(gson, "GSON")) {
        gsid2name <- gson@gsid2name
        res <- as.character(term)
        if (!is.null(gsid2name) && nrow(gsid2name) > 0L) {
            i <- match(as.character(term), as.character(gsid2name$gsid))
            j <- !is.na(i)
            res[j] <- as.character(gsid2name$name[i[j]])
        }
    } else {
        res <- as.character(term)
    }

    names(res) <- term
    return(res) 
}

TERMID2EXTID <- function(term, gson) {
    if (inherits(gson, "GSON")) {
        gsid2gene <- gson@gsid2gene
        res <- split(
            as.character(gsid2gene$gene),
            as.character(gsid2gene$gsid)
        )
        res <- res[as.character(term)]
        names(res) <- as.character(term)
        i <- vapply(res, is.null, logical(1))
        res[i] <- list(character(0))
        return(res)
    } else if (inherits(gson, "environment")) {
        PATHID2EXTID <- get("PATHID2EXTID", envir = gson)
        res <- PATHID2EXTID[term]
        return(res)
    } else {
        stop("gson not supported")
    }
}

library(testthat)
library(enrichit)

test_that("calculate_qvalue handles invalid p-values without poisoning all results", {
  skip_if_not_installed("qvalue")

  pvals <- c(0.01, 0.5, NA_real_, -0.1, 1.1)
  qv <- expect_warning(enrichit:::calculate_qvalue(pvals), "Invalid p-values detected")

  expect_length(qv, length(pvals))
  expect_true(is.double(qv))

  expect_false(is.na(qv[1]))
  expect_false(is.na(qv[2]))
  expect_true(qv[1] >= 0 && qv[1] <= 1)
  expect_true(qv[2] >= 0 && qv[2] <= 1)

  expect_true(is.na(qv[3]))
  expect_true(is.na(qv[4]))
  expect_true(is.na(qv[5]))
})

library(testthat)
library(enrichit)

test_that("TERM2NAME maps GSON term ids to names and falls back to ids", {
  x <- gson::gson(
    gsid2gene = data.frame(
      gsid = c("T1", "T1", "T2", "T3"),
      gene = c("g1", "g2", "g3", "g4")
    ),
    gsid2name = data.frame(
      gsid = c("T1", "T2"),
      name = c("Term 1", "Term 2")
    )
  )

  res <- enrichit:::TERM2NAME(c("T1", "T3", "T9"), x)

  expect_type(res, "character")
  expect_named(res, c("T1", "T3", "T9"))
  expect_identical(unname(res), c("Term 1", "T3", "T9"))
})

test_that("TERMID2EXTID returns genes grouped by requested GSON term ids", {
  x <- gson::gson(
    gsid2gene = data.frame(
      gsid = c("T1", "T1", "T2", "T3"),
      gene = c("g1", "g2", "g3", "g4")
    ),
    gsid2name = data.frame(
      gsid = c("T1", "T2"),
      name = c("Term 1", "Term 2")
    )
  )

  res <- enrichit:::TERMID2EXTID(c("T2", "T1", "T9"), x)

  expect_type(res, "list")
  expect_named(res, c("T2", "T1", "T9"))
  expect_true(all(vapply(res, is.character, logical(1))))
  expect_identical(res$T2, "g3")
  expect_identical(res$T1, c("g1", "g2"))
  expect_identical(res$T9, character(0))
})

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

ora_cpp <- function(gene, universe, gene_sets, gene_set_names) {
    .Call(`_enrichit_ora_cpp`, gene, universe, gene_sets, gene_set_names)
}

gsea_cpp <- function(stats, gene_sets, gene_set_names, nPerm = 1000L, exponent = 1.0, method = "sample", seed = 12345L) {
    .Call(`_enrichit_gsea_cpp`, stats, gene_sets, gene_set_names, nPerm, exponent, method, seed)
}

gsea_adaptive_cpp <- function(stats, gene_sets, gene_set_names, minPerm = 1000L, maxPerm = 100000L, pvalThreshold = 0.1, exponent = 1.0, method = "sample", seed = 12345L) {
    .Call(`_enrichit_gsea_adaptive_cpp`, stats, gene_sets, gene_set_names, minPerm, maxPerm, pvalThreshold, exponent, method, seed)
}

gsea_multilevel_cpp <- function(geneList, gene_sets, gene_set_names, minPerm, maxPerm, pvalThreshold, exponent, method, eps, sampleSize, seed, nPermSimple = 1000L, scoreType = "std") {
    .Call(`_enrichit_gsea_multilevel_cpp`, geneList, gene_sets, gene_set_names, minPerm, maxPerm, pvalThreshold, exponent, method, eps, sampleSize, seed, nPermSimple, scoreType)
}

gsea_scores_cpp <- function(stats, in_set, exponent) {
    .Call(`_enrichit_gsea_scores_cpp`, stats, in_set, exponent)
}

#' Random Walk with Restart using Eigen Sparse Matrix
#'
#' @param A column-normalized sparse matrix
#' @param v initial restart vector
#' @param restart restart probability (e.g., 0.5)
#' @param threshold convergence threshold
#' @param max_iter maximal number of iterations
#' @return list containing stationary probabilities and iterations
#' @noRd
rwr_eigen_cpp <- function(A, v, restart, threshold, max_iter) {
    .Call(`_enrichit_rwr_eigen_cpp`, A, v, restart, threshold, max_iter)
}

#' Prepare multi-layer network for repeated propagation
#'
#' @param networks named list of layer-specific networks.
#' @param couplings data.frame of inter-layer edges with columns
#'   `from_layer`, `from_id`, `to_layer`, `to_id`, and optional `weight`.
#' @param directed logical, whether the multi-layer graph is directed.
#' @param intra_normalize one of "column", "row", or "none".
#' @param inter_normalize one of "column", "row", or "none".
#' @param interlayer_strength numeric scalar used to scale all coupling edges.
#' @param layer_order explicit layer order. Defaults to `names(networks)`.
#'
#' @return A `multilayer_network` object.
#' @export
prepare_multilayer_network <- function(networks,
                                       couplings,
                                       directed = FALSE,
                                       intra_normalize = "column",
                                       inter_normalize = "column",
                                       interlayer_strength = 1,
                                       layer_order = names(networks)) {
    intra_normalize <- match.arg(intra_normalize, c("column", "row", "none"))
    inter_normalize <- match.arg(inter_normalize, c("column", "row", "none"))
    
    if (!is.list(networks) || length(networks) == 0) {
        stop("networks must be a non-empty named list.")
    }
    if (is.null(names(networks)) || any(!nzchar(names(networks)))) {
        stop("networks must be a named list with non-empty layer names.")
    }
    if (is.null(layer_order) || !setequal(layer_order, names(networks))) {
        stop("layer_order must contain exactly the same layer names as networks.")
    }
    if (!is.numeric(interlayer_strength) || length(interlayer_strength) != 1 || is.na(interlayer_strength) || interlayer_strength < 0) {
        stop("interlayer_strength must be a single non-negative number.")
    }
    
    layer_order <- as.character(layer_order)
    prepared <- vector("list", length(layer_order))
    names(prepared) <- layer_order
    node_index <- vector("list", length(layer_order))
    offsets <- integer(length(layer_order))
    total_nodes <- 0L
    
    intra_i <- integer(0)
    intra_j <- integer(0)
    intra_x <- numeric(0)
    
    for (k in seq_along(layer_order)) {
        layer <- layer_order[[k]]
        A <- prepare_network(networks[[layer]], directed = directed, normalize = intra_normalize)
        prepared[[layer]] <- A
        nodes <- rownames(A)
        n_nodes <- length(nodes)
        offsets[[k]] <- total_nodes
        
        node_index[[k]] <- data.frame(
            layer = layer,
            node = nodes,
            key = paste(layer, nodes, sep = "::"),
            index = seq_len(n_nodes) + total_nodes,
            stringsAsFactors = FALSE
        )
        
        sm <- Matrix::summary(A)
        if (nrow(sm) > 0) {
            intra_i <- c(intra_i, sm$i + total_nodes)
            intra_j <- c(intra_j, sm$j + total_nodes)
            intra_x <- c(intra_x, sm$x)
        }
        
        total_nodes <- total_nodes + n_nodes
    }
    
    node_index <- do.call(rbind, node_index)
    rownames(node_index) <- NULL
    
    key_to_index <- node_index$index
    names(key_to_index) <- node_index$key
    
    if (missing(couplings) || is.null(couplings)) {
        couplings <- data.frame(
            from_layer = character(0),
            from_id = character(0),
            to_layer = character(0),
            to_id = character(0),
            weight = numeric(0),
            stringsAsFactors = FALSE
        )
    } else {
        couplings <- as.data.frame(couplings, stringsAsFactors = FALSE)
        required_cols <- c("from_layer", "from_id", "to_layer", "to_id")
        if (!all(required_cols %in% colnames(couplings))) {
            stop("couplings must contain columns 'from_layer', 'from_id', 'to_layer', and 'to_id'.")
        }
        if (!"weight" %in% colnames(couplings)) {
            couplings$weight <- 1
        }
        couplings$from_layer <- as.character(couplings$from_layer)
        couplings$from_id <- as.character(couplings$from_id)
        couplings$to_layer <- as.character(couplings$to_layer)
        couplings$to_id <- as.character(couplings$to_id)
        couplings$weight <- as.numeric(couplings$weight)
    }
    
    coupling_matrix <- Matrix::sparseMatrix(
        i = integer(0),
        j = integer(0),
        x = numeric(0),
        dims = c(total_nodes, total_nodes),
        dimnames = list(node_index$key, node_index$key)
    )
    
    if (nrow(couplings) > 0) {
        from_key <- paste(couplings$from_layer, couplings$from_id, sep = "::")
        to_key <- paste(couplings$to_layer, couplings$to_id, sep = "::")
        
        if (any(!from_key %in% names(key_to_index))) {
            bad <- unique(from_key[!from_key %in% names(key_to_index)])
            stop("Unknown coupling source nodes: ", paste(bad, collapse = ", "))
        }
        if (any(!to_key %in% names(key_to_index))) {
            bad <- unique(to_key[!to_key %in% names(key_to_index)])
            stop("Unknown coupling target nodes: ", paste(bad, collapse = ", "))
        }
        
        ci <- unname(key_to_index[from_key])
        cj <- unname(key_to_index[to_key])
        cx <- couplings$weight
        
        if (!directed) {
            ci_orig <- ci
            cj_orig <- cj
            ci <- c(ci_orig, cj_orig)
            cj <- c(cj_orig, ci_orig)
            cx <- c(cx, cx)
        }
        
        coupling_matrix <- Matrix::sparseMatrix(
            i = ci,
            j = cj,
            x = cx,
            dims = c(total_nodes, total_nodes),
            dimnames = list(node_index$key, node_index$key)
        )
        
        if (inter_normalize == "column") {
            cs <- Matrix::colSums(coupling_matrix)
            cs[cs == 0] <- 1
            coupling_matrix <- coupling_matrix %*% Matrix::Diagonal(x = 1 / cs)
        } else if (inter_normalize == "row") {
            rs <- Matrix::rowSums(coupling_matrix)
            rs[rs == 0] <- 1
            coupling_matrix <- Matrix::Diagonal(x = 1 / rs) %*% coupling_matrix
        }
    }
    
    adjacency <- Matrix::sparseMatrix(
        i = intra_i,
        j = intra_j,
        x = intra_x,
        dims = c(total_nodes, total_nodes),
        dimnames = list(node_index$key, node_index$key)
    )
    adjacency <- adjacency + interlayer_strength * coupling_matrix
    
    structure(
        list(
            adjacency = adjacency,
            intra_matrices = prepared,
            coupling_matrix = coupling_matrix,
            couplings = couplings,
            node_index = node_index,
            layer_order = layer_order,
            directed = directed,
            intra_normalize = intra_normalize,
            inter_normalize = inter_normalize,
            interlayer_strength = interlayer_strength
        ),
        class = "multilayer_network"
    )
}

#' Propagate signals on a multi-layer network
#'
#' @param seed_list named list of named numeric vectors, one per layer.
#' @param network a prepared `multilayer_network` object.
#' @param mode one of "evidence" or "signed".
#' @param p restart probability.
#' @param threshold convergence threshold.
#' @param maxIter maximum number of iterations.
#' @param layer_weights optional named numeric vector of layer weights.
#' @param target_layer optional layer name to focus on downstream.
#'
#' @return A `multilayer_propagation` object.
#' @export
propagate_multilayer <- function(seed_list,
                                 network,
                                 mode = c("evidence", "signed"),
                                 p = 0.5,
                                 threshold = 1e-9,
                                 maxIter = 100,
                                 layer_weights = NULL,
                                 target_layer = NULL) {
    mode <- match.arg(mode)
    
    if (!inherits(network, "multilayer_network")) {
        stop("network must be a multilayer_network object.")
    }
    if (!is.list(seed_list) || length(seed_list) == 0) {
        stop("seed_list must be a non-empty named list.")
    }
    if (is.null(names(seed_list)) || any(!nzchar(names(seed_list)))) {
        stop("seed_list must be a named list with non-empty layer names.")
    }
    
    layer_order <- network$layer_order
    if (!all(names(seed_list) %in% layer_order)) {
        bad <- setdiff(names(seed_list), layer_order)
        stop("Unknown layers in seed_list: ", paste(bad, collapse = ", "))
    }
    
    if (is.null(layer_weights)) {
        layer_weights <- rep(1, length(layer_order))
        names(layer_weights) <- layer_order
    } else {
        if (is.null(names(layer_weights))) {
            if (length(layer_weights) != length(layer_order)) {
                stop("Unnamed layer_weights must match the number of layers.")
            }
            names(layer_weights) <- layer_order
        }
        layer_weights <- layer_weights[layer_order]
        layer_weights[is.na(layer_weights)] <- 1
    }
    
    if (!is.null(target_layer) && !target_layer %in% layer_order) {
        stop("target_layer must be one of the layers in the network.")
    }
    
    node_index <- network$node_index
    node_lookup <- split(node_index, node_index$layer)
    seed_vec <- numeric(nrow(node_index))
    names(seed_vec) <- node_index$key
    
    for (layer in names(seed_list)) {
        s <- seed_list[[layer]]
        if (!is.numeric(s) || is.null(names(s))) {
            stop("Each element in seed_list must be a named numeric vector.")
        }
        idx_df <- node_lookup[[layer]]
        common_nodes <- intersect(names(s), idx_df$node)
        if (length(common_nodes) == 0) {
            next
        }
        match_idx <- idx_df$index[match(common_nodes, idx_df$node)]
        seed_vec[match_idx] <- as.numeric(s[common_nodes]) * layer_weights[[layer]]
    }
    
    if (mode == "evidence" && any(seed_vec < 0)) {
        warning("seed_list contains negative values but mode is 'evidence'. Consider using mode = 'signed'.")
    }
    
    propagation <- .run_multilayer_rwr(
        A = network$adjacency,
        seed = seed_vec,
        mode = mode,
        p = p,
        threshold = threshold,
        maxIter = maxIter
    )
    
    global_scores <- propagation$score
    names(global_scores) <- node_index$key
    layer_scores <- lapply(layer_order, function(layer) {
        idx_df <- node_lookup[[layer]]
        scores <- global_scores[idx_df$key]
        names(scores) <- idx_df$node
        scores
    })
    names(layer_scores) <- layer_order
    
    structure(
        list(
            scores = global_scores,
            layer_scores = layer_scores,
            mode = mode,
            iterations = as.integer(propagation$iterations),
            restart_prob = p,
            layer_weights = layer_weights,
            target_layer = target_layer,
            network = network
        ),
        class = "multilayer_propagation"
    )
}

#' Collapse multi-layer diffusion scores
#'
#' @param x result from `propagate_multilayer()`.
#' @param collapse one of "weighted_mean", "sum", "mean", or "max_abs".
#' @param layer_weights optional named numeric vector used when
#'   `collapse = "weighted_mean"`.
#' @param output_space one of "union" or "gene".
#' @param mapping optional mapping data.frame with `source_id`, `target_id`, and
#'   optional `layer` columns.
#' @param target_layer optional layer name to extract before collapsing.
#'
#' @return A `multilayer_collapsed` object with a `score` vector.
#' @export
collapse_multilayer_scores <- function(x,
                                       collapse = c("weighted_mean", "sum", "mean", "max_abs"),
                                       layer_weights = NULL,
                                       output_space = c("union", "gene"),
                                       mapping = NULL,
                                       target_layer = NULL) {
    collapse <- match.arg(collapse)
    output_space <- match.arg(output_space)
    
    if (!inherits(x, "multilayer_propagation")) {
        stop("x must be a multilayer_propagation object.")
    }
    
    layer_scores <- x$layer_scores
    if (!is.null(target_layer)) {
        if (!target_layer %in% names(layer_scores)) {
            stop("target_layer is not available in the propagation result.")
        }
        layer_scores <- layer_scores[target_layer]
    }
    
    long_df <- do.call(rbind, lapply(names(layer_scores), function(layer) {
        scores <- layer_scores[[layer]]
        data.frame(
            layer = layer,
            source_id = names(scores),
            score = as.numeric(scores),
            stringsAsFactors = FALSE
        )
    }))
    
    if (nrow(long_df) == 0) {
        stop("No layer scores available to collapse.")
    }
    
    if (is.null(mapping)) {
        long_df$target_id <- long_df$source_id
    } else {
        mapping <- as.data.frame(mapping, stringsAsFactors = FALSE)
        if (!all(c("source_id", "target_id") %in% colnames(mapping))) {
            stop("mapping must contain 'source_id' and 'target_id' columns.")
        }
        
        if ("layer" %in% colnames(mapping)) {
            merged <- merge(long_df, mapping, by = c("layer", "source_id"), all.x = TRUE, sort = FALSE)
        } else {
            merged <- merge(long_df, mapping, by = "source_id", all.x = TRUE, sort = FALSE)
        }
        merged$target_id[is.na(merged$target_id)] <- merged$source_id[is.na(merged$target_id)]
        long_df <- merged
    }
    
    feature_ids <- unique(long_df$target_id)
    layer_names <- unique(long_df$layer)
    score_mat <- matrix(
        NA_real_,
        nrow = length(feature_ids),
        ncol = length(layer_names),
        dimnames = list(feature_ids, layer_names)
    )
    
    for (feature in feature_ids) {
        for (layer in layer_names) {
            vals <- long_df$score[long_df$target_id == feature & long_df$layer == layer]
            if (length(vals) == 0) {
                next
            }
            score_mat[feature, layer] <- .collapse_vector(vals, method = collapse)
        }
    }
    
    if (is.null(layer_weights)) {
        layer_weights <- x$layer_weights[layer_names]
    } else {
        if (is.null(names(layer_weights))) {
            if (length(layer_weights) != length(layer_names)) {
                stop("Unnamed layer_weights must match the number of layers used for collapsing.")
            }
            names(layer_weights) <- layer_names
        }
        layer_weights <- layer_weights[layer_names]
    }
    layer_weights[is.na(layer_weights)] <- 1
    
    collapsed <- apply(score_mat, 1, function(s) {
        .collapse_vector(s, method = collapse, weights = layer_weights)
    })
    collapsed <- collapsed[!is.na(collapsed)]
    
    structure(
        list(
            score = collapsed,
            score_matrix = score_mat[rownames(score_mat) %in% names(collapsed), , drop = FALSE],
            feature_id = names(collapsed),
            output_space = output_space,
            collapse_method = collapse,
            layer_weights = layer_weights,
            target_layer = target_layer
        ),
        class = "multilayer_collapsed"
    )
}

#' Multi-layer Network-based Gene Set Enrichment Analysis
#'
#' @param seed_list named list of named numeric vectors, one per layer.
#' @param networks named list of layer-specific networks.
#' @param couplings data.frame of inter-layer edges.
#' @param gene_sets list of gene sets.
#' @param mode one of "evidence" or "signed".
#' @param layer_weights optional named numeric vector.
#' @param collapse one of "weighted_mean", "sum", "mean", or "max_abs".
#' @param target_layer optional layer name to export scores from.
#' @param output_space one of "union" or "gene".
#' @param p restart probability.
#' @param interlayer_strength global scaling factor for coupling edges.
#' @param specific_weight logical.
#' @param minGSSize minimal size of each gene set.
#' @param maxGSSize maximal size of genes annotated for testing.
#' @param threshold convergence threshold.
#' @param maxIter maximal number of iterations.
#' @param verbose logical.
#' @param ... additional arguments passed to `gsea()`.
#'
#' @return A `mnseaResult` object.
#' @export
mnsea <- function(seed_list,
                  networks,
                  couplings,
                  gene_sets,
                  mode = c("evidence", "signed"),
                  layer_weights = NULL,
                  collapse = c("weighted_mean", "sum", "mean", "max_abs"),
                  target_layer = NULL,
                  output_space = c("union", "gene"),
                  p = 0.5,
                  interlayer_strength = 1,
                  specific_weight = FALSE,
                  minGSSize = 10,
                  maxGSSize = 500,
                  threshold = 1e-9,
                  maxIter = 100,
                  verbose = TRUE,
                  ...) {
    mode <- match.arg(mode)
    collapse <- match.arg(collapse)
    output_space <- match.arg(output_space)
    
    if (verbose) message("Preparing multi-layer network...")
    ml_net <- prepare_multilayer_network(
        networks = networks,
        couplings = couplings,
        interlayer_strength = interlayer_strength
    )
    
    if (verbose) message("Running multi-layer propagation...")
    prop <- propagate_multilayer(
        seed_list = seed_list,
        network = ml_net,
        mode = mode,
        p = p,
        threshold = threshold,
        maxIter = maxIter,
        layer_weights = layer_weights,
        target_layer = target_layer
    )
    
    collapsed <- collapse_multilayer_scores(
        x = prop,
        collapse = collapse,
        layer_weights = layer_weights,
        output_space = output_space,
        target_layer = target_layer
    )
    
    collapsed_scores <- collapsed$score
    if (length(collapsed_scores) == 0) {
        stop("No collapsed scores available for downstream enrichment.")
    }
    if (specific_weight) {
        if (verbose) message("Applying gene specificity weighting (TF-IDF)...")
        N <- length(gene_sets)
        gene_freq <- table(unlist(gene_sets))
        df_vals <- as.numeric(gene_freq[names(collapsed_scores)])
        df_vals[is.na(df_vals)] <- 1
        w <- log(N / df_vals)
        w[w < 0] <- 0
        collapsed_scores <- collapsed_scores * w
    }
    collapsed_scores <- sort(collapsed_scores, decreasing = TRUE)
    
    if (verbose) message("Running GSEA...")
    if (mode == "evidence") {
        res <- gsea(
            geneList = collapsed_scores,
            gene_sets = gene_sets,
            minGSSize = minGSSize,
            maxGSSize = maxGSSize,
            scoreType = "pos",
            ...
        )
    } else {
        res <- gsea(
            geneList = collapsed_scores,
            gene_sets = gene_sets,
            minGSSize = minGSSize,
            maxGSSize = maxGSSize,
            scoreType = "std",
            ...
        )
    }
    if (is.null(res)) {
        return(NULL)
    }
    
    build_mnsea_result(
        base_result = res,
        gene_sets = gene_sets,
        propagation = prop,
        collapsed = collapsed_scores,
        collapse_method = collapse,
        output_space = output_space,
        p = p,
        params = list(
            pvalueCutoff = 1.0,
            pAdjustMethod = "BH",
            minGSSize = minGSSize,
            maxGSSize = maxGSSize
        )
    )
}

#' Multi-layer NSEA using a GSON object
#'
#' @param seed_list named list of named numeric vectors, one per layer.
#' @param networks named list of layer-specific networks.
#' @param couplings data.frame of inter-layer edges.
#' @param gson a GSON object.
#' @param mode one of "evidence" or "signed".
#' @param layer_weights optional named numeric vector.
#' @param collapse one of "weighted_mean", "sum", "mean", or "max_abs".
#' @param target_layer optional layer name to export scores from.
#' @param output_space one of "union" or "gene".
#' @param p restart probability.
#' @param interlayer_strength global scaling factor for coupling edges.
#' @param specific_weight logical.
#' @param minGSSize minimal size of each gene set.
#' @param maxGSSize maximal size of genes annotated for testing.
#' @param threshold convergence threshold.
#' @param maxIter maximal number of iterations.
#' @param verbose logical.
#' @param ... additional arguments passed to `gsea_gson()`.
#'
#' @return A `mnseaResult` object.
#' @export
mnsea_gson <- function(seed_list,
                       networks,
                       couplings,
                       gson,
                       mode = c("evidence", "signed"),
                       layer_weights = NULL,
                       collapse = c("weighted_mean", "sum", "mean", "max_abs"),
                       target_layer = NULL,
                       output_space = c("union", "gene"),
                       p = 0.5,
                       interlayer_strength = 1,
                       specific_weight = FALSE,
                       minGSSize = 10,
                       maxGSSize = 500,
                       threshold = 1e-9,
                       maxIter = 100,
                       verbose = TRUE,
                       ...) {
    mode <- match.arg(mode)
    collapse <- match.arg(collapse)
    output_space <- match.arg(output_space)
    
    if (verbose) message("Preparing multi-layer network...")
    ml_net <- prepare_multilayer_network(
        networks = networks,
        couplings = couplings,
        interlayer_strength = interlayer_strength
    )
    
    if (verbose) message("Running multi-layer propagation...")
    prop <- propagate_multilayer(
        seed_list = seed_list,
        network = ml_net,
        mode = mode,
        p = p,
        threshold = threshold,
        maxIter = maxIter,
        layer_weights = layer_weights,
        target_layer = target_layer
    )
    
    collapsed <- collapse_multilayer_scores(
        x = prop,
        collapse = collapse,
        layer_weights = layer_weights,
        output_space = output_space,
        target_layer = target_layer
    )
    
    collapsed_scores <- collapsed$score
    if (length(collapsed_scores) == 0) {
        stop("No collapsed scores available for downstream enrichment.")
    }
    if (specific_weight) {
        if (verbose) message("Applying gene specificity weighting (TF-IDF)...")
        gsid2gene <- gson@gsid2gene
        N <- length(unique(gsid2gene$gsid))
        gene_freq <- table(gsid2gene$gene)
        df_vals <- as.numeric(gene_freq[names(collapsed_scores)])
        df_vals[is.na(df_vals)] <- 1
        w <- log(N / df_vals)
        w[w < 0] <- 0
        collapsed_scores <- collapsed_scores * w
    }
    collapsed_scores <- sort(collapsed_scores, decreasing = TRUE)
    
    if (verbose) message("Running GSEA...")
    if (mode == "evidence") {
        res <- gsea_gson(
            geneList = collapsed_scores,
            gson = gson,
            minGSSize = minGSSize,
            maxGSSize = maxGSSize,
            scoreType = "pos",
            ...
        )
    } else {
        res <- gsea_gson(
            geneList = collapsed_scores,
            gson = gson,
            minGSSize = minGSSize,
            maxGSSize = maxGSSize,
            scoreType = "std",
            ...
        )
    }
    if (is.null(res)) {
        return(NULL)
    }
    
    build_mnsea_result(
        base_result = res,
        gene_sets = NULL,
        propagation = prop,
        collapsed = collapsed_scores,
        collapse_method = collapse,
        output_space = output_space,
        p = p,
        params = NULL
    )
}

build_mnsea_result <- function(base_result,
                               gene_sets,
                               propagation,
                               collapsed,
                               collapse_method,
                               output_space,
                               p,
                               params = NULL) {
    if (inherits(base_result, "gseaResult")) {
        result_df <- base_result@result
        gene_sets <- base_result@geneSets
        organism <- base_result@organism
        setType <- base_result@setType
        keytype <- base_result@keytype
        params <- base_result@params
        gene_list <- base_result@geneList
        perm_scores <- base_result@permScores
        gene2symbol <- base_result@gene2Symbol
        readable <- base_result@readable
        termsim <- base_result@termsim
        method <- base_result@method
        dr <- base_result@dr
    } else if (is.data.frame(base_result)) {
        result_df <- base_result
        if (is.null(params)) {
            params <- list(pvalueCutoff = 1.0, pAdjustMethod = "BH")
        }
        gene_list <- collapsed
        perm_scores <- matrix(0, nrow = 0, ncol = 0)
        gene2symbol <- character(0)
        readable <- FALSE
        termsim <- matrix(0, nrow = 0, ncol = 0)
        method <- "MNSEA"
        dr <- list()
        organism <- "UNKNOWN"
        setType <- "UNKNOWN"
        keytype <- "UNKNOWN"
    } else {
        stop("base_result must be a gseaResult or data.frame.")
    }
    
    if (!"Description" %in% colnames(result_df)) {
        result_df$Description <- result_df$ID
    }
    if (!"p.adjust" %in% colnames(result_df)) {
        result_df$p.adjust <- stats::p.adjust(result_df$pvalue, method = params$pAdjustMethod %||% "BH")
    }
    if (!"qvalue" %in% colnames(result_df)) {
        result_df$qvalue <- calculate_qvalue(result_df$pvalue)
    }
    rownames(result_df) <- as.character(result_df$ID)
    
    explain_cache <- .build_mnsea_explain_cache(
        result_df = result_df,
        gene_sets = gene_sets,
        layer_scores = propagation$layer_scores
    )
    
    new("mnseaResult",
        result = result_df,
        organism = organism,
        setType = setType,
        geneSets = gene_sets,
        geneList = gene_list,
        keytype = keytype,
        permScores = perm_scores,
        params = params,
        gene2Symbol = gene2symbol,
        readable = readable,
        termsim = termsim,
        method = method,
        dr = dr,
        multilayer_network = propagation$network,
        layer_scores = propagation$layer_scores,
        collapsed_scores = collapsed,
        layer_weights = propagation$layer_weights,
        coupling_table = propagation$network$couplings,
        mode = propagation$mode,
        iterations = as.integer(propagation$iterations),
        restart_prob = p,
        collapse_method = collapse_method,
        target_layer = propagation$target_layer %||% "",
        output_space = output_space,
        pathway_contribution = explain_cache$pathway_contribution,
        feature_contribution = explain_cache$feature_contribution)
}

.run_multilayer_rwr <- function(A, seed, mode, p, threshold, maxIter) {
    if (mode == "evidence") {
        sum_seed <- sum(seed)
        if (sum_seed <= 0) {
            stop("The sum of seed scores in the multi-layer network is zero.")
        }
        v <- seed / sum_seed
        rwr_res <- rwr_eigen_cpp(A, v, restart = p, threshold = threshold, max_iter = maxIter)
        return(list(score = rwr_res$score, iterations = rwr_res$iterations))
    }
    
    v_up <- ifelse(seed > 0, seed, 0)
    v_down <- ifelse(seed < 0, abs(seed), 0)
    if (sum(v_up) > 0) v_up <- v_up / sum(v_up)
    if (sum(v_down) > 0) v_down <- v_down / sum(v_down)
    
    rwr_up <- rep(0, length(seed))
    rwr_down <- rep(0, length(seed))
    iter_up <- 0
    iter_down <- 0
    
    if (sum(v_up) > 0) {
        up_res <- rwr_eigen_cpp(A, v_up, restart = p, threshold = threshold, max_iter = maxIter)
        rwr_up <- up_res$score
        iter_up <- up_res$iterations
    }
    if (sum(v_down) > 0) {
        down_res <- rwr_eigen_cpp(A, v_down, restart = p, threshold = threshold, max_iter = maxIter)
        rwr_down <- down_res$score
        iter_down <- down_res$iterations
    }
    
    list(score = rwr_up - rwr_down, iterations = max(iter_up, iter_down))
}

.collapse_vector <- function(x, method, weights = NULL) {
    x <- x[!is.na(x)]
    if (length(x) == 0) {
        return(NA_real_)
    }
    if (method == "sum") {
        return(sum(x))
    }
    if (method == "mean") {
        return(mean(x))
    }
    if (method == "max_abs") {
        return(x[which.max(abs(x))])
    }
    if (method == "weighted_mean") {
        if (is.null(weights)) {
            return(mean(x))
        }
        w <- weights[names(weights) %in% names(x)]
        if (length(w) == 0) {
            return(mean(x))
        }
        return(sum(x * w) / sum(w))
    }
    stop("Unsupported collapse method: ", method)
}

.build_mnsea_explain_cache <- function(result_df, gene_sets, layer_scores) {
    layer_names <- names(layer_scores)
    if (length(layer_names) == 0 || nrow(result_df) == 0) {
        empty_pathway <- data.frame(
            ID = character(0),
            Description = character(0),
            layer = character(0),
            contribution = numeric(0),
            share = numeric(0),
            n_feature = integer(0),
            stringsAsFactors = FALSE
        )
        empty_feature <- data.frame(
            ID = character(0),
            Description = character(0),
            Feature = character(0),
            layer = character(0),
            score = numeric(0),
            abs_score = numeric(0),
            is_core = logical(0),
            stringsAsFactors = FALSE
        )
        return(list(
            pathway_contribution = empty_pathway,
            feature_contribution = empty_feature
        ))
    }
    
    layer_feature_union <- unique(unlist(lapply(layer_scores, names), use.names = FALSE))
    pathway_rows <- vector("list", nrow(result_df))
    feature_rows <- vector("list", nrow(result_df))
    
    for (i in seq_len(nrow(result_df))) {
        path_id <- as.character(result_df$ID[i])
        path_desc <- as.character(result_df$Description[i])
        gene_col <- if ("core_enrichment" %in% colnames(result_df) &&
            !is.na(result_df$core_enrichment[i]) &&
            nzchar(as.character(result_df$core_enrichment[i]))) {
            "core_enrichment"
        } else if ("geneID" %in% colnames(result_df) &&
            !is.na(result_df$geneID[i]) &&
            nzchar(as.character(result_df$geneID[i]))) {
            "geneID"
        } else {
            NA_character_
        }
        
        if (!is.na(gene_col)) {
            selected_features <- unique(unlist(strsplit(as.character(result_df[[gene_col]][i]), "/")))
        } else if (!is.null(gene_sets) && path_id %in% names(gene_sets)) {
            selected_features <- unique(as.character(gene_sets[[path_id]]))
        } else {
            selected_features <- character(0)
        }
        
        selected_features <- intersect(selected_features, layer_feature_union)
        if (length(selected_features) == 0) {
            next
        }
        
        if ("core_enrichment" %in% colnames(result_df) &&
            !is.na(result_df$core_enrichment[i]) &&
            nzchar(as.character(result_df$core_enrichment[i]))) {
            core_features <- unique(unlist(strsplit(as.character(result_df$core_enrichment[i]), "/")))
        } else {
            core_features <- selected_features
        }
        
        layer_contrib <- numeric(length(layer_names))
        names(layer_contrib) <- layer_names
        per_layer_rows <- vector("list", length(layer_names))
        
        for (j in seq_along(layer_names)) {
            layer <- layer_names[[j]]
            scores <- layer_scores[[layer]]
            valid_features <- intersect(selected_features, names(scores))
            if (length(valid_features) == 0) {
                per_layer_rows[[j]] <- NULL
                layer_contrib[[layer]] <- 0
                next
            }
            
            layer_vals <- scores[valid_features]
            layer_contrib[[layer]] <- mean(abs(layer_vals))
            per_layer_rows[[j]] <- data.frame(
                ID = path_id,
                Description = path_desc,
                Feature = valid_features,
                layer = layer,
                score = as.numeric(layer_vals),
                abs_score = abs(as.numeric(layer_vals)),
                is_core = valid_features %in% core_features,
                stringsAsFactors = FALSE
            )
        }
        
        share <- if (sum(layer_contrib) > 0) layer_contrib / sum(layer_contrib) else rep(0, length(layer_contrib))
        pathway_rows[[i]] <- data.frame(
            ID = path_id,
            Description = path_desc,
            layer = layer_names,
            contribution = as.numeric(layer_contrib),
            share = as.numeric(share),
            n_feature = length(selected_features),
            stringsAsFactors = FALSE
        )
        feature_rows[[i]] <- do.call(rbind, per_layer_rows)
    }
    
    pathway_contribution <- do.call(rbind, pathway_rows)
    feature_contribution <- do.call(rbind, feature_rows)
    
    if (is.null(pathway_contribution)) {
        pathway_contribution <- data.frame(
            ID = character(0),
            Description = character(0),
            layer = character(0),
            contribution = numeric(0),
            share = numeric(0),
            n_feature = integer(0),
            stringsAsFactors = FALSE
        )
    }
    if (is.null(feature_contribution)) {
        feature_contribution <- data.frame(
            ID = character(0),
            Description = character(0),
            Feature = character(0),
            layer = character(0),
            score = numeric(0),
            abs_score = numeric(0),
            is_core = logical(0),
            stringsAsFactors = FALSE
        )
    }
    
    list(
        pathway_contribution = pathway_contribution,
        feature_contribution = feature_contribution
    )
}

library(testthat)
library(enrichit)

test_that("prepare_multilayer_network builds a valid supra-graph", {
  networks <- list(
    rna = data.frame(
      from = c("A", "B"),
      to = c("B", "C"),
      weight = c(1, 1),
      stringsAsFactors = FALSE
    ),
    protein = data.frame(
      from = c("A", "C"),
      to = c("C", "D"),
      weight = c(1, 1),
      stringsAsFactors = FALSE
    )
  )
  couplings <- data.frame(
    from_layer = c("rna", "rna"),
    from_id = c("A", "C"),
    to_layer = c("protein", "protein"),
    to_id = c("A", "C"),
    weight = c(1, 2),
    stringsAsFactors = FALSE
  )

  ml_net <- prepare_multilayer_network(networks, couplings)

  expect_s3_class(ml_net, "multilayer_network")
  expect_equal(nrow(ml_net$node_index), 6)
  expect_s4_class(ml_net$adjacency, "dgCMatrix")
  expect_equal(nrow(ml_net$couplings), 2)
})

test_that("propagate_multilayer and collapse_multilayer_scores stay lightweight", {
  networks <- list(
    rna = data.frame(
      from = c("A", "B", "C"),
      to = c("B", "C", "D"),
      weight = c(1, 1, 1),
      stringsAsFactors = FALSE
    ),
    protein = data.frame(
      from = c("A", "C", "D"),
      to = c("C", "D", "A"),
      weight = c(1, 1, 1),
      stringsAsFactors = FALSE
    )
  )
  couplings <- data.frame(
    from_layer = c("rna", "rna", "rna"),
    from_id = c("A", "C", "D"),
    to_layer = c("protein", "protein", "protein"),
    to_id = c("A", "C", "D"),
    weight = c(1, 1, 1),
    stringsAsFactors = FALSE
  )
  seed_list <- list(
    rna = c(A = 1, D = -0.5),
    protein = c(A = 0.5, C = 0.8)
  )

  ml_net <- prepare_multilayer_network(networks, couplings)
  prop <- propagate_multilayer(
    seed_list = seed_list,
    network = ml_net,
    mode = "signed",
    maxIter = 50
  )
  collapsed <- collapse_multilayer_scores(
    x = prop,
    collapse = "weighted_mean",
    layer_weights = c(rna = 1, protein = 2)
  )

  expect_s3_class(prop, "multilayer_propagation")
  expect_equal(sort(names(prop$layer_scores)), c("protein", "rna"))
  expect_s3_class(collapsed, "multilayer_collapsed")
  expect_true(all(c("A", "C", "D") %in% names(collapsed$score)))
})

test_that("mnsea works on a small multi-layer example", {
  networks <- list(
    rna = data.frame(
      from = c("A", "B", "C"),
      to = c("B", "C", "D"),
      weight = c(1, 1, 1),
      stringsAsFactors = FALSE
    ),
    protein = data.frame(
      from = c("A", "C", "D"),
      to = c("C", "D", "A"),
      weight = c(1, 1, 1),
      stringsAsFactors = FALSE
    )
  )
  couplings <- data.frame(
    from_layer = c("rna", "rna", "rna"),
    from_id = c("A", "C", "D"),
    to_layer = c("protein", "protein", "protein"),
    to_id = c("A", "C", "D"),
    weight = c(1, 1, 1),
    stringsAsFactors = FALSE
  )
  seed_list <- list(
    rna = c(A = 1.2, B = 0.8, D = -0.9),
    protein = c(A = 0.4, C = 1.1, D = -0.3)
  )
  gene_sets <- list(
    Pathway1 = c("A", "B", "C"),
    Pathway2 = c("D")
  )

  res <- mnsea(
    seed_list = seed_list,
    networks = networks,
    couplings = couplings,
    gene_sets = gene_sets,
    mode = "signed",
    collapse = "weighted_mean",
    layer_weights = c(rna = 1, protein = 1.5),
    minGSSize = 1,
    maxGSSize = 10,
    method = "sample",
    nPerm = 30,
    verbose = FALSE
  )

  expect_s4_class(res, "mnseaResult")
  expect_identical(res@mode, "signed")
  expect_true(length(res@layer_scores) == 2)
  expect_true(length(res@collapsed_scores) > 0)
  expect_true(nrow(res@result) > 0)
  expect_true(nrow(res@pathway_contribution) > 0)
  expect_true(nrow(res@feature_contribution) > 0)
  expect_true(all(c("ID", "layer", "contribution", "share") %in% colnames(res@pathway_contribution)))
  expect_true(all(c("ID", "Feature", "layer", "score", "abs_score", "is_core") %in% colnames(res@feature_contribution)))

  path_tbl <- get_mnsea_contribution(res, level = "pathway")
  feat_tbl <- get_mnsea_contribution(res, pathway_id = path_tbl$ID[1], level = "feature")
  expect_true(nrow(path_tbl) > 0)
  expect_true(nrow(feat_tbl) > 0)

  subnet <- extract_mnsea_subnetwork(res, pathway_id = path_tbl$ID[1])
  expect_true(all(c("pathway", "layer_contribution", "nodes", "edges") %in% names(subnet)))
  expect_true(nrow(subnet$nodes) > 0)
  expect_true(all(c("node_key", "collapsed_score", "layer_weight") %in% colnames(subnet$nodes)))
  expect_true(all(c("from", "to", "weight", "edge_type") %in% colnames(subnet$edges)))
  expect_true(any(subnet$edges$edge_type %in% c("intra", "coupling")))
})

test_that("mnsea_gson works on a small multi-layer example", {
  skip_if_not_installed("gson")

  gsid2gene <- data.frame(
    gsid = c("Path1", "Path1", "Path2"),
    gene = c("A", "B", "D"),
    stringsAsFactors = FALSE
  )
  gsid2name <- data.frame(
    gsid = c("Path1", "Path2"),
    name = c("Pathway 1", "Pathway 2"),
    stringsAsFactors = FALSE
  )
  gson_obj <- gson::gson(
    gsid2gene = gsid2gene,
    gsid2name = gsid2name,
    species = "test",
    gsname = "test",
    version = "test",
    accessed_date = as.character(Sys.Date()),
    keytype = "SYMBOL"
  )

  networks <- list(
    rna = data.frame(from = c("A", "B"), to = c("B", "D"), weight = c(1, 1), stringsAsFactors = FALSE),
    protein = data.frame(from = c("A", "D"), to = c("D", "A"), weight = c(1, 1), stringsAsFactors = FALSE)
  )
  couplings <- data.frame(
    from_layer = c("rna", "rna"),
    from_id = c("A", "D"),
    to_layer = c("protein", "protein"),
    to_id = c("A", "D"),
    weight = c(1, 1),
    stringsAsFactors = FALSE
  )
  seed_list <- list(
    rna = c(A = 1, B = 0.5, D = -0.3),
    protein = c(A = 0.2, D = -0.5)
  )

  res <- mnsea_gson(
    seed_list = seed_list,
    networks = networks,
    couplings = couplings,
    gson = gson_obj,
    mode = "signed",
    collapse = "mean",
    minGSSize = 1,
    maxGSSize = 10,
    pvalueCutoff = 1,
    method = "sample",
    nPerm = 30,
    verbose = FALSE
  )

  expect_s4_class(res, "mnseaResult")
  expect_true(nrow(res@result) > 0)
})

#' Get gene-level omics contribution for a specific pathway
#'
#' Extract the original multi-omics statistics for genes in a specific enriched pathway.
#'
#' @param res An \code{enrichResult} or \code{gseaResult} object.
#' @param agg An \code{omics_aggregated} object from \code{aggregate_omics()}.
#' @param pathway_id Character, the ID of the pathway to extract. If NULL, the top pathway is used.
#'
#' @return A data.frame containing the genes, their original omics statistics, the aggregated score, and whether they belong to the core enrichment.
#' @export
get_omics_contribution <- function(res, agg, pathway_id = NULL) {
    if (!inherits(agg, "omics_aggregated") || is.null(agg$original_matrix)) {
        stop("agg must be an omics_aggregated object containing original_matrix.")
    }
    
    res_df <- as.data.frame(res)
    if (nrow(res_df) == 0) {
        stop("The enrichment result is empty.")
    }
    
    if (is.null(pathway_id)) {
        pathway_id <- res_df$ID[1]
        message("pathway_id not provided. Using the top pathway: ", pathway_id)
    }
    
    if (!pathway_id %in% res_df$ID) {
        stop("pathway_id not found in the enrichment result.")
    }
    
    pathway_row <- res_df[res_df$ID == pathway_id, , drop = FALSE]
    
    # Extract genes in the pathway
    gene_col <- if ("core_enrichment" %in% colnames(pathway_row)) "core_enrichment" else "geneID"
    if (!gene_col %in% colnames(pathway_row)) {
        stop("Cannot find 'core_enrichment' or 'geneID' column in the result.")
    }
    
    pathway_genes <- unlist(strsplit(as.character(pathway_row[[gene_col]]), "/"))
    
    # Get the matrix and score
    mat <- agg$original_matrix
    all_genes <- rownames(mat)
    
    valid_genes <- intersect(pathway_genes, all_genes)
    if (length(valid_genes) == 0) {
        warning("No matching genes found between the pathway and the aggregated object.")
        return(NULL)
    }
    
    df <- as.data.frame(mat[valid_genes, , drop = FALSE])
    df$Aggregated_Score <- agg$score[valid_genes]
    
    if (!is.null(agg$pvalue)) {
        df$Aggregated_Pvalue <- agg$pvalue[valid_genes]
    }
    
    df$Feature <- valid_genes
    rownames(df) <- NULL
    
    # Reorder columns
    cols <- c("Feature", colnames(mat), "Aggregated_Score")
    if (!is.null(agg$pvalue)) cols <- c(cols, "Aggregated_Pvalue")
    
    return(df[, cols, drop = FALSE])
}

#' Classify pathway-level multi-omics patterns
#'
#' Compare merged enrichment results with single-omics enrichment results to classify the contribution pattern of each pathway.
#'
#' @param merged_res An \code{enrichResult} or \code{gseaResult} object from the merged multi-omics analysis.
#' @param single_res A named list of \code{enrichResult} or \code{gseaResult} objects from single-omics analyses.
#' @param p_cutoff Numeric, the significance cutoff. Default is 0.05.
#' @param by Character, the column to use for significance threshold. Default is "p.adjust".
#'
#' @return The \code{merged_res} object with an additional column \code{Omics_Pattern} in its result data.frame.
#' @export
classify_omics_pattern <- function(merged_res, single_res, p_cutoff = 0.05, by = "p.adjust") {
    if (!is.list(single_res) || is.null(names(single_res))) {
        stop("single_res must be a named list of enrichment results.")
    }
    
    merged_df <- as.data.frame(merged_res)
    if (!by %in% colnames(merged_df)) {
        stop(sprintf("Column '%s' not found in merged_res.", by))
    }
    
    # Prepare a lookup table for single-omics results
    single_lookups <- lapply(single_res, function(res) {
        df <- as.data.frame(res)
        if (nrow(df) > 0 && by %in% colnames(df)) {
            stats::setNames(df[[by]], df$ID)
        } else {
            stats::setNames(numeric(0), character(0))
        }
    })
    
    omics_names <- names(single_res)
    
    patterns <- sapply(seq_len(nrow(merged_df)), function(i) {
        path_id <- merged_df$ID[i]
        merged_p <- merged_df[[by]][i]
        
        if (is.na(merged_p) || merged_p > p_cutoff) {
            return("Not Significant")
        }
        
        sig_count <- 0
        sig_omics <- character(0)
        
        for (om in omics_names) {
            p_val <- single_lookups[[om]][path_id]
            if (!is.na(p_val) && p_val <= p_cutoff) {
                sig_count <- sig_count + 1
                sig_omics <- c(sig_omics, om)
            }
        }
        
        if (sig_count == 0) {
            return("Enhanced (1+1>2)")
        } else if (sig_count == 1) {
            return(paste0(sig_omics[1], "-Specific"))
        } else {
            return("Shared")
        }
    })
    
    if (inherits(merged_res, "enrichResult") || inherits(merged_res, "gseaResult")) {
        merged_res@result$Omics_Pattern <- patterns
    } else if (is.data.frame(merged_res)) {
        merged_res$Omics_Pattern <- patterns
    }
    return(merged_res)
}

#' interal method for enrichment analysis
#'
#' using the hypergeometric model
#' @title ora-gson
#' @param gene a vector of entrez gene id.
#' @param qvalueCutoff cutoff of qvalue
#' @param universe background genes, default is the intersection of the 'universe' with genes that have annotations. 
#' Users can set `options(enrichment_force_universe = TRUE)` to force the 'universe' untouched.
#' @param weight A named numeric vector of weights for background genes. If provided, Weighted ORA will be performed.
#' @inheritParams enrichit_params
#' @return  A `enrichResult` instance.
#' @importClassesFrom methods data.frame
#' @importFrom methods new
#' @importFrom stats p.adjust
#' @keywords manip
#' @author Guangchuang Yu <https://yulab-smu.top>
#' @export
ora_gson <- function(gene,
                              pvalueCutoff,
                              pAdjustMethod="BH",
                              universe = NULL,
                              weight = NULL,
                              minGSSize=10,
                              maxGSSize=500,
                              qvalueCutoff=0.2,
                              gson){

    if (!inherits(gson, "GSON")) {
        stop("gson should be a GSON object")
    }

    ## query external ID to Term ID
    gene <- as.character(unique(gene))
    
    # Extract gene sets from GSON
    gsid2gene <- gson@gsid2gene
    
    # ID Match Check
    if (!check_gene_id(gene, gsid2gene)) {
        return(NULL)
    }

    # Handle universe
    extID <- unique(gsid2gene$gene)
    if (missing(universe))
        universe <- NULL
    if(!is.null(universe)) {
        if (is.character(universe)) {
            force_universe <- getOption("enrichment_force_universe", FALSE)
            if (force_universe) {
                extID <- universe
            } else {
                extID <- intersect(extID, universe)
            }
        } else {
            ## https://github.com/YuLab-SMU/clusterProfiler/issues/217
            message("`universe` is not in character and will be ignored...")
        }
    }

    # Prepare Gene Sets
    geneSets <- split(gsid2gene$gene, gsid2gene$gsid)
    # Intersect with universe
    geneSets <- lapply(geneSets, intersect, extID)
    
    # Filter by size
    idx <- get_geneSet_index(geneSets, minGSSize, maxGSSize)
    if (sum(idx) == 0) {
        msg <- paste("No gene sets have size between", minGSSize, "and", maxGSSize, "...")
        message(msg)
        message("--> return NULL...")
        return (NULL)
    }
    geneSets <- geneSets[idx]

    ora_res <- ora(gene, geneSets, universe = extID, weight = weight)

    if (is.null(ora_res) || nrow(ora_res) == 0) {
        return(NULL)
    }

    # Rename columns to match expectation
    if ("PValue" %in% names(ora_res)) {
        names(ora_res)[names(ora_res) == "PValue"] <- "pvalue"
    }
    if ("GeneSet" %in% names(ora_res)) {
        names(ora_res)[names(ora_res) == "GeneSet"] <- "ID"
    }
    if ("DEInSet" %in% names(ora_res)) {
        names(ora_res)[names(ora_res) == "DEInSet"] <- "Count"
    }

    # Only gene sets with at least one matched input gene should contribute to
    # multiple testing correction. Keeping zero-overlap rows here inflates the
    # number of hypotheses and shifts adjusted statistics relative to the
    # historical DOSE/clusterProfiler behavior.
    if ("Count" %in% names(ora_res)) {
        ora_res <- ora_res[ora_res$Count > 0, , drop = FALSE]
    }

    if (nrow(ora_res) == 0) {
        return(NULL)
    }
    
    # Calculate ratios
    if (all(c("Count", "DESize") %in% names(ora_res))) {
        ora_res$GeneRatio <- paste0(ora_res$Count, "/", ora_res$DESize)
    }
    if (all(c("SetSize", "UniverseSize") %in% names(ora_res))) {
        ora_res$BgRatio <- paste0(ora_res$SetSize, "/", ora_res$UniverseSize)
    }

    # Calculate RichFactor and FoldEnrichment
    if (all(c("Count", "SetSize", "DESize", "UniverseSize") %in% names(ora_res))) {
        ora_res$RichFactor <- ora_res$Count / ora_res$SetSize
        ora_res$FoldEnrichment <- (ora_res$Count / ora_res$DESize) / (ora_res$SetSize / ora_res$UniverseSize)
    }

    p <- suppressWarnings(as.numeric(ora_res$pvalue))
    invalid <- is.na(p) | !is.finite(p) | p < 0 | p > 1
    if (any(invalid)) {
        ids <- as.character(ora_res$ID)
        bad_i <- which(invalid)
        head_i <- bad_i[seq_len(min(length(bad_i), 10))]
        bad_preview <- paste0(ids[head_i], "=", format(p[head_i], digits = 4, scientific = TRUE), collapse = ", ")
        warning(
            "Invalid p-values detected in ORA result (",
            length(bad_i),
            "/",
            length(p),
            "). Examples: ",
            bad_preview
        )
        too_small <- !is.na(p) & is.finite(p) & p < 0
        too_large <- !is.na(p) & is.finite(p) & p > 1
        p[too_small] <- 0
        p[too_large] <- 1
        ora_res$pvalue <- p
    } else if (!isTRUE(all.equal(p, ora_res$pvalue))) {
        ora_res$pvalue <- p
    }

    # Calculate p.adjust
    ora_res$p.adjust <- p.adjust(ora_res$pvalue, method=pAdjustMethod)

    # Calculate qvalue
    ora_res$qvalue <- calculate_qvalue(ora_res$pvalue)

    # Calculate zScore
    # Need k, M, n, N    
    N <- length(extID)
    n <- length(intersect(gene, extID))
    k <- ora_res$Count
    
    # M = size of gene set (in universe)
    # Map ID to M
    # ora_res$ID should match names(geneSets)
    M <- sapply(geneSets[ora_res$ID], length)
    
    mu <- M * n / N
    sigma <- mu * (N - n) * (N - M) / N / (N - 1)
    zScore <- (k - mu) / sqrt(sigma)
    ora_res$zScore <- zScore

    # Add Description
    gsid2name <- gson@gsid2name
    if (!is.null(gsid2name) && "ID" %in% names(ora_res)) {
        description <- gsid2name$name[match(ora_res$ID, gsid2name$gsid)]
        na_idx <- is.na(description)
        description[na_idx] <- ora_res$ID[na_idx]
        ora_res$Description <- description
    } 

    if (!"Description" %in% names(ora_res)) {
        ora_res$Description <- ora_res$ID
    }

    # Reorder columns
    expected_cols <- c("ID", "Description", "GeneRatio", "BgRatio", "RichFactor", "FoldEnrichment", "zScore", "pvalue", "p.adjust", "qvalue", "geneID", "Count")
    
    # Ensure all expected columns exist
    missing_cols <- setdiff(expected_cols, names(ora_res))
    if (length(missing_cols) > 0) {
        warning("Missing columns in ORA result: ", paste(missing_cols, collapse=", "), 
                ". Filling with NA. (Note: 'qvalue' may be missing if p-value distribution does not allow q-value calculation)")
        for (col in missing_cols) {
            ora_res[[col]] <- NA
        }
    }

    #other_cols <- setdiff(names(ora_res), expected_cols)
    ora_res <- ora_res[, expected_cols]
    
    # Sort by pvalue
    ora_res <- ora_res[order(ora_res$pvalue), ]
    
    # Set row names
    # Ensure IDs are robust (handle NA and duplicates)
    ids <- as.character(ora_res$ID)
    
    # Handle NA IDs (convert to string "NA")
    if (any(is.na(ids))) {
        warning("NA values detected in gene set IDs. Replacing with string 'NA'.")
        ids[is.na(ids)] <- "NA"
    }
    
    # Handle Duplicate IDs
    if (any(duplicated(ids))) {
        dup_ids <- unique(ids[duplicated(ids)])
        warning("Duplicate gene set IDs detected: ", paste(head(dup_ids), collapse=", "), 
                "... (Total ", length(dup_ids), "). Unique suffixes added.")
        ids <- make.unique(ids)
    }
    
    rownames(ora_res) <- ids

    x <- new("enrichResult",
             result         = ora_res,
             pvalueCutoff   = pvalueCutoff,
             pAdjustMethod  = pAdjustMethod,
             qvalueCutoff   = qvalueCutoff,
             gene           = as.character(gene),
             universe       = extID,
             geneSets       = geneSets,
             organism       = if (!is.null(gson@species)) gson@species else "UNKNOWN",
             keytype        = if (!is.null(gson@keytype)) gson@keytype else "UNKNOWN",
             ontology       = if (!is.null(gson@gsname)) gsub(".*;", "", gson@gsname) else "UNKNOWN",
             readable       = FALSE
             )
             
    return(x)
}

library(testthat)
library(enrichit)

make_enrich_result_accessor_test <- function() {
  result_df <- data.frame(
    ID = c("Path1", "Path2"),
    Description = c("Path 1", "Path 2"),
    GeneRatio = c("2/3", "1/3"),
    BgRatio = c("2/10", "4/10"),
    RichFactor = c(1, 0.25),
    FoldEnrichment = c(5, 1.25),
    pvalue = c(0.001, 0.02),
    p.adjust = c(0.002, 0.03),
    qvalue = c(0.002, 0.03),
    geneID = c("GeneA/GeneB", "GeneC"),
    Count = c(2, 1),
    stringsAsFactors = FALSE
  )
  rownames(result_df) <- result_df$ID

  new("enrichResult",
    result = result_df,
    pvalueCutoff = 1,
    pAdjustMethod = "BH",
    qvalueCutoff = 1,
    organism = "human",
    ontology = "TEST",
    gene = c("GeneA", "GeneB", "GeneC"),
    keytype = "SYMBOL",
    universe = paste0("Gene", LETTERS[1:10]),
    gene2Symbol = character(0),
    geneSets = list(Path1 = c("GeneA", "GeneB"), Path2 = "GeneC"),
    readable = FALSE,
    termsim = matrix(0, nrow = 0, ncol = 0),
    method = "ORA",
    dr = list()
  )
}

make_gsea_result_accessor_test <- function() {
  result_df <- data.frame(
    ID = c("Top", "Bottom"),
    Description = c("Top pathway", "Bottom pathway"),
    setSize = c(2L, 2L),
    enrichmentScore = c(0.7, -0.6),
    NES = c(1.8, -1.5),
    pvalue = c(0.01, 0.03),
    p.adjust = c(0.02, 0.03),
    qvalue = c(0.02, 0.03),
    rank = c(2L, 5L),
    leading_edge = c("tags=100%, list=40%, signal=70%", "tags=100%, list=40%, signal=70%"),
    core_enrichment = c("GeneA/GeneB", "GeneD/GeneE"),
    stringsAsFactors = FALSE
  )
  rownames(result_df) <- result_df$ID

  new("gseaResult",
    result = result_df,
    organism = "human",
    setType = "TEST",
    geneSets = list(Top = c("GeneA", "GeneB"), Bottom = c("GeneD", "GeneE")),
    geneList = c(GeneA = 3, GeneB = 2, GeneC = 1, GeneD = -2, GeneE = -3),
    keytype = "SYMBOL",
    permScores = matrix(0, nrow = 0, ncol = 0),
    params = list(pvalueCutoff = 1, pAdjustMethod = "BH"),
    gene2Symbol = character(0),
    readable = FALSE,
    termsim = matrix(0, nrow = 0, ncol = 0),
    method = "GSEA",
    dr = list()
  )
}

test_that("geneID and geneInCategory return expected feature mappings", {
  enrich_res <- make_enrich_result_accessor_test()
  gsea_res <- make_gsea_result_accessor_test()

  expect_equal(geneID(enrich_res), c("GeneA/GeneB", "GeneC"))
  expect_equal(geneID(gsea_res), c("GeneA/GeneB", "GeneD/GeneE"))

  enrich_cat <- geneInCategory(enrich_res)
  gsea_cat <- geneInCategory(gsea_res)

  expect_equal(enrich_cat$Path1, c("GeneA", "GeneB"))
  expect_equal(gsea_cat$Bottom, c("GeneD", "GeneE"))
})

test_that("gsfilter filters enrichResult by Count and GSSize", {
  enrich_res <- make_enrich_result_accessor_test()

  by_count <- gsfilter(enrich_res, by = "Count", min = 2)
  expect_s4_class(by_count, "enrichResult")
  expect_equal(by_count@result$ID, "Path1")

  by_size <- gsfilter(enrich_res, by = "GSSize", max = 2)
  expect_equal(by_size@result$ID, "Path1")
})

test_that("summary delegates to as.data.frame for enrichResult and gseaResult", {
  enrich_res <- make_enrich_result_accessor_test()
  gsea_res <- make_gsea_result_accessor_test()

  expect_warning(
    expect_equal(summary(enrich_res), as.data.frame(enrich_res)),
    "deprecated"
  )
  expect_warning(
    expect_equal(summary(gsea_res), as.data.frame(gsea_res)),
    "deprecated"
  )
})

---
title: "Introduction to enrichit"
format: 
    html:
        theme: none
        embed-resources: true
        fontsize: 1.1em
        linestretch: 1.7
author: "Guangchuang Yu\\

        School of Basic Medical Sciences, Southern Medical University"
date: "`r Sys.Date()`"
vignette: >
  %\VignetteIndexEntry{enrichit}
  %\VignetteEngine{quarto::html}
  %\VignetteEncoding{UTF-8}
---


# Introduction

Functional enrichment analysis is a staple in bioinformatics for interpreting lists of genes identified from omics experiments. `enrichit` provides fast, C++-based implementations of two of the most widely used methods:

1.  **Over-Representation Analysis (ORA)**
2.  **Gene Set Enrichment Analysis (GSEA)**

The package is designed to be efficient and easy to integrate into existing workflows, with a focus on performance and standardized output formats.

# Installation

You can install `enrichit` from GitHub:

```{r eval=FALSE}
devtools::install_github("YuLab-SMU/enrichit")
```

# Over-Representation Analysis (ORA)

ORA determines whether a set of genes of interest (e.g., differentially expressed genes) is enriched in a known gene set (e.g., a biological pathway) more than would be expected by chance.

## Method

`enrichit` implements ORA using the **hypergeometric distribution** (one-sided Fisher's exact test). The p-value is calculated as the probability of observing at least $k$ genes from the specific gene set in the selected list of $n$ genes, given a background population (universe) of $N$ genes containing $M$ genes from that set.

$$
p = 1 - \sum_{i=0}^{k-1} \frac{\binom{M}{i} \binom{N-M}{n-i}}{\binom{N}{n}}
$$

## Example

```{r}
library(enrichit)

# Simulate a universe of 1000 genes
universe <- paste0("Gene", 1:1000)

# Define gene sets
gene_sets <- list(
  PathwayA = paste0("Gene", 1:50),       # Genes 1-50
  PathwayB = paste0("Gene", 800:850)     # Genes 800-850
)

# Select 'significant' genes (e.g., top 20 genes)
# PathwayA should be enriched
sig_genes <- paste0("Gene", 1:20)

# Run ORA
ora_result <- ora(
  gene = sig_genes,
  gene_sets = gene_sets,
  universe = universe
)

# View results
as.data.frame(ora_result)
```

# Gene Set Enrichment Analysis (GSEA)

GSEA evaluates whether a defined set of genes shows statistically significant, concordant differences between two biological states. Unlike ORA, GSEA uses the entire ranked list of genes, avoiding the need for arbitrary thresholds to select "significant" genes.

## Method

`enrichit` offers a fast C++ implementation of GSEA. It calculates an **Enrichment Score (ES)** that reflects the degree to which a gene set is over-represented at the top or bottom of a ranked list of genes.

The package supports different methods for p-value calculation:

1.  **Multilevel (`method = "multilevel"`)**: This is the default and recommended method. It uses an adaptive multi-level splitting Monte Carlo approach to estimate low p-values efficiently with high accuracy, similar to the `fgsea` package.
2.  **Simple Permutation (`method = "permute"`)**: Standard permutation of gene labels.
3.  **Sample Permutation (`method = "sample"`)**: Random sampling of gene sets (faster but less rigorous for some null hypotheses).

## Example

```{r}
# Generate synthetic ranked gene list
set.seed(42)
geneList <- sort(rnorm(1000), decreasing = TRUE)
names(geneList) <- paste0("Gene", 1:1000)

# Define gene sets
# PathwayTop is enriched at the top (positive ES)
# PathwayBottom is enriched at the bottom (negative ES)
gene_sets <- list(
  PathwayTop = names(geneList)[1:50],
  PathwayBottom = names(geneList)[951:1000],
  PathwayRandom = sample(names(geneList), 50)
)

# Run GSEA using the multilevel method
gsea_result <- gsea(
  geneList = geneList,
  gene_sets = gene_sets,
  method = "multilevel",
  nPerm = 1000,    # Base permutations
  minGSSize = 10,
  maxGSSize = 500
)

# View results
head(gsea_result)
```

# Working with GSON

`enrichit` works seamlessly with `GSON` objects, which are used to store gene set information along with metadata. The `GSON` class is defined in the `r yulab.utils::CRANpkg("gson")` package. It provides a structured way to handle gene sets, including gene identifiers, gene set names, and other associated information.

```{r eval=FALSE}
# Assuming you have a GSON object 'g'
# result <- gsea_gson(geneList = geneList, gson = g)
```

# Session Info

```{r}
sessionInfo()
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/00-AllClasses.R
\docType{class}
\name{mnseaResult-class}
\alias{mnseaResult-class}
\title{Class "mnseaResult"
This class represents the result of multi-layer Network-based Set Enrichment Analysis.}
\description{
Class "mnseaResult"
This class represents the result of multi-layer Network-based Set Enrichment Analysis.
}
\section{Slots}{

\describe{
\item{\code{result}}{enrichment analysis}

\item{\code{organism}}{organism label for the enrichment result}

\item{\code{setType}}{gene set collection type}

\item{\code{geneSets}}{gene sets}

\item{\code{geneList}}{order rank geneList}

\item{\code{keytype}}{ID type of gene}

\item{\code{permScores}}{permutation score matrix inherited from \code{gseaResult}}

\item{\code{gene2Symbol}}{gene ID to symbol mapping}

\item{\code{readable}}{logical flag of gene ID in symbol or not.}

\item{\code{termsim}}{Calculation matrix of termsim.}

\item{\code{method}}{Method of termsim.}

\item{\code{params}}{parameters}

\item{\code{dr}}{dimension reduction result}

\item{\code{multilayer_network}}{prepared multi-layer network object.}

\item{\code{layer_scores}}{list of layer-specific diffusion score vectors.}

\item{\code{collapsed_scores}}{numeric vector used for downstream enrichment.}

\item{\code{layer_weights}}{numeric vector of layer weights.}

\item{\code{coupling_table}}{data.frame of inter-layer couplings.}

\item{\code{mode}}{character, "evidence" or "signed".}

\item{\code{iterations}}{integer, the actual number of iterations RWR took to converge.}

\item{\code{restart_prob}}{numeric, the restart probability used in RWR.}

\item{\code{collapse_method}}{character collapse method used on layer scores.}

\item{\code{target_layer}}{optional layer name used for downstream export.}

\item{\code{output_space}}{character output space of collapsed scores.}

\item{\code{pathway_contribution}}{pathway-by-layer contribution table precomputed for explanation.}

\item{\code{feature_contribution}}{feature-by-layer contribution table precomputed for explanation.}
}}

\author{
Guangchuang Yu \url{https://yulab-smu.top}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/multi_omics.R
\name{select_features_for_ora}
\alias{select_features_for_ora}
\title{Select features for ORA}
\usage{
select_features_for_ora(x, cutoff = 0.05, by = c("pvalue", "score"), ...)
}
\arguments{
\item{x}{A structured result from \code{aggregate_omics()} or \code{harmonize_ids()}.}

\item{cutoff}{Numeric, the threshold to apply.}

\item{by}{Character, metric to apply the threshold on. One of "pvalue" or "score".}

\item{...}{Additional arguments.}
}
\value{
A list containing \code{gene} (the selected feature IDs) and \code{universe} (all feature IDs).
}
\description{
Convert continuous aggregated statistics into a discrete list of genes and a universe for Over-Representation Analysis.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/00-AllClasses.R
\docType{class}
\name{nseaResult-class}
\alias{nseaResult-class}
\title{Class "nseaResult"
This class represents the result of Network-based Set Enrichment Analysis (NSEA).}
\description{
Class "nseaResult"
This class represents the result of Network-based Set Enrichment Analysis (NSEA).
}
\section{Slots}{

\describe{
\item{\code{result}}{enrichment analysis}

\item{\code{organism}}{organism label for the enrichment result}

\item{\code{setType}}{gene set collection type}

\item{\code{geneSets}}{gene sets}

\item{\code{geneList}}{order rank geneList}

\item{\code{keytype}}{ID type of gene}

\item{\code{permScores}}{permutation score matrix inherited from \code{gseaResult}}

\item{\code{gene2Symbol}}{gene ID to symbol mapping}

\item{\code{readable}}{logical flag of gene ID in symbol or not.}

\item{\code{termsim}}{Calculation matrix of termsim.}

\item{\code{method}}{Method of termsim.}

\item{\code{params}}{parameters}

\item{\code{dr}}{dimension reduction result}

\item{\code{network}}{sparse matrix or data.frame representing the underlying network.}

\item{\code{diffusion_scores}}{numeric vector of RWR diffusion scores for each node.}

\item{\code{mode}}{character, "evidence" or "signed", describing the RWR propagation mode.}

\item{\code{iterations}}{integer, the actual number of iterations RWR took to converge.}

\item{\code{restart_prob}}{numeric, the restart probability used in RWR.}
}}

\author{
Guangchuang Yu \url{https://yulab-smu.top}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mnsea.R
\name{propagate_multilayer}
\alias{propagate_multilayer}
\title{Propagate signals on a multi-layer network}
\usage{
propagate_multilayer(
  seed_list,
  network,
  mode = c("evidence", "signed"),
  p = 0.5,
  threshold = 1e-09,
  maxIter = 100,
  layer_weights = NULL,
  target_layer = NULL
)
}
\arguments{
\item{seed_list}{named list of named numeric vectors, one per layer.}

\item{network}{a prepared \code{multilayer_network} object.}

\item{mode}{one of "evidence" or "signed".}

\item{p}{restart probability.}

\item{threshold}{convergence threshold.}

\item{maxIter}{maximum number of iterations.}

\item{layer_weights}{optional named numeric vector of layer weights.}

\item{target_layer}{optional layer name to focus on downstream.}
}
\value{
A \code{multilayer_propagation} object.
}
\description{
Propagate signals on a multi-layer network
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.R
\docType{methods}
\name{show}
\alias{show}
\alias{show,nseaResult-method}
\alias{show,mnseaResult-method}
\title{show method}
\usage{
show(object)

show(object)

show(object)

show(object)
}
\arguments{
\item{object}{A \code{enrichResult} instance.}
}
\value{
message

message

message

message
}
\description{
show method for \code{gseaResult} instance

show method for \code{nseaResult} instance

show method for \code{mnseaResult} instance

show method for \code{enrichResult} instance
}
\author{
Guangchuang Yu \url{https://yulab-smu.top}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ora.R
\name{ora}
\alias{ora}
\title{Over-Representation Analysis (ORA)}
\usage{
ora(gene, gene_sets, universe, weight = NULL)
}
\arguments{
\item{gene}{Character vector of differentially expressed genes (or gene list of interest).}

\item{gene_sets}{A named list of gene sets. Each element is a character vector of genes.}

\item{universe}{Character vector of background genes (e.g., all genes in the platform).}

\item{weight}{A named numeric vector of weights for background genes. If provided, Weighted ORA will be performed using Wallenius' noncentral hypergeometric distribution (requires 'BiasedUrn' package). The names should match the universe genes.}
}
\value{
A data.frame with columns:
\item{GeneSet}{Gene set name}
\item{SetSize}{Number of genes in the gene set (intersected with universe)}
\item{DEInSet}{Number of differentially expressed genes in the gene set}
\item{DESize}{Total number of differentially expressed genes in universe}
\item{PValue}{Raw p-value from hypergeometric test}
}
\description{
Perform over-representation analysis using hypergeometric test (Fisher's exact test).
}
\examples{
# Example data
de_genes <- c("Gene1", "Gene2", "Gene3", "Gene4", "Gene5")
all_genes <- paste0("Gene", 1:1000)

gs1 <- paste0("Gene", 1:50)
gs2 <- paste0("Gene", 51:150)
gs3 <- paste0("Gene", 151:300)
gene_sets <- list(Pathway1 = gs1, Pathway2 = gs2, Pathway3 = gs3)

result <- ora(gene=de_genes, gene_sets=gene_sets, universe=all_genes)
head(result)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/setReadable.R
\name{EXTID2NAME}
\alias{EXTID2NAME}
\title{EXTID2NAME}
\usage{
EXTID2NAME(OrgDb, geneID, keytype, toType = "SYMBOL")
}
\arguments{
\item{OrgDb}{OrgDb}

\item{geneID}{entrez gene ID}

\item{keytype}{keytype}

\item{toType}{ID type of the output}
}
\value{
gene symbol
}
\description{
mapping gene ID to gene Symbol
}
\author{
Guangchuang Yu \url{https://yulab-smu.top}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bayes_enrich.R
\name{bayes_enrich}
\alias{bayes_enrich}
\title{Bayesian term selection for enrichment results}
\usage{
bayes_enrich(
  x,
  candidate = c("top", "significant", "all"),
  n_terms = 200,
  by = "p.adjust",
  prior = 0.1,
  false_positive = 0.01,
  false_negative = 0.1,
  n_iter = 5000,
  burnin = 1000,
  thin = 1,
  posterior_cutoff = 0.5,
  seed = NULL,
  verbose = FALSE
)
}
\arguments{
\item{x}{An \code{enrichResult} object, typically from \code{ora_gson()} or a package
that builds on enrichit, such as clusterProfiler.}

\item{candidate}{Candidate terms to include. \code{"significant"} uses
\code{as.data.frame(x)}, \code{"all"} uses \code{x@result}, \code{"top"} uses the top
\code{n_terms} rows from \code{x@result} ordered by \code{by}; or provide a character
vector of term IDs.}

\item{n_terms}{Maximum number of candidate terms when \code{candidate = "top"} or
when more candidates are supplied than this value. Use \code{Inf} to disable.}

\item{by}{Column used to order candidate terms.}

\item{prior}{Prior probability that a term is active.}

\item{false_positive}{Probability of observing a gene not covered by active
terms.}

\item{false_negative}{Probability of missing a gene covered by active terms.}

\item{n_iter}{Total number of MCMC iterations.}

\item{burnin}{Number of initial iterations discarded.}

\item{thin}{Keep one sample every \code{thin} iterations after burn-in.}

\item{posterior_cutoff}{Terms with posterior greater than or equal to this
value are marked active.}

\item{seed}{Optional random seed.}

\item{verbose}{Print sampler progress.}
}
\value{
The input \code{enrichResult} object with additional columns in
\verb{@result}: \code{posterior}, \code{posterior_odds}, \code{bayes_rank},
\code{bayes_active}, \code{bayes_covered_gene}, and \code{bayes_covered_count}.
}
\description{
\code{bayes_enrich()} adds a model-based selection layer on top of ORA results.
It estimates the posterior probability that each candidate term is an
active biological program explaining the observed input genes.
}
\details{
The implementation uses a lightweight Metropolis-Hastings sampler over
binary latent term states. Given active terms, each gene is modeled as
observed with probability \code{1 - false_negative} if covered by at least one
active term, and with probability \code{false_positive} otherwise. The prior
probability that a candidate term is active is \code{prior}.

This is intended as a result-compression and interpretation layer, not as a
replacement for ORA p-values.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/multi_omics.R
\name{aggregate_omics}
\alias{aggregate_omics}
\title{Aggregate multi-omics gene/protein-level statistics}
\usage{
aggregate_omics(
  x,
  method = c("fisher", "stouffer", "brown", "mean", "weighted_mean", "max_abs"),
  input = c("pvalue", "signed_score"),
  feature_type = "gene",
  conflict_policy = c("keep_all", "strict", "penalty"),
  ...
)
}
\arguments{
\item{x}{A list of named numeric vectors, a data.frame, or a matrix. Row names (or names for vectors) must represent feature IDs.}

\item{method}{Character, aggregation method. One of "fisher", "stouffer", "mean", or "max_abs".}

\item{input}{Character, input type. One of "pvalue" or "signed_score".}

\item{feature_type}{Character, type of the features (e.g., "gene", "protein"). Default is "gene".}

\item{conflict_policy}{Character, strategy to handle directional conflicts when input is "signed_score". One of "keep_all" (default, ignore conflicts), "strict" (set to NA if any signs conflict), or "penalty" (divide final score by 2 if signs conflict).}

\item{...}{Additional arguments.}
}
\value{
An object of class \code{omics_aggregated} containing \code{score}, \code{pvalue} (if input is "pvalue"), \code{input_type}, \code{feature_type}, and \code{feature_id}.
}
\description{
Aggregate multi-omics or multi-source statistics into a unified object for downstream enrichment analysis.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utilities.R
\name{enrichit_params}
\alias{enrichit_params}
\title{Common parameters for enrichit functions}
\arguments{
\item{geneList}{A named numeric vector of gene statistics (e.g., log fold change), ranked in descending order.}

\item{gene_sets}{A named list of gene sets. Each element is a character vector of genes.}

\item{nPerm}{Number of permutations for p-value calculation (default: 1000).}

\item{exponent}{Weighting exponent for enrichment score (default: 1.0).}

\item{minGSSize}{minimal size of each geneSet for analyzing}

\item{maxGSSize}{maximal size of each geneSet for analyzing}

\item{pvalueCutoff}{P-value cutoff.}

\item{pAdjustMethod}{P-value adjustment method (e.g., "BH").}

\item{verbose}{Logical. Print progress messages.}

\item{gson}{A GSON object containing gene set information.}

\item{method}{Permutation method.}

\item{adaptive}{Logical. Use adaptive permutation.}

\item{minPerm}{Minimum permutations for adaptive mode.}

\item{maxPerm}{Maximum permutations for adaptive mode.}

\item{pvalThreshold}{P-value threshold for early stopping.}
}
\description{
Common parameters for enrichit functions
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gsea.R
\name{gsea_gson}
\alias{gsea_gson}
\title{gsea_gson}
\usage{
gsea_gson(
  geneList,
  gson,
  weight = NULL,
  nPerm = 1000,
  exponent = 1,
  minGSSize = 10,
  maxGSSize = 500,
  pvalueCutoff = 0.05,
  pAdjustMethod = "BH",
  method = "multilevel",
  adaptive = FALSE,
  minPerm = 101,
  maxPerm = 1e+05,
  pvalThreshold = 0.1,
  verbose = TRUE,
  ...
)
}
\arguments{
\item{geneList}{A named numeric vector of gene statistics (e.g., log fold change), ranked in descending order.}

\item{gson}{A GSON object containing gene set information.}

\item{weight}{A named numeric vector of weights for genes.}

\item{nPerm}{Number of permutations for p-value calculation (default: 1000).}

\item{exponent}{Weighting exponent for enrichment score (default: 1.0).}

\item{minGSSize}{minimal size of each geneSet for analyzing}

\item{maxGSSize}{maximal size of each geneSet for analyzing}

\item{pvalueCutoff}{P-value cutoff.}

\item{pAdjustMethod}{P-value adjustment method (e.g., "BH").}

\item{method}{Permutation method.}

\item{adaptive}{Logical. Use adaptive permutation.}

\item{minPerm}{Minimum permutations for adaptive mode.}

\item{maxPerm}{Maximum permutations for adaptive mode.}

\item{pvalThreshold}{P-value threshold for early stopping.}

\item{verbose}{Logical. Print progress messages.}

\item{...}{Additional parameters passed to gsea()}
}
\value{
gseaResult object
}
\description{
generic function for gene set enrichment analysis
}
\author{
Guangchuang Yu
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mnsea.R
\name{prepare_multilayer_network}
\alias{prepare_multilayer_network}
\title{Prepare multi-layer network for repeated propagation}
\usage{
prepare_multilayer_network(
  networks,
  couplings,
  directed = FALSE,
  intra_normalize = "column",
  inter_normalize = "column",
  interlayer_strength = 1,
  layer_order = names(networks)
)
}
\arguments{
\item{networks}{named list of layer-specific networks.}

\item{couplings}{data.frame of inter-layer edges with columns
\code{from_layer}, \code{from_id}, \code{to_layer}, \code{to_id}, and optional \code{weight}.}

\item{directed}{logical, whether the multi-layer graph is directed.}

\item{intra_normalize}{one of "column", "row", or "none".}

\item{inter_normalize}{one of "column", "row", or "none".}

\item{interlayer_strength}{numeric scalar used to scale all coupling edges.}

\item{layer_order}{explicit layer order. Defaults to \code{names(networks)}.}
}
\value{
A \code{multilayer_network} object.
}
\description{
Prepare multi-layer network for repeated propagation
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mnsea_contribution.R
\name{get_mnsea_contribution}
\alias{get_mnsea_contribution}
\title{Get cached contribution tables from a \code{mnseaResult}}
\usage{
get_mnsea_contribution(res, pathway_id = NULL, level = c("pathway", "feature"))
}
\arguments{
\item{res}{A \code{mnseaResult} object.}

\item{pathway_id}{Optional pathway ID. If \code{NULL}, returns all pathways for
\code{level = "pathway"} and uses the top pathway for \code{level = "feature"}.}

\item{level}{One of \code{"pathway"} or \code{"feature"}.}
}
\value{
A data.frame containing cached contribution information.
}
\description{
Get cached contribution tables from a \code{mnseaResult}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/omics_contribution.R
\name{get_omics_contribution}
\alias{get_omics_contribution}
\title{Get gene-level omics contribution for a specific pathway}
\usage{
get_omics_contribution(res, agg, pathway_id = NULL)
}
\arguments{
\item{res}{An \code{enrichResult} or \code{gseaResult} object.}

\item{agg}{An \code{omics_aggregated} object from \code{aggregate_omics()}.}

\item{pathway_id}{Character, the ID of the pathway to extract. If NULL, the top pathway is used.}
}
\value{
A data.frame containing the genes, their original omics statistics, the aggregated score, and whether they belong to the core enrichment.
}
\description{
Extract the original multi-omics statistics for genes in a specific enriched pathway.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gsea.R
\name{gsea}
\alias{gsea}
\title{Gene Set Enrichment Analysis (GSEA)}
\usage{
gsea(
  geneList,
  gene_sets,
  weight = NULL,
  minGSSize = 10,
  maxGSSize = 500,
  nPerm = 1000,
  exponent = 1,
  method = "multilevel",
  adaptive = FALSE,
  minPerm = 101,
  maxPerm = 1e+05,
  pvalThreshold = 0.1,
  eps = 1e-10,
  sampleSize = 101,
  seed = FALSE,
  nPermSimple = 1000,
  scoreType = "std",
  verbose = TRUE
)
}
\arguments{
\item{geneList}{A named numeric vector of gene statistics (e.g., log fold change), ranked in descending order.}

\item{gene_sets}{A named list of gene sets. Each element is a character vector of genes.}

\item{weight}{A named numeric vector of weights for genes. The names should match the names of geneList. If provided, the geneList will be multiplied by the weight and resorted before GSEA (default: NULL).}

\item{minGSSize}{minimal size of each geneSet for analyzing}

\item{maxGSSize}{maximal size of each geneSet for analyzing}

\item{nPerm}{Number of permutations for p-value calculation (default: 1000).}

\item{exponent}{Weighting exponent for enrichment score (default: 1.0).}

\item{method}{Permutation method.}

\item{adaptive}{Logical. Use adaptive permutation.}

\item{minPerm}{Minimum permutations for adaptive mode.}

\item{maxPerm}{Maximum permutations for adaptive mode.}

\item{pvalThreshold}{P-value threshold for early stopping.}

\item{eps}{Epsilon for multilevel methods (default: 1e-10). Sets the smallest p-value that can be estimated.}

\item{sampleSize}{Sample size for multilevel methods (default: 101).}

\item{seed}{Random seed for reproducibility (default: FALSE). If FALSE, a random seed is generated.}

\item{nPermSimple}{Number of permutations for the simple method (default: 1000).}

\item{scoreType}{Type of enrichment score calculation: "std", "pos", "neg" (default: "std").}

\item{verbose}{Logical. Print progress messages.}
}
\value{
A data.frame with columns:
\itemize{
\item \strong{ID}: Gene set name
\item \strong{enrichmentScore}: Enrichment Score
\item \strong{NES}: Normalized Enrichment Score
\item \strong{pvalue}: Empirical p-value from permutation test
\item \strong{setSize}: Size of the gene set (number of genes found in geneList)
\item \strong{nPerm}: (adaptive mode only) Actual number of permutations used
\item \strong{rank}: Rank at which the maximum enrichment score is attained
\item \strong{leading_edge}: Leading edge statistics (tags, list, signal)
\item \strong{core_enrichment}: Genes in the leading edge, separated by '/'
}
}
\description{
Perform Gene Set Enrichment Analysis (GSEA) using a ranked gene list.
}
\examples{
# Example data
stats <- rnorm(1000)
names(stats) <- paste0("Gene", 1:1000)
stats <- sort(stats, decreasing = TRUE)

gs1 <- paste0("Gene", 1:50)
gs2 <- paste0("Gene", 500:550)
gene_sets <- list(Pathway1 = gs1, Pathway2 = gs2)

# Use default fixed permutation method
result <- gsea(geneList=stats, gene_sets=gene_sets, nPerm=100)

# Use adaptive permutation for more accurate p-values
\donttest{
result_adaptive <- gsea(geneList=stats, gene_sets=gene_sets, adaptive=TRUE)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mnsea.R
\name{collapse_multilayer_scores}
\alias{collapse_multilayer_scores}
\title{Collapse multi-layer diffusion scores}
\usage{
collapse_multilayer_scores(
  x,
  collapse = c("weighted_mean", "sum", "mean", "max_abs"),
  layer_weights = NULL,
  output_space = c("union", "gene"),
  mapping = NULL,
  target_layer = NULL
)
}
\arguments{
\item{x}{result from \code{propagate_multilayer()}.}

\item{collapse}{one of "weighted_mean", "sum", "mean", or "max_abs".}

\item{layer_weights}{optional named numeric vector used when
\code{collapse = "weighted_mean"}.}

\item{output_space}{one of "union" or "gene".}

\item{mapping}{optional mapping data.frame with \code{source_id}, \code{target_id}, and
optional \code{layer} columns.}

\item{target_layer}{optional layer name to extract before collapsing.}
}
\value{
A \code{multilayer_collapsed} object with a \code{score} vector.
}
\description{
Collapse multi-layer diffusion scores
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mnsea_subnetwork.R
\name{extract_mnsea_subnetwork}
\alias{extract_mnsea_subnetwork}
\title{Extract pathway subnetwork data from a \code{mnseaResult}}
\usage{
extract_mnsea_subnetwork(
  res,
  pathway_id = NULL,
  include_couplings = TRUE,
  include_isolated = TRUE
)
}
\arguments{
\item{res}{A \code{mnseaResult} object.}

\item{pathway_id}{Optional pathway ID. If \code{NULL}, the top pathway is used.}

\item{include_couplings}{Logical, whether to include inter-layer coupling
edges. Default is \code{TRUE}.}

\item{include_isolated}{Logical, whether to keep nodes without retained
edges. Default is \code{TRUE}.}
}
\value{
A list with \code{pathway}, \code{layer_contribution}, \code{nodes}, and \code{edges}.
}
\description{
Extract pathway subnetwork data from a \code{mnseaResult}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ora_gson.R
\name{ora_gson}
\alias{ora_gson}
\title{ora-gson}
\usage{
ora_gson(
  gene,
  pvalueCutoff,
  pAdjustMethod = "BH",
  universe = NULL,
  weight = NULL,
  minGSSize = 10,
  maxGSSize = 500,
  qvalueCutoff = 0.2,
  gson
)
}
\arguments{
\item{gene}{a vector of entrez gene id.}

\item{pvalueCutoff}{P-value cutoff.}

\item{pAdjustMethod}{P-value adjustment method (e.g., "BH").}

\item{universe}{background genes, default is the intersection of the 'universe' with genes that have annotations.
Users can set \code{options(enrichment_force_universe = TRUE)} to force the 'universe' untouched.}

\item{weight}{A named numeric vector of weights for background genes. If provided, Weighted ORA will be performed.}

\item{minGSSize}{minimal size of each geneSet for analyzing}

\item{maxGSSize}{maximal size of each geneSet for analyzing}

\item{qvalueCutoff}{cutoff of qvalue}

\item{gson}{A GSON object containing gene set information.}
}
\value{
A \code{enrichResult} instance.
}
\description{
interal method for enrichment analysis
}
\details{
using the hypergeometric model
}
\author{
Guangchuang Yu \url{https://yulab-smu.top}
}
\keyword{manip}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mnsea.R
\name{mnsea}
\alias{mnsea}
\title{Multi-layer Network-based Gene Set Enrichment Analysis}
\usage{
mnsea(
  seed_list,
  networks,
  couplings,
  gene_sets,
  mode = c("evidence", "signed"),
  layer_weights = NULL,
  collapse = c("weighted_mean", "sum", "mean", "max_abs"),
  target_layer = NULL,
  output_space = c("union", "gene"),
  p = 0.5,
  interlayer_strength = 1,
  specific_weight = FALSE,
  minGSSize = 10,
  maxGSSize = 500,
  threshold = 1e-09,
  maxIter = 100,
  verbose = TRUE,
  ...
)
}
\arguments{
\item{seed_list}{named list of named numeric vectors, one per layer.}

\item{networks}{named list of layer-specific networks.}

\item{couplings}{data.frame of inter-layer edges.}

\item{gene_sets}{list of gene sets.}

\item{mode}{one of "evidence" or "signed".}

\item{layer_weights}{optional named numeric vector.}

\item{collapse}{one of "weighted_mean", "sum", "mean", or "max_abs".}

\item{target_layer}{optional layer name to export scores from.}

\item{output_space}{one of "union" or "gene".}

\item{p}{restart probability.}

\item{interlayer_strength}{global scaling factor for coupling edges.}

\item{specific_weight}{logical.}

\item{minGSSize}{minimal size of each gene set.}

\item{maxGSSize}{maximal size of genes annotated for testing.}

\item{threshold}{convergence threshold.}

\item{maxIter}{maximal number of iterations.}

\item{verbose}{logical.}

\item{...}{additional arguments passed to \code{gsea()}.}
}
\value{
A \code{mnseaResult} object.
}
\description{
Multi-layer Network-based Gene Set Enrichment Analysis
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gseaScores.R
\name{gseaScores}
\alias{gseaScores}
\title{Calculate GSEA Running Enrichment Scores}
\usage{
gseaScores(geneList, geneSet, exponent = 1, fortify = FALSE)
}
\arguments{
\item{geneList}{a named numeric vector of gene statistics (e.g., t-statistics or log-fold changes), sorted in decreasing order.}

\item{geneSet}{a character vector of gene IDs belonging to the gene set.}

\item{exponent}{a numeric value defining the weight of the running enrichment score. Default is 1.}

\item{fortify}{logical. If TRUE, returns a data frame with columns \code{x}, \code{runningScore}, and \code{position}.
If FALSE (default), returns the enrichment score (ES).}
}
\value{
If \code{fortify = TRUE}, a data frame containing the running enrichment scores and positions.
If \code{fortify = FALSE}, a numeric value representing the Enrichment Score (ES).
}
\description{
Calculate GSEA Running Enrichment Scores
}
\author{
Guangchuang Yu
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/omics_contribution.R
\name{classify_omics_pattern}
\alias{classify_omics_pattern}
\title{Classify pathway-level multi-omics patterns}
\usage{
classify_omics_pattern(
  merged_res,
  single_res,
  p_cutoff = 0.05,
  by = "p.adjust"
)
}
\arguments{
\item{merged_res}{An \code{enrichResult} or \code{gseaResult} object from the merged multi-omics analysis.}

\item{single_res}{A named list of \code{enrichResult} or \code{gseaResult} objects from single-omics analyses.}

\item{p_cutoff}{Numeric, the significance cutoff. Default is 0.05.}

\item{by}{Character, the column to use for significance threshold. Default is "p.adjust".}
}
\value{
The \code{merged_res} object with an additional column \code{Omics_Pattern} in its result data.frame.
}
\description{
Compare merged enrichment results with single-omics enrichment results to classify the contribution pattern of each pathway.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/enrichit-package.R
\docType{package}
\name{enrichit-package}
\alias{enrichit}
\alias{enrichit-package}
\title{enrichit: 'C++' Implementations of Functional Enrichment Analysis}
\description{
Fast implementations of functional enrichment analysis methods using 'C++' via 'Rcpp'. Currently provides Over-Representation Analysis (ORA), Gene Set Enrichment Analysis (GSEA), Weighted Enrichment Analysis for ORA and GSEA, Network-based Set Enrichment Analysis (NSEA), multi-layer network-based enrichment, and multi-omics integration workflows. Additional features include early fusion at the feature level, late fusion at the pathway level, multi-omics contribution tracing, topology-aware explanation helpers, Bayesian term selection, and extremely fast Random Walk with Restart (RWR) using 'RcppEigen'. The enrichment methods build on GSEA by Subramanian et al. (2005) \doi{10.1073/pnas.0506580102}, the multilevel strategy derived from 'fgsea' by Korotkevich et al. (2021) \doi{10.1101/060012}, and network-based enrichment ideas described by Glaab et al. (2012) \doi{10.1093/bioinformatics/bts389}.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://yulab-smu.top/biomedical-knowledge-mining-book/}
}

}
\author{
\strong{Maintainer}: Guangchuang Yu \email{guangchuangyu@gmail.com}

Authors:
\itemize{
  \item Guangchuang Yu \email{guangchuangyu@gmail.com}
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/multi_omics.R
\name{harmonize_ids}
\alias{harmonize_ids}
\title{Harmonize feature IDs to a target space}
\usage{
harmonize_ids(
  x,
  mapping,
  from = "protein",
  to = "gene",
  collapse = c("max_abs", "mean", "min_p")
)
}
\arguments{
\item{x}{A structured result from \code{aggregate_omics()}.}

\item{mapping}{A data.frame with \code{source_id} and \code{target_id} columns.}

\item{from}{Character, source feature type. Default is "protein".}

\item{to}{Character, target feature type. Default is "gene".}

\item{collapse}{Character, method to collapse multiple source IDs mapped to a single target ID. One of "max_abs", "mean", or "min_p".}
}
\value{
A harmonized \code{omics_aggregated} object.
}
\description{
Map protein-level or other feature-level statistics to a unified gene-level space.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bayes_enrich.R
\name{bayes_summary}
\alias{bayes_summary}
\title{Summarize Bayesian enrichment results}
\usage{
bayes_summary(x, active = FALSE, n = Inf)
}
\arguments{
\item{x}{An \code{enrichResult} object processed by \code{bayes_enrich()}.}

\item{active}{Logical. If \code{TRUE}, keep only rows with \code{bayes_active = TRUE}.}

\item{n}{Number of rows to return. Use \code{Inf} to return all rows.}
}
\value{
A data frame ordered by decreasing \code{posterior}.
}
\description{
Return a data frame sorted by posterior probability from a result processed
by \code{bayes_enrich()}. This is a convenience wrapper around sorting
\code{as.data.frame(x)} by decreasing \code{posterior}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nsea.R
\name{nsea_gson}
\alias{nsea_gson}
\title{Network-based GSEA using a GSON object}
\usage{
nsea_gson(
  geneList,
  network,
  gson,
  mode = c("evidence", "signed"),
  p = 0.5,
  specific_weight = FALSE,
  minGSSize = 10,
  maxGSSize = 500,
  threshold = 1e-09,
  maxIter = 100,
  verbose = TRUE,
  ...
)
}
\arguments{
\item{geneList}{named numeric vector. In "evidence" mode, must be non-negative. In "signed" mode, can contain both positive and negative values.}

\item{network}{edge list (data.frame) or sparse matrix.}

\item{gson}{a GSON object.}

\item{mode}{character, either "evidence" (default) or "signed".}

\item{p}{restart probability for RWR (default is 0.5).}

\item{specific_weight}{logical, whether to apply gene specificity weighting (TF-IDF style) based on gene frequencies in the GSON object. Default is FALSE.}

\item{minGSSize}{minimal size of each gene set for analyzing. default here is 10.}

\item{maxGSSize}{maximal size of genes annotated for testing. default here is 500.}

\item{threshold}{convergence threshold for RWR (default is 1e-9).}

\item{maxIter}{maximal number of RWR iterations (default is 100).}

\item{verbose}{logical, print messages.}

\item{...}{other arguments passed to \code{gsea_gson()}.}
}
\value{
A \code{nseaResult} object.
}
\description{
Network-based GSEA using a GSON object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/setReadable.R
\name{setReadable}
\alias{setReadable}
\title{setReadable}
\usage{
setReadable(x, OrgDb, keyType = "auto", toType = "SYMBOL")
}
\arguments{
\item{x}{enrichResult Object}

\item{OrgDb}{OrgDb}

\item{keyType}{keyType of gene}

\item{toType}{ID type of the output}
}
\value{
enrichResult Object
}
\description{
mapping geneID to gene Symbol
}
\author{
Guangchuang Yu
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/00-AllClasses.R
\docType{class}
\name{compareClusterResult-class}
\alias{compareClusterResult-class}
\alias{show,compareClusterResult-method}
\alias{summary,compareClusterResult-method}
\alias{plot,compareClusterResult-method}
\title{Class "compareClusterResult"
This class represents the comparison result of gene clusters by GO
categories at specific level or GO enrichment analysis.}
\description{
Class "compareClusterResult"
This class represents the comparison result of gene clusters by GO
categories at specific level or GO enrichment analysis.
}
\section{Slots}{

\describe{
\item{\code{compareClusterResult}}{cluster comparing result}

\item{\code{geneClusters}}{a list of genes}

\item{\code{fun}}{one of groupGO, enrichGO and enrichKEGG}

\item{\code{gene2Symbol}}{gene ID to Symbol}

\item{\code{keytype}}{Gene ID type}

\item{\code{readable}}{logical flag of gene ID in symbol or not.}

\item{\code{.call}}{function call}

\item{\code{termsim}}{Similarity between term}

\item{\code{method}}{method of calculating the similarity between nodes}

\item{\code{dr}}{dimension reduction result}

\item{\code{organism}}{organism}
}}

\seealso{
\code{\linkS4class{enrichResult}}
}
\author{
Guangchuang Yu \url{https://yulab-smu.top}
}
\keyword{classes}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/multi_omics.R
\name{aggregate_enrichment}
\alias{aggregate_enrichment}
\title{Aggregate multiple enrichment results (Late Fusion)}
\usage{
aggregate_enrichment(res_list, method = c("brown", "fisher", "stouffer"), ...)
}
\arguments{
\item{res_list}{A named list of enrichment result objects (e.g., \code{enrichResult}, \code{gseaResult}, \code{nseaResult}).}

\item{method}{Character, aggregation method for p-values. One of "brown", "fisher", or "stouffer".}

\item{...}{Additional arguments passed to \code{aggregate_omics} (e.g., \code{cov_matrix} for Brown's method).}
}
\value{
An \code{enrichResult} object containing the aggregated p-values, FDR, and combined gene lists.
}
\description{
Combine pathway-level enrichment results from multiple omics or independent analyses.
P-values of identical pathways are merged using statistical methods (e.g., Brown's method).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nsea.R
\name{prepare_network}
\alias{prepare_network}
\title{Prepare network for repeated NSEA runs}
\usage{
prepare_network(network, directed = FALSE, normalize = "column")
}
\arguments{
\item{network}{edge list (data.frame with 2 or 3 columns) or sparse matrix.}

\item{directed}{logical, whether the network is directed. Default is FALSE.}

\item{normalize}{one of "column", "row", or "none". Default is "column".}
}
\value{
A sparse matrix (dgCMatrix) that has been properly formatted and normalized.
}
\description{
Prepare network for repeated NSEA runs
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gsfilter.R
\name{gsfilter}
\alias{gsfilter}
\title{gsfilter}
\usage{
gsfilter(x, by = "GSSize", min = NA, max = NA)
}
\arguments{
\item{x}{instance of enrichResult or compareClusterResult}

\item{by}{one of 'GSSize' or 'Count'}

\item{min}{minimal size}

\item{max}{maximal size}
}
\value{
update object
}
\description{
filter enriched result by gene set size or gene count
}
\author{
Guangchuang Yu
}
